/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team
  * July, 2002 
  */
 

package lotos;


import lotos.*; 
import java.util.*;
import java.lang.*;

/**
 * This class simulate the 802.11 protocol under ad hoc mode
 * Currenly only the broadcast funcation (basic DCF access) is 
 * implemented
 * 
 * @author Guangyu Dong, Jan 6, 2004
 */
 
interface MAC_Callback{
	void messageSentFromMAC(SimEventMsg msg);
	void messageReceivedToMAC(SimEventMsg msg);
}

public class AdhocMAC implements MACInterface{
		
	/** Parameters used in 802.11 */
	/** Short Interframe Space */
	final static long SIFS=10;
	
	/** Distributed Interframe Space */
	final static long DIFS=50;
	/** The Backoff Time Slot */
	final static long SLOT_TIME=20;
	
	/** The Time needed to Sense the Signal */
	final static long SENSING_TIME=5;
	
	/** Maximum Contention Window Size */
	final static int CWmax=1023;
	/** Minimum Contention Window Size */
	final static int CWmin=31;
	
	/** Current Contention Window Size */
	int current_cw=31;
	
	/** Size of Outgoing Buffer*/
	final static int BUFFER_SIZE=1000;
	
	/** Outgoing Buffer */
	Vector msg_queue;
	
	/** Packet Speed (m/s): distance between node/The time between a bit was sent out and it's received */
	final static float PACKET_SPEED=WirelessPhysicalLayer.PACKET_SPEED;
	/** Transmission Range (m)*/
	final static float TRANS_RADIUS=100f;
	
	/** The Timer Index for Backoff Event */
	final static int BACKOFF_TIMER=1;
	/** The Timer Index for DIFS Event */
	final static int DIFS_TIMER=2;
	/** The Timer Index for Event Sending Out Packet */
	final static int SENDOUT_TIMER=3;
	
	/** The next DIFS Event.
	 *  Used to eliminate old events that has not been executed 
	 */
	SimEventMAC nextDIFS;
	/** The next Backoff Event 
	 *  Used to eliminate old events that has not been executed
	 */
	SimEventMAC nextBackoff;
	
	/** Callback object for sending out incoming message */
	MAC_Callback callback;
	
	/** the overlay network running the node*/
	OverlayNetwork net;
	/** the node running this MAC protocol */
	OverlayNode node;
	
	/** A random number generator */
	static Random ran;
	
	/* Constructor. 
	 * @param _node the OverlayNode running this MAC
	 */
    public AdhocMAC(OverlayNode _node){
		net=_node.net;
		node=_node;
		
		if(ran==null)
			ran=new Random();
		
		
		MACInit();
	}
	/* 802.11 functions */
	
	/* @return current CW value */
	int getCW(){
		return current_cw;
	}
	
	boolean cw_set=false;
	/* Reset the current CW to the minimum value */
	void resetCW(){
		current_cw=CWmin;	
		
		//lotusDebug.errorln("reset CW="+getCW());
	}
	/* Calculate the next CW after detecting a contention */
	void nextCW(){
		int new_cw=(current_cw*2+1);
		if(new_cw<=CWmax)current_cw=new_cw;	
		
		//lotusDebug.errorln("next CW="+getCW());
	}
	
	int backoffTimer=0;
	/*
	 *  Backoff timer is initiated randomly and
	 *  decreased every slot time until it reaches 0
	 * 
	 *  @return return the next backoff timer 
	 */
	int getNextBackoffTimer(){
		if(backoffTimer<=0)backoffTimer=getRandomBackoffTime();
		else backoffTimer--;
		
		return backoffTimer;	
	}
	
	/*
	 * @return a randomly generated backoff time
	 */
	int getRandomBackoffTime(){
		//lotusDebug.errorln("CW="+getCW());
		return ran.nextInt(getCW()+1);	
	}
	/*
	 * Send out a message if the buffer is not empty
	 * 
	 * @return whether any message has been sent
	 */
	 
	boolean sendOutMessage(){
		
		
		if(msg_queue.size()>0){
			SimEventMsg msg=(SimEventMsg)msg_queue.elementAt(0);
			msg_queue.remove(0);
			
			//lotusDebug.errorln("Send out message");
			
			node.net.wl_phyLayer.sendMsg(msg);
			if(callback!=null)callback.messageSentFromMAC(msg);
			
			//sending success
			//set DIFS timers of all neighbors
			ms_timeLastSent = ms_timeLastReceived;
			resetCW();
			cw_set=false;
			return true;
			
			
			
		}else {
			//lotusDebug.errorln("No message waiting");
		}
		return false;
	}
	/*
	 * @see MACInterface#getHeaderSize()
	 */
	public int getHeaderSize(){
		return 34;
	}
	
	long ms_timeLastReceived=0;
	long ms_timeLastSent=0;
	long ms_timeLastSensed=0;
	/* when a signal is sensed */
	public void senseMessage(long ms_time_start, long ms_time_end){
		if(ms_timeLastSensed<ms_time_end){
			ms_timeLastSensed = ms_time_end;
			insertDIFSEvent(ms_time_end-simTime());
		}
	}
	
	/*
	 * @see MACInterface#recordMessage(long, long)
	 */
	public void recordMessage(long ms_time_start, long ms_time_end){
		
		//lotusDebug.errorln("record message");
		
		if(ms_timeLastReceived > ms_time_start)
		{
			//lotusDebug.errorln("collision happend at receiver");
			//collision happens at the receiver
			//and the sender will not know
			//collision = true;
			
			if(ms_timeLastSent > ms_time_start){
				//sending fails
				//lotusDebug.errorln("Sending collision detected.");
			}
			
			
			node.wl_lastMsgCollision=true;
			if(ms_timeLastReceived<ms_time_end){
				ms_timeLastReceived = ms_time_end;
				insertDIFSEvent(ms_time_end-simTime());
			}
			//break;
		}else{
			//lotusDebug.errorln("no collisino");
			//record the message end time on receiver
			ms_timeLastReceived = ms_time_end;
			node.wl_lastMsgCollision = false;
			insertDIFSEvent(ms_time_end-simTime());
		}
		
	}
	
	/* handle the DIFS event */
	void handleDIFS(){
		
		long cf_time1=ms_timeLastReceived-simTime()+DIFS;
		long cf_time2=ms_timeLastSensed-simTime()+DIFS;
		
		if(cf_time1<=0 && cf_time2<=0){
			//no signal for a period of DIFS	
			//handle DIFS
			
			//lotusDebug.errorln("Medium is free");
			
			int bo_timer=getNextBackoffTimer();
			
			//lotusDebug.errorln("backoff timer "+bo_timer);
			
			if(bo_timer==0){
				//send message
				insertSendoutEvent();
				
			}else{
				//enter the backoff phase
				insertBackoffEvent();
			}
		}else{
			//lotusDebug.errorln("Medium is not free");
			
		}
		
	}
	/* handle the backoff event */
	void handleBackoff(){
		
		handleDIFS();
		
	}
	
	/* handle the SENDOUT event */
	void handleSendout(){
		sendOutMessage();
	}
	
	
	/* 
	 * insert the point that the sending ends and the backoff 
	 * should be started. 
	 * Both sender and receiver should be inserted this point.
	 */
	void insertDIFSEvent(long trans_time){
		//lotusDebug.errorln("Insert the next DIFS");
		nextDIFS=node.setMSMACTimer(trans_time+DIFS, DIFS_TIMER);
		
	}
	
	boolean idle=false;
	/*
	 * insert an backoff event happening at a slottime from now
	 */
	void insertBackoffEvent(){
		if(queueEmpty()){
			idle=true;
		}else{
			//lotusDebug.errorln("Insert the next backoff");
			nextBackoff=node.setMSMACTimer(SLOT_TIME, BACKOFF_TIMER);
			idle=false;
		}
	}
	/*
	 * Insert a event to send out message
	 */
	void insertSendoutEvent(){
		node.setMSMACTimer(SENSING_TIME, SENDOUT_TIMER);
	}
		
	/* message buffer functions */
	boolean queueEmpty(){
		return msg_queue.size()<=0;	
	}
	void insertMsg(SimEventMsg msg){
		if(msg_queue.size()<BUFFER_SIZE){
			msg_queue.addElement(msg);	
		}else 
			lotusDebug.errorln("message dropped due to full buffer");
	}
	SimEventMsg getNextPendingMsg(){
		if(msg_queue.size()>0){
			SimEventMsg msg=(SimEventMsg)msg_queue.elementAt(0);
			msg_queue.remove(0);
			return msg;
		}else return null;	
	}
	
	/*
	 * @return current time
	 */
	long simTime(){
		return net.getMSSimTime();
	}
	
	/*
	 * @see MACInterface#sendMsg(SimEventMsg)
	 */
	public boolean sendMsg(SimEventMsg msg){
		//debug
		////lotusDebug.errorln("insert a message");
		msg.setMSTime(simTime());
		
		insertMsg(msg);
		if(idle){
			idle=false;
			
			long channel_idle_time= (simTime()-(ms_timeLastReceived+DIFS));
			if(channel_idle_time>=0){
				long next_bo_time= SLOT_TIME - (channel_idle_time%SLOT_TIME) ;
				nextBackoff=node.setMSMACTimer(next_bo_time, BACKOFF_TIMER);
				//handleBackoff();
			}
		}
		
		return true;
	}
	
	/*
	 * @see MACInterface#MACInit()
	 */
	public void MACInit(){
		msg_queue=new Vector();
		
		nextDIFS = null;
		nextBackoff = null;
		current_cw=31;
		backoffTimer=0;
		idle=false;
		cw_set=false;
		
		ms_timeLastSent=-1;
		ms_timeLastReceived=-1;
		ms_timeLastSensed=-1;
		node.wl_lastMsgCollision=false;
		
		//check if some transmision is going around
		/*
		Vector neighbors = node.getWirelessNeighbors();
		if(neighbors!=null){
	   	   	for(int i=0;i<neighbors.size();i++) {
	   	   		OverlayNode nnode = (OverlayNode) neighbors.elementAt(i);
	   	   		AdhocMAC nmac= (AdhocMAC)nnode.wl_MAC;
				if(nmac.ms_timeLastSent>simTime()){
					if(ms_timeLastReceived>=0)node.wl_lastMsgCollision=true;
					if(ms_timeLastReceived < nmac.ms_timeLastSent)
						ms_timeLastReceived = nmac.ms_timeLastSent;
				}
			}
		}
		
		
	    //set DIFS timer
	    long trans_time;
	    if(ms_timeLastReceived>=0)trans_time=ms_timeLastReceived-simTime();
	    else trans_time=(long)(Math.random()*DIFS);
	    */
	    //insertDIFSEvent(trans_time + ran.nextInt(1000000));
	    insertDIFSEvent(ran.nextInt(1000000));
	    
	}
	/*
	 * @see MACInterface#MACStep()
	 */
	public void MACStep(){
		
		
	}
	/*
	 * @see MACInterface#handleMACTimer(SimEventMAC)
	 */
	public void handleMACTimer(SimEventMAC mac_event){
		
		int timer_index= mac_event.getIndex() ;
		
		switch(timer_index){
			case DIFS_TIMER:{
				if(mac_event==nextDIFS){
					//lotusDebug.errorln("DIFS event time="+simTime());
					handleDIFS();
				}
				break;
			}
			
			case BACKOFF_TIMER:{
				if(mac_event==nextBackoff){
					//lotusDebug.errorln("Backoff event time="+simTime());			
					handleBackoff();
					
				}
				break;
			}
			case SENDOUT_TIMER:{
				handleSendout();
				break;	
			}
			
			default:
				lotusDebug.errorln("Unknown MAC timer: "+timer_index);
		}
	}

}
