/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/

package lotos;

import lotos.*; 
import hypercast.HyperCastConfig;
import hypercast.I_Node;
import hypercast.I_PhysicalAddress;
import hypercast.I_UnicastAdapter;
import hypercast.DT.GNP_Landmark;
import hypercast.util.XmlUtil;

/**
 * This class tries to assign a DT address to a protocol node according to 
 * the configuration file 
 * @author HyperCast Team
 * @author Guangyu Dong
 *
 */
public class DTAddressManager {
	private simHCNetwork network;
	String prefix;
	
	GNP_Landmark landmarks[];
	I_PhysicalAddress landmarkAddrs[];
	
	public DTAddressManager(simHCNetwork network, String prefix){
		this.network = network;
		this.prefix = prefix;
	}
	public I_Node createServer(HyperCastConfig config, I_UnicastAdapter adapter){
		String coordMode = config.getTextAttribute(XmlUtil.createXPath(prefix +"Coords"));
		
		if(!coordMode.equals("USE_LM"))
			return null;
		
		if(landmarks==null){
			int numOfLandmark = Integer.parseInt(config.getTextAttribute(XmlUtil.createXPath(prefix +"Coords/USE_LM/LandmarkNum")));
			landmarks = new GNP_Landmark[numOfLandmark];
			landmarkAddrs = new I_PhysicalAddress[numOfLandmark];
		}
		
		GNP_Landmark newLandmark = null;
		int index = -1;
		
		for(int i=0;i<landmarks.length;i++){
			if(landmarks[i]==null){
				index = i;
				break;
			}
		}
		
		if(index>=0){
		
			newLandmark = new GNP_Landmark(config, adapter, prefix);
			landmarks[index] = newLandmark;
			landmarkAddrs[index] = adapter.createPhysicalAddress();
		}
		
		return newLandmark;
	}
	public void processConfig(XYAddress coords, HyperCastConfig config){
		String coordMode = config.getTextAttribute(XmlUtil.createXPath(prefix +"Coords"));
		
		if(coordMode.equals("FIXED")){
			config.setTextAttribute(XmlUtil.createXPath(prefix+"Coords/FIXED/coordinate"), ""+coords.getX()+","+coords.getY());
		
		}else{
			//TODO
		}
	}
	public boolean needServers(HyperCastConfig config){
		
		String coordMode = config.getTextAttribute(XmlUtil.createXPath(prefix +"Coords"));
		
		if(coordMode.equals("USE_LM")){
			if(landmarks==null){
				int numOfLandmark = Integer.parseInt(config.getTextAttribute(XmlUtil.createXPath(prefix +"Coords/USE_LM/LandmarkNum")));
				landmarks = new GNP_Landmark[numOfLandmark];
				landmarkAddrs = new I_PhysicalAddress[numOfLandmark];
			}
			for(int i=0;i<landmarks.length;i++){
				if(landmarks[i] == null)
					return true;
			}
		}
		return false;
	}
	public void nodeRemoved(I_Node node){
		for(int i=0;i<landmarks.length;i++){
			if(node == landmarks[i])
				landmarks[i] =null;
		}
	}
	
	
}
