/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team
  */
 

package lotos;

import lotos.*; 
import hypercast.HyperCastConfig;
import hypercast.I_Node;
import hypercast.I_PhysicalAddress;
import hypercast.I_UnderlayAddress;
import hypercast.I_UnicastAdapter;
import hypercast.DT.DT_LogicalAddress;
import hypercast.DT.DT_Node_BuddyList;
import hypercast.DT.GNP_Landmark;
import hypercast.util.XmlUtil;

import java.awt.Color;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.Vector;


public class DTBuddyListAgent implements I_SimHyperCastAgent {
	
	simHCNetwork network;
	Random r = new Random();
	DTAddressManager dtAddrMngr ;
	String prefix = "/Public/Node/DTBuddyList/";
	
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getName()
	 */
	public String getName() {
		return "DTBuddyList";
	}
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#setNetwork(simHCNetwork)
	 */
	public void setNetwork(simHCNetwork network) {
		this.network=network;
		dtAddrMngr = new DTAddressManager(network, prefix);
	}
	
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#createNode(XYAddress, HC_Adapter)
	 */
	public I_Node createNode(XYAddress coords, HyperCastConfig config, I_UnicastAdapter adapter, String info) {
		I_Node newNode;
		if(info==null){
			if(this.dtAddrMngr.needServers(config)){
				newNode =  this.dtAddrMngr.createServer(config, adapter);
			}else{
				dtAddrMngr.processConfig(coords, config);
				
				newNode = new DT_Node_BuddyList(config, adapter, prefix);
			}
		}else{
			StringTokenizer tokens = new StringTokenizer(info, ":");
			String type = tokens.nextToken();
			String laStr = tokens.nextToken();
			DT_LogicalAddress la = new DT_LogicalAddress(laStr);
			newNode = new DT_Node_BuddyList(config, adapter, la, prefix);
		}
		
		return newNode;
	}
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getInfoForNode(hypercast.I_Node)
	 */
	public String getInfoForNode(I_Node node) {
		return "dtnode:"+node.getAddressPair().getLogicalAddress();
	}
	
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getMenuItems()
	 */
	public String[] getMenuItems() {
		//TODO
		return new String[0];
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#menuFired(java.lang.String)
	 */
	public void menuFired(String menu) {
		// TODO Auto-generated method stub
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getColorForNode(hypercast.I_Node)
	 */
	public Color getColorForNode(I_Node node) {
		if(node instanceof DT_Node_BuddyList){
			DT_Node_BuddyList dtbuddynode = (DT_Node_BuddyList)node;
			if(dtbuddynode.getState()==DT_Node_BuddyList.LEADER_INCOMPLETE || dtbuddynode.getState()==DT_Node_BuddyList.LEADER_STABLE)
				return Color.ORANGE;
			return Color.YELLOW;
		}else if(node instanceof GNP_Landmark){
			return Color.MAGENTA;
		}
		return Color.YELLOW;
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getShapeForNode(hypercast.I_Node)
	 */
	public int getShapeForNode(I_Node node) {
		if(node instanceof DT_Node_BuddyList)
			return protocolInterface.shapeCircle;
		else
			return protocolInterface.shapeTriangle;
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#nodeRemoved(hypercast.I_Node)
	 */
	public void nodeRemoved(I_Node node) {
		//do nothing
		
	}

	

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#needServers()
	 */
	public boolean needServers(HyperCastConfig config) {
		return this.dtAddrMngr.needServers(config);
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getLabelForNode(hypercast.I_Node)
	 */
	public String getLabelForNode(I_Node node) {
		return "dtbuddy";
	}

	
}
