/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team
  * 2005 
  */
 
package lotos;

import lotos.*; 
import hypercast.HyperCastConfig;
import hypercast.HyperCastConfigException;
import hypercast.I_Node;
import hypercast.I_PhysicalAddress;
import hypercast.I_UnderlayAddress;
import hypercast.I_UnicastAdapter;
import hypercast.DT.DT_LogicalAddress;
import hypercast.DT.DT_Node_BuddyList;
import hypercast.DT.DT_Node_ServerJoin;
import hypercast.DT.DT_Server;
import hypercast.DT.GNP_Landmark;
import hypercast.util.XmlUtil;

import java.awt.Color;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.Vector;


public class DTServerAgent implements I_SimHyperCastAgent {
	
	simHCNetwork network;
	Random r = new Random();
	DTAddressManager dtAddrMngr ;
	String prefix = "/Public/Node/DTServer/";
	
	private DT_Server dtServer = null;
	
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getName()
	 */
	public String getName() {
		return "DTServer";
	}
	
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getLabelForNode(hypercast.I_Node)
	 */
	public String getLabelForNode(I_Node node) {
		if(node instanceof DT_Server)
			return "dtserver";
		else
			return "dtnode";
	}
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#setNetwork(simHCNetwork)
	 */
	public void setNetwork(simHCNetwork network) {
		this.network=network;
		dtAddrMngr = new DTAddressManager(network, prefix);
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#createNode(XYAddress, HC_Adapter)
	 */
	public I_Node createNode(XYAddress coords, HyperCastConfig config, I_UnicastAdapter adapter, String info) {
		if(info==null){
			if(this.isServer(config, adapter)){
				dtServer = new DT_Server(config, adapter);
				return dtServer;
			}else if(this.dtAddrMngr.needServers(config)){
				return this.dtAddrMngr.createServer(config, adapter);
			}else{
				
				dtAddrMngr.processConfig(coords, config);
				
				I_Node newNode = new DT_Node_ServerJoin(config, adapter, prefix);
				
				return newNode;
			}
		}else{
			if(info.startsWith("dtserver")){
				//create a dt server
				dtServer = new DT_Server(config, adapter);
				return dtServer;
			}else{
				//create a dt node
				StringTokenizer tokens = new StringTokenizer(info, ":");
				String type = tokens.nextToken();
				String laStr = tokens.nextToken();
				DT_LogicalAddress la = new DT_LogicalAddress(laStr);
				I_Node newNode = new DT_Node_ServerJoin(config, adapter, la, prefix);
				return newNode;
			}
		}
	}

    /** Read the IP addresses of DT servers from an configuration object */
    private boolean isServer(HyperCastConfig config, I_UnicastAdapter adapter)
    {
    	I_UnderlayAddress nodePhysicalAddress = adapter.createUnderlayAddress();
		/*
		 * XPath index starts from 1 not 0.
		 */
        for (int i = 1; true; i++)
        {
        	String addrType;
        	try
			{
        		addrType = config.getTextAttribute(XmlUtil.createXPath(prefix + DT_Node_ServerJoin.SERVER_PREFIX + "[" + i + "]/UnderlayAddress"));
			}catch (HyperCastConfigException e)
			{
				/*
				 * The attribute doesn't exist.
				 */
				break;
			}
        	if(addrType == null)
        		break;
        	
			String propName = prefix + DT_Node_ServerJoin.SERVER_PREFIX + "[" + i + "]/UnderlayAddress/" + addrType;

			String s = null;
			try
			{            
				s = config.getTextAttribute(XmlUtil.createXPath(propName));
			}
			catch (HyperCastConfigException e)
			{
				/*
				 * The attribute doesn't exist.
				 */
				break;
			}
            
            I_UnderlayAddress serverAddr = adapter.createUnderlayAddress(s);

            if (serverAddr != null)
            {
            	if(nodePhysicalAddress.equals(serverAddr))
            		return true;
            } 
            else
            {
                config.err.println("Property " + prefix + DT_Node_ServerJoin.SERVER_PREFIX + i + " contained a bad address. ");         
            }
        }
        return false;
    }
	
	/**
	 * create a DTServer node
	 * @return
	 */
	private I_Node createDTServer(HyperCastConfig config, I_UnicastAdapter adapter){
		I_Node dtServer = new DT_Server(config, adapter);
		return dtServer;
	}
	
	
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getMenuItems()
	 */
	public String[] getMenuItems() {
		//TODO
		return new String[0];
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#menuFired(java.lang.String)
	 */
	public void menuFired(String menu) {
		// TODO Auto-generated method stub
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getColorForNode(hypercast.I_Node)
	 */
	public Color getColorForNode(I_Node node) {
		if(node instanceof DT_Server)
			return Color.GREEN;
		if(node instanceof DT_Node_ServerJoin){
			DT_Node_ServerJoin dtnode = (DT_Node_ServerJoin)node;
			if(dtnode.getState()==DT_Node_ServerJoin.LEADER_INCOMPLETE || dtnode.getState()==DT_Node_ServerJoin.LEADER_STABLE)
				return Color.ORANGE;
			return Color.YELLOW;
		}else if(node instanceof GNP_Landmark){
			return Color.MAGENTA;
		}
		return Color.YELLOW;
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getShapeForNode(hypercast.I_Node)
	 */
	public int getShapeForNode(I_Node node) {
		if(node instanceof DT_Server)
			return protocolInterface.shapeSquare;
		else if(node instanceof DT_Node_ServerJoin)
			return protocolInterface.shapeCircle;
		else
			return protocolInterface.shapeTriangle;
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#nodeRemoved(hypercast.I_Node)
	 */
	public void nodeRemoved(I_Node node) {
		if(node == dtServer)
			dtServer =null;
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#needServers()
	 */
	public boolean needServers(HyperCastConfig config) {
		if(dtServer==null)
			return true;
		else
			return this.dtAddrMngr.needServers(config);
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getInfoForNode(hypercast.I_Node)
	 */
	public String getInfoForNode(I_Node node) {
		if((node instanceof DT_Server)){
			return "dtserver";
		}else{
			return "dtnode:"+node.getAddressPair().getLogicalAddress();
		}
	}
	

	
}
