/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team
  * July 2002 
  */
 

package lotos;

import lotos.*; 
import java.lang.Math;
import java.io.*;
import java.util.*;

/**
 * Exception report class
 */

class Panic {
    static public void panic(String m) {
        System.err.println(m);
        System.exit(1);
    }
}

/*
 * Wrapper class for basic int type for pass by reference.
 */

class DTInt {
    int i;

    public void DTInt() {
	i = 0;
    }

    public void DTInt(int i) {
	this.i = i;
    }

    public void setValue(int i) {
	this.i = i;
    }

    public int getValue() {
	return i;
    }
}

/*
 * Point class.  RealPoint to avoid clash with java.awt.Point.
 */
class RealPoint {
    float x, y;

    RealPoint() { x = y = 0.0f; }
    RealPoint(float x, float y) { this.x = x; this.y = y; }
    RealPoint(RealPoint p) { x = p.x; y = p.y; }
    public float x() { return this.x; }
    public float y() { return this.y; }
    public void set(float x, float y) { this.x = x; this.y = y; }

    public float distance(RealPoint p) {
	float dx, dy;

	dx = p.x - x;
	dy = p.y - y;
	return (float)Math.sqrt((double)(dx * dx + dy * dy));
    }

    public float distanceSq(RealPoint p) {
	float dx, dy;

	dx = p.x - x;
	dy = p.y - y;
	return (float)(dx * dx + dy * dy);
    }
}

/*
 * Edge class. Edges have two vertices, s and t, and two faces,
 * l (left) and r (right). The triangulation representation and
 * the Delaunay triangulation algorithms require edges.
 */

 class Edge {
    int s, t;
    int l, r;

    Edge() { s = t = 0; }
    Edge(int s, int t) { this.s =s; this.t = t; }
    int s() { return this.s; }
    int t() { return this.t; }
    int l() { return this.l; }
    int r() { return this.r; }
}


/*
 * Vector class.  A few elementary vector operations.
 */
class DTVector {
    float u, v;

    DTVector() { u = v = 0.0f; }
    DTVector(RealPoint p1, RealPoint p2) {
	u = p2.x() - p1.x();
	v = p2.y() - p1.y();
    }
    DTVector(float u, float v) { this.u = u; this.v = v; }

    float dotProduct(DTVector v) { return u * v.u + this.v * v.v; }

    static float dotProduct(RealPoint p1, RealPoint p2, RealPoint p3) {
	float u1, v1, u2, v2;

	u1 =  p2.x() - p1.x();
	v1 =  p2.y() - p1.y();
	u2 =  p3.x() - p1.x();
	v2 =  p3.y() - p1.y();

	return u1 * u2 + v1 * v2;
    }

    float crossProduct(DTVector v) { return u * v.v - this.v * v.u; }

    static float crossProduct(RealPoint p1, RealPoint p2, RealPoint p3) {
	float u1, v1, u2, v2;

	u1 =  p2.x() - p1.x();
	v1 =  p2.y() - p1.y();
	u2 =  p3.x() - p1.x();
	v2 =  p3.y() - p1.y();

	return u1 * v2 - v1 * u2;
    }

    void setRealPoints(RealPoint p1, RealPoint p2) {
	u = p2.x() - p1.x();
	v = p2.y() - p1.y();
    }
}

/*
 * Circle class. Circles are fundamental to computation of Delaunay
 * triangulations.  In particular, an operation which computes a
 * circle defined by three points is required.
 */
 class Circle {
    RealPoint c;
    float r;

    Circle() { c = new RealPoint(); r = 0.0f; }
    Circle(RealPoint c, float r) { this.c = c; this.r = r; }
    public RealPoint center() { return c; }
    public float radius() { return r; }
    public void set(RealPoint c, float r) { this.c = c; this.r = r; }

    /*
     * Tests if a point lies inside the circle instance.
     */
    public boolean inside(RealPoint p) {
	if (c.distanceSq(p) < r * r)
	    return true;
	else
	    return false;
    }

    /*
     * Compute the circle defined by three points (circumcircle).
     */
    public void circumCircle(RealPoint p1, RealPoint p2, RealPoint p3) {
	float cp;

	cp = DTVector.crossProduct(p1, p2, p3);
	if (cp != 0.0)
	    {
		float p1Sq, p2Sq, p3Sq;
		float num, den;
		float cx, cy;

		p1Sq = p1.x() * p1.x() + p1.y() * p1.y();
		p2Sq = p2.x() * p2.x() + p2.y() * p2.y();
		p3Sq = p3.x() * p3.x() + p3.y() * p3.y();
		num = p1Sq*(p2.y() - p3.y()) + p2Sq*(p3.y() - p1.y()) + p3Sq*(p1.y() - p2.y());
		cx = num / (2.0f * cp);
		num = p1Sq*(p3.x() - p2.x()) + p2Sq*(p1.x() - p3.x()) + p3Sq*(p2.x() - p1.x());
		cy = num / (2.0f * cp);

		c.set(cx, cy);
	    }

	// Radius
	r = c.distance(p1);
    }
}

/*
 * Triangulation class.  A triangulation is represented as a set of
 * points and the edges which form the triangulation.
 */
class Triangulation {
    static final int Undefined = -1;
    static final int Universe = 0;
    int nPoints;
    RealPoint point[];
    int nEdges;
    int maxEdges;
    Edge edge[];

    Triangulation( DTAddress[] nodes ) {

	// Allocate points.
	this.nPoints = nodes.length;
	this.point = new RealPoint[nPoints];
	for (int i = 0; i < nPoints; i++)
	    point[i] = new RealPoint();

	// Allocate edges.
	maxEdges = 3 * nPoints - 6;	// Max number of edges.
	edge = new Edge[maxEdges];
	for (int i = 0; i < maxEdges; i++)
	    edge[i] = new Edge();
	nEdges = 0;

	for (int i = 0; i < nPoints; i++) {
	    point[i].x = nodes[i].x;
	    point[i].y = nodes[i].y;
	}


    }


    void addTriangle(int s, int t, int u) {
	addEdge(s, t);
	addEdge(t, u);
	addEdge(u, s);
    }

    public int addEdge(int s, int t) {
	return addEdge(s, t, Undefined, Undefined);
    }

    /*
     * Adds an edge to the triangulation. Store edges with lowest
     * vertex first (easier to debug and makes no other difference).
     */
    public int addEdge(int s, int t, int l, int r) {
	int e;

	// Add edge if not already in the triangulation.
	e = findEdge(s, t);
	if (e == Undefined)
	    if (s < t)
		{
		    edge[nEdges].s = s;
		    edge[nEdges].t = t;
		    edge[nEdges].l = l;
		    edge[nEdges].r = r;
		    return nEdges++;
		}
	    else
		{
		    edge[nEdges].s = t;
		    edge[nEdges].t = s;
		    edge[nEdges].l = r;
		    edge[nEdges].r = l;
		    return nEdges++;
		}
	else
	    return Undefined;
    }

    public int findEdge(int s, int t) {
	boolean edgeExists = false;
	int i;

	for (i = 0; i < nEdges; i++)
	    if (edge[i].s == s && edge[i].t == t ||
		edge[i].s == t && edge[i].t == s) {
		edgeExists = true;
		break;
	    }

	if (edgeExists)
	    return i;
	else
	    return Undefined;
    }

    /*
     * Update the left face of an edge.
     */
    public void updateLeftFace(int eI, int s, int t, int f) {
	if (!((edge[eI].s == s && edge[eI].t == t) ||
	      (edge[eI].s == t && edge[eI].t == s)))
	    Panic.panic("updateLeftFace: adj. matrix and edge table mismatch");
	if (edge[eI].s == s && edge[eI].l == Triangulation.Undefined)
	    edge[eI].l = f;
	else if (edge[eI].t == s && edge[eI].r == Triangulation.Undefined)
	    edge[eI].r = f;
	else
	    Panic.panic("updateLeftFace: attempt to overwrite edge info");
    }

    public void print(PrintStream p) {
	printPoints(p);
	printEdges(p);
    }

    public void printPoints(PrintStream p) {
	for (int i = 0; i < nPoints; i++)
	    p.println(String.valueOf(point[i].x) + " " + String.valueOf(point[i].y));
    }

    public void printEdges(PrintStream p) {
	for (int i = 0; i < nEdges; i++)
	    p.println(String.valueOf(edge[i].s) + " " + String.valueOf(edge[i].t));
    }
}

/*
 * QuadraticAlgorithm class.  O(n^2) algorithm.
 */
class TriangulationAlgorithm {
    int s, t, u, bP;
    Circle bC = new Circle();
      /*final static String algName = "O(n^2)";
    int nFaces;*/

    public void triangulate(Triangulation tri) {
	int seedEdge, currentEdge;
	int nFaces;
	DTInt s, t;

	// Initialise.
	nFaces = 0;
	s = new DTInt();
	t = new DTInt();

	// Find closest neighbours and add edge to triangulation.
	findClosestNeighbours(tri.point, tri.nPoints, s, t);

	// Create seed edge and add it to the triangulation.
	seedEdge = tri.addEdge(s.getValue(), t.getValue(),
			       Triangulation.Undefined,
			       Triangulation.Undefined);

	currentEdge = 0;
	while (currentEdge < tri.nEdges) {
	    if (tri.edge[currentEdge].l == Triangulation.Undefined) {
		completeFacet(currentEdge, tri, nFaces);
	    }
	    if (tri.edge[currentEdge].r == Triangulation.Undefined) {
		completeFacet(currentEdge, tri, nFaces);
	    }
	    currentEdge++;
	}
    }

    // Find the two closest points.
    public void findClosestNeighbours(RealPoint p[], int nPoints,
				      DTInt u, DTInt v) {
	int i, j;
	float d, min;
	int s, t;

	s = t = 0;
	min = Float.MAX_VALUE;
	for (i = 0; i < nPoints-1; i++)
	    for (j = i+1; j < nPoints; j++)
		{
		    d = p[i].distanceSq(p[j]);
		    if (d < min)
			{
			    s = i;
			    t = j;
			    min = d;
			}
		}
	u.setValue(s);
	v.setValue(t);
    }

    /*
     * Complete a facet by looking for the circle free point to the left
     * of the edge "e_i".  Add the facet to the triangulation.
     *
     * This function is a bit long and may be better split.
     */
    public void completeFacet(int eI, Triangulation tri, int nFaces) {
	float cP;
	int i;
	Edge e[] = tri.edge;
	RealPoint p[] = tri.point;

	// Cache s and t.
	if (e[eI].l == Triangulation.Undefined)
	    {
		s = e[eI].s;
		t = e[eI].t;
	    }
	else if (e[eI].r == Triangulation.Undefined)
	    {
		s = e[eI].t;
		t = e[eI].s;
	    }
	else
	    // Edge already completed.
	    return;


	// Find a point on left of edge.
	for (u = 0; u < tri.nPoints; u++)
	    {
		if (u == s || u == t)
		    continue;
		if (DTVector.crossProduct(p[s], p[t], p[u]) > 0.0)
		    break;
	    }

	// Find best point on left of edge.
	bP = u;
	if (bP < tri.nPoints)
	    {
		bC.circumCircle(p[s], p[t], p[bP]);

		for (u = bP+1; u < tri.nPoints; u++)
		    {
			if (u == s || u == t)
			    continue;

			cP = DTVector.crossProduct(p[s], p[t], p[u]);

			if (cP > 0.0)
			    if (bC.inside(p[u]))
				{
				    bP = u;
				    bC.circumCircle(p[s], p[t], p[u]);
				}
		    }
	    }

	// Add new triangle or update edge info if s-t is on hull.
	if (bP < tri.nPoints)
	    {
		// Update face information of edge being completed.
		tri.updateLeftFace(eI, s, t, nFaces);
		nFaces++;

		// Add new edge or update face info of old edge.
		eI = tri.findEdge(bP, s);
		if (eI == Triangulation.Undefined)
		    // New edge.
		    eI = tri.addEdge(bP, s, nFaces, Triangulation.Undefined);
		else
		    // Old edge.
		    tri.updateLeftFace(eI, bP, s, nFaces);

		// Add new edge or update face info of old edge.
		eI = tri.findEdge(t, bP);
		if (eI == Triangulation.Undefined)
		    // New edge.
		    eI = tri.addEdge(t, bP, nFaces, Triangulation.Undefined);
		else
		    // Old edge.
		    tri.updateLeftFace(eI, t, bP, nFaces);
	    } else
		tri.updateLeftFace(eI, s, t, Triangulation.Universe);
    }
}

/**
 * This class encapsulates the algorithm of Delaunay Triangulation by hiding
 * the details of the algorithm and providing a simple interface for the outside
 * classes to invoke the algorithm
 */

class DelaunayTriangulation{

    /**
     * Calculate Delaunay Triangulation of the nodes
     * @param nodes points on the logical plan with two dimensional coordinates
     * @return      the Delaunay Triangulation of the points
     */
    static DTEdge[] QuadraDT( DTAddress [] nodes ){

    lotusDebug.println("before triangulate");
	Triangulation rtri = new Triangulation( nodes );
    lotusDebug.println("step 1");
	TriangulationAlgorithm ralg = new TriangulationAlgorithm( );
    lotusDebug.println("Step 2");
	ralg.triangulate(rtri);
    lotusDebug.println("Step 3");
	DTEdge[] edges = new DTEdge[rtri.nEdges];

    lotusDebug.println("prepare to return");
	for( int i=0; i<edges.length; i++){
	    edges[i] = new DTEdge( rtri.edge[i].s, rtri.edge[i].t );
	}

	return edges;
    }

}
