/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team
  * July 2002 
  */
 
package lotos;

import lotos.*; 
import java.applet.*;
import java.awt.*;
import java.awt.event.*;
import java.awt.image.*;
import java.net.*;
import java.util.*;
import java.io.*;

/**
 * the display window of the GUI.
 */

class GUI_DisplayWindow
    extends Panel
    implements MouseMotionListener, MouseListener, ActionListener {
    // This is a constant used to scale distances in the network.
    // A packet in the network will travel at a distance of PIXELS_PER_SECOND
    // pixels in one simulated second.

    //Edit Mode
    final static int EDIT_ARROW = 0;
    final static int EDIT_NODE = 1;
    final static int EDIT_DENODE = 2;
    
    final static int EDIT_LINE = 3;
    final static int EDIT_DELINE = 4;
    
    final static int EDIT_START = 5;
    final static int EDIT_TREE = 6;
    final static int EDIT_SRC = 7;
    final static int EDIT_DST = 8;
    
    //added 04-08-04
    final static int EDIT_MOVE = 9;
    
    int node_type=0;
    
    final static int NODE_STATIC_WIRED=0;
    final static int NODE_STATIC_WIRELESS=1;
    final static int NODE_STATIC_BOTH=2;
    final static int NODE_MOBILE_WIRELESS=3;
    
    boolean appendSelectedNode;
    
    String mobilityMode;
	MouseAdapter current_listener;

      /** the current edit mode of the display window */
    int edit_mode;
    /** the display scale */
    float scale; // node address * scale ==> point on the screen.

    /** view point, used to move the part of display shown */
    int viewX, viewY;

    /** Popup Menu as interface for different protocols */
    PopupMenu pm_edit;

    MenuItem[] mi_protocol;

    //static final float PIXELS_PER_SECOND = (float) 100000;  //  2/3speed of light * 1pix/2km

    /** a reference to class GUI */
    GUI the_gui;

    // These variables determine what will be drawn in the window.
    boolean display_on;
    boolean display_logical;
    boolean display_physical;
    boolean display_node_labels;
    boolean display_messages;
    boolean display_message_labels;
    boolean display_multicast;
    boolean display_wirelessrange;

    static Font nodeFont = new Font("Helvetica", Font.PLAIN, 10);
	Font   time_font = new Font("Helvetica", Font.BOLD, 14);

    /** record the root of the multicast tree */
    OverlayNode multicastRoot;
    OverlayNode routeSrc;
    OverlayNode routeDst;

    /**
     * Constructor. It initializes display reated setting
     */
    public GUI_DisplayWindow(GUI gui_arg) {

    //set up node modes
    edit_mode = EDIT_ARROW;
    node_type = NODE_STATIC_WIRELESS;
    
    scale = 1.0f;
	mobilityMode = "Linear Motion";

    //pm_edit = new PopupMenu("Edit");

    //this.add(pm_edit);

    the_gui = gui_arg;
	
	display_on = true;
    display_logical = true;
    display_physical = true;
    display_node_labels = true;
    display_messages = true;
    display_message_labels = true;
    display_multicast = true;
    display_wirelessrange = true;

    setBackground(GUI.dwColor);
    addMouseListener(this);
    addMouseMotionListener(this);
    
    }


    private Image offScreenBuffer;
    private Graphics offGraphics;

    /**
     * Update the display in the window.
     * Use double buffers to avoid the flickering when updating
     */
    public void update(Graphics g)
    {
        if (offScreenBuffer==null ||
            (! (offScreenBuffer.getWidth(this) == this.getSize().width &&
                offScreenBuffer.getHeight(this) == this.getSize().height)))
        {
            offScreenBuffer = this.createImage(getSize().width, getSize().height);
        }

        // We need to use our buffer Image as a Graphics object:
        offGraphics = offScreenBuffer.getGraphics();
        paint(offGraphics);
        // Passes our off-screen buffer to our paint method, which,
        // unsuspecting, paints on it just as it would on the Graphics
        // passed by the browser or applet viewer.
        g.drawImage(offScreenBuffer, 0, 0, this);
        // And now we transfer the info in the buffer onto the
        // graphics context we got from the browser in one smooth motion.

    }

    /**
     * Paint the window and invoke the paint subroutine based on
     * the setting
     */
    public void paint(Graphics g)
    {
    	Dimension dim = this.getSize();
        g.setColor( GUI.dwColor );
        g.fillRect(0, 0, dim.width, dim.height);
        
        g.setColor(Color.black);
        g.drawLine(0, 0, 0, dim.height-1);
        g.drawLine(0, 0, dim.width-1, 0);
        g.drawLine(dim.width-1, 0, dim.width-1, dim.height-1);
        g.drawLine(0, dim.height-1, dim.width-1, dim.height-1);
		
		if(!display_on){
    		//do nothing but draw the time
    		String[] sarray = new String[4];
			OverlayNetwork psim = the_gui.getOverlay();
        	
        	sarray[0] = "Time = " + Float.toString(psim.simTime());
        	sarray[1] = ""+psim.nodes.size()+" nodes";
        	sarray[2] = ""+psim.getNumOfLeaf()+" leaves";
        	sarray[3] = "Display is OFF";
    		g.setFont(time_font);
    		FontMetrics time_font_met = g.getFontMetrics();
        	drawStringArray(sarray, 4, 4, g, time_font_met);
        	
    		return;	
    	}
		
        if( display_physical ) {
        	paint_physical_edges(g);
        	//paint_physical_transmission(g);
        }
        if( display_logical ) paint_logical_edges(g);
        paint_nodes(g);
        if( display_node_labels ) paint_nodes_name(g);

        if( display_messages ) paint_messages(g);
        if( display_message_labels ) paint_message_labels(g);


        g.setColor( Color.black );
        
		// Draw current time
        // Draw the time it went stable
        String[] sarray = new String[2];
		OverlayNetwork psim = the_gui.getOverlay();
        //sarray[0] = psim.getProtocols().;
        sarray[0] = "Time = " + Float.toString(psim.simTime());
        sarray[1] = ""+psim.nodes.size()+" nodes";
        //sarray[2] = ""+psim.getNumOfLeaf()+" leaves";
        //sarray[2] = "Collision rate = " + ((float)psim.sum_collision *100.0f / (float)psim.sum_sent) +"%";

        //float time_stable = topo.stableSinceTime();
        //if (time_stable > 0.0)
        //    sarray[2] = "Stable at Time = " + Float.toString(time_stable);
        //else
        //    sarray[2] = "Not Stable";
        
        g.setFont(time_font);
        FontMetrics time_font_metrics = g.getFontMetrics();
        drawStringArray(sarray, 4, 4, g, time_font_metrics);
        
        paint_legends(g);
        
        drawMulticastTree( g );
        drawRoute( g);
        
        if(win_select){
        	//draw an rectangle
        	g.setColor(Color.RED);
        	int startX, startY, endX, endY;
        	if(arrowStart.x<arrowEnd.x){
        		startX = arrowStart.x;
        		endX = arrowEnd.x;
        	}
        	else{
        		startX = arrowEnd.x;
        		endX = arrowStart.x;
        	}
        	
        	if(arrowStart.y<arrowEnd.y){
        		startY = arrowStart.y;
        		endY = arrowEnd.y;
        	}
        	else{
        		startY = arrowEnd.y;
        		endY = arrowStart.y;
        	}
        	
        	g.drawRect(startX, startY, endX-startX,  endY-startY);
        	
        }
        
    }
    
    private void paint_legends(Graphics g){
    	Color legendColor = Color.GRAY;
    	int xStart = 45;
    	
    	String sarray[] = new String[2];
    	sarray[0] = "Static Wired";
    	sarray[1] = "Static Wireless";
    	
    	g.setFont(time_font);
        FontMetrics time_font_metrics = g.getFontMetrics();
        
        OverlayNode onode = new OverlayNode(new XYAddress(xStart+97, 14), the_gui.getOverlay(), "test");
        onode.wired_link = true;
        onode.wireless_link = false;
        drawPhyNode(g, onode);
        
        onode.setAddress(new XYAddress(xStart+97, 33));
        onode.wired_link = false;
        onode.wireless_link = true;
        drawPhyNode(g, onode);
        
        g.setColor(legendColor);
        drawStringArray(sarray, xStart+105, 4, g, time_font_metrics);
        
        onode.setAddress(new XYAddress(xStart+227, 14));
        onode.wired_link = true;
        onode.wireless_link = true;
        drawPhyNode(g, onode);
        
        onode.setAddress(new XYAddress(xStart+227, 33));
        onode.wired_link = false;
        onode.wireless_link = true;
        onode.setMobility("Linear Motion");
        drawPhyNode(g, onode);
        
        sarray[0] = "Static Both";
    	sarray[1] = "Mobile Wireless";
    	
    	g.setColor(legendColor);
    	drawStringArray(sarray, xStart+235, 4, g, time_font_metrics);
    	
    	g.setColor(OverlayNetwork.phyEdgeColor);
    	g.drawLine(xStart+358, 14, xStart+380, 14);
    	
    	g.setColor(Color.yellow);
    	int width = 2;

    	for( int i=0; i<width; i++) g.drawLine(xStart+358, 33+i,xStart+380, 33+i);
    	
    	sarray[0] = "Physical Link";
    	sarray[1] = "Logical Link";
    	
    	g.setColor(legendColor);
    	drawStringArray(sarray, xStart+385, 4, g, time_font_metrics);
    	
    	
    	g.setColor(Color.blue);
    	g.drawLine(xStart+485, 14, xStart+485+10, 14);
    	g.drawLine(xStart+485, 10, xStart+485, 14);
    	g.drawLine(xStart+485+10, 10, xStart+485+10, 14);
    	g.drawLine(xStart+485, 33, xStart+485+((int)WirelessPhysicalLayer.SIGNAL_RANGE), 33);
    	for(int i=0;i<=WirelessPhysicalLayer.SIGNAL_RANGE;i+=10){
    		g.drawLine(xStart+485+i, 29, xStart+485+i, 33);
    	}
    	
    	sarray[0] = "10 meters";
    	sarray[1] = "Wireless Range ("+WirelessPhysicalLayer.SIGNAL_RANGE+"m)";
    	
    	g.setColor(legendColor);
    	drawStringArray(sarray, xStart+490+(int)WirelessPhysicalLayer.SIGNAL_RANGE, 4, g, time_font_metrics);
    	
    }
    
    class DelayedPopupShower extends Thread{
    	
    	public void run(){
    		try {
				Thread.sleep(200);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			if(showPopup){
	        	showPopup = false;
	        	pm_edit.show(GUI_DisplayWindow.this, popupPoint.x, popupPoint.y);
	        	repaint();
	        }
    	}
    }
	private void drawStringArray(String[] sarray, int x, int y, Graphics g, FontMetrics fm) 
    {
        y += fm.getAscent();
        for (int i = 0; i < sarray.length; i++) 
        {
            g.drawString(sarray[i], x, y);
            y += fm.getHeight();
        }
    }
    /**
     * Paint the messages currently in the network
     */
    private void paint_messages(Graphics g) {
        OverlayNetwork psim = the_gui.getOverlay();
        Enumeration enumTmp = psim.getPendingEvent();

        while( enumTmp.hasMoreElements() ){
        	
            SimEvent se = (SimEvent)enumTmp.nextElement();
            if( se.eventType() != SimEvent.MESSAGE_EVENT ) continue;
            SimEventMsg msg = (SimEventMsg) se;
            
            XYAddress src = msg.getSender().getAddress();
		    XYAddress dst;
			float fraction;
				
	        
	        if(msg.getRecipient()==null){
            	
            	//lotusDebug.errorln("drawing wireless message: "+msg);
            	
            	//fraction=( ( msg.getTime() - psim.simTime() ) * OverlayNode.PACKET_SPEED ) / msg.getSender().wireless_trans_radius;
            	
            	fraction=0;
            	
            	Point p = scaleToPoint( src );
	            g.setColor( Color.green );
				int r = (int)(msg.getSender().GetRadiusF()*(1-fraction)/scale);
	            g.drawOval( p.x-r, p.y-r,2*r,2*r );
	            
	        }
	        else {
	        	
	        	dst= msg.getRecipient().getAddress();
				
				fraction = ( ( msg.getTime() - psim.simTime() ) * OverlayNode.PACKET_SPEED ) / se.getSender().getDirectDistanceTo(se.getRecipient());
				
				if(fraction<0)
					fraction = 0;
				else if(fraction>1)
					fraction =1;
				
				//lotusDebug.errorln("fraction="+fraction);
					
				XYAddress middle = new XYAddress( (int) (src.getX() * fraction + dst.getX() * (1-fraction)),
		                                              (int) (src.getY() * fraction + dst.getY() * (1-fraction)) );
		        Point p = scaleToPoint( middle );
		        g.setColor( GUI.packetColor );
	        		
	        	//lotusDebug.errorln("drawing wired message: "+msg);
	        	
	            g.fillRect( p.x-GUI.packetSRadius/2, p.y-GUI.packetSRadius/2, GUI.packetSRadius, GUI.packetSRadius);
	            
	        }
	        
        }
        return;
    }

    /**
     * Paint message label
     */
    private void paint_message_labels(Graphics g){
        OverlayNetwork psim = the_gui.getOverlay();
        Enumeration enumTmp = psim.getPendingEvent();
        while( enumTmp.hasMoreElements() ){
            SimEvent se = (SimEvent)enumTmp.nextElement();
            if( se.eventType() != SimEvent.MESSAGE_EVENT ) continue;
            SimEventMsg msg = (SimEventMsg) se;
            XYAddress src = msg.getSender().getAddress();
             XYAddress dst;
			float fraction;
			
			if(msg.getRecipient()==null){
            	
            	//lotusDebug.errorln("drawing wireless message: "+msg);
            	
            	//fraction=( ( msg.getTime() - psim.simTime() ) * OverlayNode.PACKET_SPEED ) / msg.getSender().wireless_trans_radius;
            	
            	fraction=0;
            	
            	Point p = scaleToPoint( src );
	            g.setColor( Color.green );
				int r = (int)(msg.getSender().GetRadiusF()*0.8*(1-fraction)/scale);
	            
	            //g.drawString( msg.getMsgLabel(msg.getMsgType()), p.x-r, p.y-r);
				g.drawString( msg.getMsgLabel(msg.getMsgType()), p.x+ 15, p.y + 1 );
	        }
	        else{
	        	
	        	dst= msg.getRecipient().getAddress();
				
				fraction = ( ( msg.getTime() - psim.simTime() ) * OverlayNode.PACKET_SPEED ) / se.getSender().getDirectDistanceTo(se.getRecipient());
				
				if(fraction<0)
					fraction = 0;
				else if(fraction>1)
					fraction =1;
				//lotusDebug.errorln("fraction="+fraction);
					
				XYAddress middle = new XYAddress( (int) (src.getX() * fraction + dst.getX() * (1-fraction)),
		                                              (int) (src.getY() * fraction + dst.getY() * (1-fraction)) );
		        Point p = scaleToPoint( middle );
		        g.setColor( GUI.packetColor );
	        		
	        	//lotusDebug.errorln("drawing wired message: "+msg);
	        	g.drawString( msg.getMsgLabel(msg.getMsgType()), p.x+GUI.packetSRadius+1, p.y+GUI.packetSRadius+1);
	        	
	            
	        }
			
        }

        return;
    }

    private boolean outsidePanel( Point p ){
        Dimension dim = this.getSize();
        boolean b = p.x < 0 ||
                    p.y < 0 ||
                    p.x > dim.width ||
                    p.y > dim.height;

        return b;
    }

    /**
     * Paint the name of the nodes
     */
    private void paint_nodes_name(Graphics g){
        Dimension dim = this.getSize();
        for( int i=0; i<the_gui.getOverlay().getPhyNodes().size(); i++){
            OverlayNode onode = (OverlayNode ) the_gui.getOverlay().getPhyNodes().elementAt(i);
            //paint each node
            //if( null == onode.getName() ) continue;
			if( null == onode.getProtocol())continue;
			
            XYAddress addr = onode.getAddress();
            Point p = scaleToPoint( addr );
            if( outsidePanel( p ) ) continue;
            g.setFont(nodeFont);
            g.setColor(Color.black);
            g.drawString( onode.getProtocol().getName(), p.x, p.y+14);
        }
    }
    private void paint_wireless_range(Graphics g, OverlayNode onode){
    	g.setColor(Color.lightGray);
    	Point p = scaleToPoint( onode.getAddress() );
    	int range = (int)(WirelessPhysicalLayer.SIGNAL_RANGE*2);
    	g.drawOval(p.x - range/2,
                p.y - range/2,
                range,
				range);
    }
    /**
     * Paint nodes, including physical and logical part.
     * The display of the logical nodes depends on the configuration
     * of each protocol.
     */
    private void paint_nodes(Graphics g)
    {
    	float SIZE_RATE = (float)1.7;
    	Enumeration nodes = the_gui.getOverlay().getPhyNodes().elements();
        while (nodes.hasMoreElements())
        {
	        OverlayNode onode = (OverlayNode) nodes.nextElement();
	        Point p = scaleToPoint( onode.getAddress() );
            if( outsidePanel(p) ) continue;
            
            if(display_wirelessrange && onode.wireless_link)
            	paint_wireless_range(g, onode);
            
            if( onode.getProtocol() == null ) {
            //only physical node is painted
            	drawPhyNode(g, onode);
            }else
            {
                protocolInterface protocol = onode.getProtocol();
                
                g.setColor( protocol.getColor() );
                
                if(onode==multicastRoot)g.setColor(Color.cyan);
                if(onode==routeSrc)g.setColor(Color.cyan);
                if(onode==routeDst)g.setColor(Color.blue);
                
                if( protocol.getShape() == protocolInterface.shapeCircle )
                    g.fillOval(p.x - (int)(GUI.ndDiameter/2*1.5),
                        p.y - (int)(GUI.ndDiameter/2*SIZE_RATE),
                        (int)(GUI.ndDiameter*SIZE_RATE),
                        (int)(GUI.ndDiameter*SIZE_RATE));
                else if( protocol.getShape() == protocolInterface.shapeSquare )
                    g.fillRect(p.x - (int)(GUI.ndDiameter/2*SIZE_RATE),
                        p.y - (int)(GUI.ndDiameter/2*1.15),
                        (int)(GUI.ndDiameter*SIZE_RATE),
                        (int)(GUI.ndDiameter*SIZE_RATE));
                else if( protocol.getShape() == protocolInterface.shapeTriangle ){
                    int []xline = new int[3];
                    int []yline = new int[3];
                    xline[0] = p.x;
                    xline[1] = p.x - (int)(GUI.ndDiameter*0.86*SIZE_RATE);
                    xline[2] = p.x + (int)(GUI.ndDiameter*0.86*SIZE_RATE);
                    yline[0] = p.y - (int)(GUI.ndDiameter*SIZE_RATE);
                    yline[1] = p.y + (int)(GUI.ndDiameter/2*SIZE_RATE);
                    yline[2] = p.y + (int)(GUI.ndDiameter/2*SIZE_RATE);
                    g.fillPolygon( xline, yline, 3 );
                }
                
                drawPhyNode(g, onode);
            }
            if(the_gui.getOverlay().isNodeSelected(onode)){
            	g.setColor(Color.RED);
            	int ndDiameter = GUI.ndDiameter + 10;
            	g.drawOval(p.x - ndDiameter/2,
                   p.y - ndDiameter/2,
				   ndDiameter,
				   ndDiameter);
            }
        		
        }
    }
    
    private void drawPhyNode(Graphics g, OverlayNode onode){
    	Point p = scaleToPoint( onode.getAddress() );
    	if(onode.isMobile()){
        	g.setColor(Color.blue);
        }
        else g.setColor(OverlayNetwork.phyNodeColor);
        if(onode.isWired() && !onode.isWireless()){
            g.fillOval(p.x - GUI.ndDiameter/2,
               p.y - GUI.ndDiameter/2,
               GUI.ndDiameter,
               GUI.ndDiameter);
    	}else if(onode.isWired() && onode.isWireless()){
    		float rectWidth = (float)(GUI.ndDiameter*Math.sqrt(2)/2);
    		g.fillRect((int)(p.x - rectWidth /2), (int)(p.y - rectWidth/2), (int)rectWidth+1, (int)rectWidth+1);
    	}else {
    		int xPoints[] = new int [3];
    		int yPoints[] = new int [3];
    		float rectWidth = (float)(GUI.ndDiameter);
    		xPoints[0] = (int)(p.x - rectWidth /2);
    		yPoints[0] = (int)(p.y - rectWidth/2)+1;
    		xPoints[1] = (int)(p.x + rectWidth /2);
    		yPoints[1] = (int)(p.y - rectWidth/2)+1;
    		xPoints[2] = p.x;
    		yPoints[2] = p.y+ GUI.ndDiameter/2+1;
    		g.fillPolygon(xPoints, yPoints, 3);
    	}
    }

    /**
     * Paint the logical edges
     */
    private void paint_logical_edges(Graphics g)
    {
        Color edge_color;

        Enumeration nodes = the_gui.getOverlay().getPhyNodes().elements();

        while (nodes.hasMoreElements())
        {
	        OverlayNode fromnode = (OverlayNode) nodes.nextElement();

            Enumeration other_ends = fromnode.getLogicalNeighbor();
            Enumeration other_colors = fromnode.getLogicalColor();
            Enumeration other_widths = fromnode.getLogicalWidth();

            while(other_ends.hasMoreElements())
            {
                OverlayNode tonode = (OverlayNode) other_ends.nextElement();

		        Point  p1 = scaleToPoint( fromnode.getAddress() );
        		Point  p2 = scaleToPoint( tonode.getAddress() );
                if( outsidePanel(p1) && outsidePanel(p2) ) continue;
                Point  pm = new Point( (int) ((p1.x+p2.x)/2), (int) ((p1.y+p2.y)/2) );
		        g.setColor((Color)other_colors.nextElement());
                int xinterval = ( p1.x - pm.x > 0 ) ? (p1.x-pm.x) : (pm.x-p1.x);
                int yinterval = ( p1.y - pm.y > 0 ) ? (p1.y-pm.y) : (pm.y-p1.y);
                int width = ((Integer)other_widths.nextElement()).intValue();

                if( xinterval > yinterval ){
                    for( int i=0; i<width; i++) g.drawLine( p1.x, p1.y-width/2+i, pm.x, pm.y-width/2+i);
                }else{
                    for( int i=0; i<width; i++) g.drawLine( p1.x-width/2+i, p1.y, pm.x-width/2+i, pm.y);
                }
            }
        }
    }
	/**
     * Paint the physical edges
     */
    private void paint_physical_edges(Graphics g)
    {
        final Color edge_color = OverlayNetwork.phyEdgeColor;

        Enumeration nodes = the_gui.getOverlay().getPhyNodes().elements();

        while (nodes.hasMoreElements())
        {
	    	OverlayNode fromnode = (OverlayNode) nodes.nextElement();
	    	
	    	
            Enumeration other_ends = fromnode.physical_neighbors.elements();
            if (other_ends == null)
                continue;

            while(other_ends.hasMoreElements())
            {
                OverlayNode tonode = (OverlayNode) other_ends.nextElement();

                Point  p1 = scaleToPoint( fromnode.getAddress() );
                Point  p2 = scaleToPoint( tonode.getAddress() );
                if( outsidePanel(p1) && outsidePanel(p2) ) continue;
                g.setColor(edge_color);
                g.drawLine( p1.x, p1.y, p2.x, p2.y );

            }
        }
    }


    /**
     * Paint the physical edges
     */
    private void paint_physical_transmission(Graphics g)
    {
        OverlayNetwork psim = the_gui.getOverlay();
        Enumeration enumTmp = psim.getPendingEvent();

        while( enumTmp.hasMoreElements() ){
            SimEvent se = (SimEvent)enumTmp.nextElement();
            if( se.eventType() != SimEvent.MESSAGE_EVENT ) continue;
            SimEventMsg msg = (SimEventMsg) se;
            XYAddress src = msg.getSender().getAddress();
            Point p = scaleToPoint( src );
            g.setColor( Color.red );
			int r = (int)(msg.getSender().GetRadiusF()/scale);
            g.drawOval( p.x-r, p.y-r,2*r,2*r );
        }
        return;
    }

	
	public final void setDisplayOn(boolean b){
		display_on=b;
			
	}
    /**
     * set the display of logical edge
     */
    public final void setDisplayLogical(boolean b)
    {
        display_logical = b;
        repaint();
    }


    /**
     * set the display of physical edge
     */
    public final void setDisplayPhysical(boolean b)
    {
        display_physical = b;
        repaint();
    }

    /**
     * set the display of message
     */
    public final void setDisplayMessages(boolean b)
    {
        display_messages = b;
        repaint();
    }

    /**
     * set the display of message labels
     */
    public final void setDisplayMessageLabels(boolean b)
    {
        display_message_labels = b;
        repaint();
    }

    /**
     * set the display of node labels
     */
    public final void setDisplayNodeLabels(boolean b)
    {
        display_node_labels = b;
        repaint();
    }

    /**
     * set whether a multicast tree is displayed
     */
    public final void setDisplayMulticast(boolean b)
    {
        display_multicast = b;
        repaint();
    }
    public final void setDisplayWirelessRange(boolean b){
    	display_wirelessrange = b;
    	repaint();
    }
	private void drawRoute( Graphics g){
		OverlayNetwork net = the_gui.getOverlay();
		
		if(routeSrc==null || routeDst==null)return;
		if(!net.active_nodes.contains(routeSrc) || !net.active_nodes.contains(routeDst))return;
		
		Color edge_color = Color.green;
		
		OverlayNode prevHop = null;
		OverlayNode thisHop=routeSrc;
		OverlayNode nextHop=net.getLogicalNextHop(thisHop, routeDst);
		
		while(nextHop!=null){
			Point  p1 = scaleToPoint( nextHop.getAddress() );
            Point  p2 = scaleToPoint( thisHop.getAddress() );
            if( outsidePanel(p1) && outsidePanel(p2) ) continue;
            g.setColor(edge_color);
            int xinterval = ( p1.x - p2.x > 0 ) ? (p1.x-p2.x) : (p2.x-p1.x);
            int yinterval = ( p1.y - p2.y > 0 ) ? (p1.y-p2.y) : (p2.y-p1.y);
            int width = 3;
            if( xinterval > yinterval ){
            	for( int i=0; i<width; i++) g.drawLine( p1.x, p1.y-width/2+i, p2.x, p2.y-width/2+i);
            }else{
                for( int i=0; i<width; i++) g.drawLine( p1.x-width/2+i, p1.y, p2.x-width/2+i, p2.y);
            }
            prevHop = thisHop; 
            thisHop=nextHop;
            nextHop=net.getLogicalNextHop(thisHop, routeDst);
            if(prevHop == nextHop)
            	break;
		}
	}
    /**
     * draw the multicast tree rooted at multicastRoot
     */
    private void drawMulticastTree( Graphics g ){
        OverlayNode root = this.multicastRoot;
        OverlayNetwork net = the_gui.getOverlay();
        if( net.active_nodes.contains(root) == false ) return ;

        Color edge_color = Color.blue;
        //Graphics g = this.getGraphics();

        OverlayNode [] nodequeue = new OverlayNode[net.active_nodes.size()];
        int tail = 0, head = 1;
        nodequeue[0] = root;

        while( tail != head ){
            OverlayNode parent = nodequeue[tail++];
            Enumeration e = parent.getLogicalNeighbor();
            while( e.hasMoreElements() ){
                OverlayNode child = (OverlayNode)e.nextElement();
                
                if( net.getLogicalNextHop(child, root) == parent ) {
                	if(head>=nodequeue.length)
                		break;
                	
                    nodequeue[head++] = child;
                    //lotusDebug.errorln("add a child");
                    
                    Point  p1 = scaleToPoint( parent.getAddress() );
                    Point  p2 = scaleToPoint( child.getAddress() );
                    if( outsidePanel(p1) && outsidePanel(p2) ) continue;
                    g.setColor(edge_color);
                    int xinterval = ( p1.x - p2.x > 0 ) ? (p1.x-p2.x) : (p2.x-p1.x);
                    int yinterval = ( p1.y - p2.y > 0 ) ? (p1.y-p2.y) : (p2.y-p1.y);
                    int width = 3;
                    if( xinterval > yinterval ){
                        for( int i=0; i<width; i++) g.drawLine( p1.x, p1.y-width/2+i, p2.x, p2.y-width/2+i);
                    }else{
                        for( int i=0; i<width; i++) g.drawLine( p1.x-width/2+i, p1.y, p2.x-width/2+i, p2.y);
                    }
                }
            }
        }
    }


    /**
     * process the selection event of the popup menu items
     */
    public void actionPerformed(ActionEvent ae){
		if( mi_sim == null ) return;
		/*
        int i;
        for( i=0; i<mi_sim.length; i++){
            if( mi_sim[i] == ae.getSource() ) break;
        }
        if( i == mi_sim.length ) return;
        //match one of the
        //System.out.println("match one of the popup");
        if(selectedNode!=null)
        	the_gui.printWarning("fire event at node (" + selectedNode.getAddress().toString() + ")" );
        synchronized(the_gui.getOverlay()){
        	try{
        	    the_gui.getOverlay().fireMenu( selectedNode, i );
			}catch(Exception e){
            	
			}
        }
        repaint();
        */
     
    }

	public void initNode(OverlayNode onode){
		float speed;
		try{
			speed=Float.parseFloat(the_gui.tf_speed.getText());
		}catch(Exception e){
			speed=0;
			
		}
		switch(node_type){
			case NODE_STATIC_WIRED:
				onode.setMobility("");
				onode.setLinkType(OverlayNode.INTERFACE_WIRED);
				break;
			case NODE_STATIC_WIRELESS:
				onode.setMobility("");
				onode.setLinkType(OverlayNode.INTERFACE_WIRELESS);
				break;
			case NODE_STATIC_BOTH:
				onode.setMobility("");
				onode.setLinkType(OverlayNode.INTERFACE_BOTH);
				break;
			case NODE_MOBILE_WIRELESS:
				onode.setMobility(mobilityMode, speed);
				onode.setLinkType(OverlayNode.INTERFACE_WIRELESS);
				break;
		}
	}

  
    Point arrowStart;
    Point arrowEnd;
    OverlayNode lineStart;
    Point linePrev;
    boolean dragging = false;
    boolean dragged = false;
    XYAddress dragStart;
    boolean selected = false;
    
    boolean win_select = false;
    XYAddress selectStart;
    XYAddress selectEnd;

    /**
     * Handle of mouse pressed event
     * Left button:
     * arrow mode, record the coordinates to move
     * node mode, add a node
     * line mode, see if it is a possible starting point
     * denode mode, see if a node is selected
     * deline mode, see if a line is selected
     * Right button:
     * popup the menu
     */
    public void mousePressed(MouseEvent me) {
        /* do not respond while in simulation mode */
        //if( the_gui.animated() ) return;
    	OverlayNode test = the_gui.getOverlay().findClosest(
							    scaleToNode(me.getPoint()),
							    (int)(GUI.ndDiameter/scale) );

    	if( test != null && test.getProtocol()!=null ) the_gui.printWarning( 
						 test.getProtocol().getName());
    	selected = false;
    	dragging = false;
    	win_select = false;
    	dragged = false;
    	
        int mods = me.getModifiers();
        
        //ignore right button
        if ((mods & MouseEvent.BUTTON3_MASK) != 0  )
        	return;
        
        if(edit_mode!=EDIT_MOVE){
        	the_gui.getOverlay().setSelectedNode(null, false);
        }
        
        /* process left button click */
        switch (edit_mode) {
           	case EDIT_ARROW:
                arrowStart = me.getPoint();
                
                break;
            case EDIT_MOVE:
            	//the_gui.printWarning("Hold CTRL to select multiple nodes");
            	if(test!=null && the_gui.getOverlay().isNodeSelected(test)){
            		//do nothing
            	}else{
	            	this.the_gui.getOverlay().setSelectedNode(test, appendSelectedNode);
	            	selected = true;
		            repaint();
            	}	
				if( test == null ) {
					win_select = true;
					selectStart = scaleToNode(me.getPoint());
					selectEnd = selectStart;
					arrowStart = me.getPoint();
					arrowEnd = arrowStart;
				}else{
	            	arrowStart = me.getPoint();
	            	dragging=true;
	            	dragged = false;
	            	dragStart = scaleToNode(me.getPoint());
				}	
            	break;
            	
            case EDIT_NODE:
                if( test != null ) return;
				if ((mods & MouseEvent.BUTTON1_MASK) == 0  ) return;
				
        	    Graphics g = this.getGraphics();
                Point p = me.getPoint();

                synchronized(the_gui.getOverlay()){
					XYAddress temp = scaleToNode(p);
					PhyNode t = new PhyNode(temp.getX(), temp.getY());
					OverlayNode onode = the_gui.gOverlay.addNode(t );
					initNode(onode);
					
					the_gui.gOverlay.activateNode(onode);
					
	                g.setColor(OverlayNetwork.phyNodeColor);
					g.fillOval(p.x - GUI.ndDiameter/2,
				       p.y - GUI.ndDiameter/2,
				       GUI.ndDiameter,
				       GUI.ndDiameter);
				    
				    this.repaint();
				    
			    }
                break;
          
            case EDIT_DENODE:
                if( test == null ) return;
                synchronized(the_gui.getOverlay()){
                	if(the_gui.getOverlay().active_nodes.contains(test)){
                		if(!the_gui.gOverlay.deactivateNode(test))
                			break;
                	}
                    the_gui.gOverlay.removeNode( test );
                }
                repaint();
                break;
                
            case EDIT_LINE:
                if( test == null ) return;
                lineStart = test;
                linePrev = me.getPoint();
                break;
            case EDIT_DELINE:
                OverlayNode []nodeArray = new OverlayNode [2];
                if( the_gui.gOverlay.findClosestLine(
                    scaleToNode( me.getPoint() ),
                    (int)(GUI.ndDiameter/scale), nodeArray ) == true ) {
                        synchronized(the_gui.getOverlay()){
                            the_gui.gOverlay.removeEdge(nodeArray[0], nodeArray[1]);
                        }
                        repaint();
                }
                break;
            
            
			case EDIT_START:
                if( test == null ) return;
                synchronized(the_gui.getOverlay()){
                	
					if(the_gui.getOverlay().active_nodes.contains(test))
						the_gui.gOverlay.deactivateNode(test);
					else the_gui.gOverlay.activateNode(test);
                }
                repaint();
                break;
            case EDIT_TREE:
                multicastRoot = test;
                this.repaint();
                //drawMulticastTree( test );
                break;
            case EDIT_SRC:
            	routeSrc=test;
            	this.repaint();
            	break;
            case EDIT_DST:
            	routeDst=test;
            	this.repaint();
            	break;
        }
    }

    /**
     * Process mouse drag event. Currently, only two edit modes
     * utilize mouse drag, arrow and line modes
     */
    public void mouseDragged(MouseEvent me) {
        //if( the_gui.animated() ) return;
        int mods = me.getModifiers();
	    if ((mods & MouseEvent.BUTTON1_MASK) == 0 ) return; //not left button

        if( edit_mode == EDIT_ARROW ) {
            Point p = me.getPoint();
            this.setView( (int)(viewX - (p.x-arrowStart.x)/scale),
                          (int)(viewY - (p.y-arrowStart.y)/scale) );
            arrowStart = p;
            //XYAddress addr = scaleToNode( new Point( p.x - arrowStart.x, p.y - arrowStart.y ) );
            //this.setView( viewX - addr.getX(), viewY - addr.getY());
            repaint();
        }else if(edit_mode == EDIT_MOVE){
        	//the_gui.printWarning("Hold CTRL to select multiple nodes");
        	Point p = me.getPoint();
        	if(dragging){
        		dragged = true;
        		OverlayNode selected []= the_gui.getOverlay().getSelectedNodes();
        		XYAddress mAddr = scaleToNode(p);
        		for(int i=0;i<selected.length;i++){
	        		OverlayNode dragNode = selected[i];
	        		
	        		XYAddress oldAddr = dragNode.getAddress();
	        		XYAddress newAddr = new XYAddress(oldAddr.getXF()+(mAddr.getXF()-dragStart.getXF()), oldAddr.getYF()+(mAddr.getYF()-dragStart.getYF()));
	        		
	        		
		        	dragNode.setAddress(newAddr);
		        	
        		}
        		dragStart = mAddr;
        		repaint();
	        	
        	}else if(win_select){
        		arrowEnd = p;
        		selectEnd = scaleToNode(p);
        		repaint();
        	}
        	
        }
        
       	else if( edit_mode == EDIT_LINE ) {
            if( lineStart == null ) return;
            Graphics g = this.getGraphics();
            Point p = me.getPoint();
            Point startP = scaleToPoint( lineStart.getAddress() );
            //the_gui.printWarning( "x:" + p.x + "," + "y:" + p.y );
            //g.setColor( GUI.dwColor );
            g.setXORMode( GUI.dwColor );
            g.drawLine( startP.x, startP.y, linePrev.x, linePrev.y);
            //repaint();
            //g.setColor( GUI.edgeColor );
            g.drawLine( startP.x, startP.y, p.x, p.y);
            linePrev = p;
        }
    }

    boolean showPopup;
    Point popupPoint;
    public void mouseReleased(MouseEvent me) {
        //if( the_gui.animated() ) return;
	    int mods = me.getModifiers();

        Point p = me.getPoint();
    	OverlayNode test = the_gui.getOverlay().findClosest(
							    scaleToNode(p),
							    (int)(GUI.ndDiameter/scale) );
    	
    	showPopup = false;
    	dragging = false;
	   
	   	if ((mods & MouseEvent.BUTTON3_MASK) != 0  ) {
            if( me.isPopupTrigger() == false ) return;
            
            if(test!=null && !selected && !this.the_gui.getOverlay().isNodeSelected(test)){
            	this.the_gui.getOverlay().setSelectedNode(test, appendSelectedNode);
            	repaint();
            }else if (test==null){
            	this.the_gui.getOverlay().setSelectedNode(test, appendSelectedNode);
            	repaint();
            }
            
            //if( test == null ) return;
            if(pm_edit!=null)
            	this.remove(pm_edit);
            
           
            updatePopupMenu();
            
            
            this.add(pm_edit);
            //pm_edit.show(this, p.x, p.y);
            showPopup = true;
            popupPoint = p;
            
            this.repaint();
            
            (new DelayedPopupShower()).start();
            
            return;
        }
        else if( (mods & MouseEvent.BUTTON1_MASK) != 0 ) {
        	if( edit_mode ==EDIT_MOVE){
        		//the_gui.printWarning("Hold CTRL to select multiple nodes");
        		if(!selected && !dragged && !win_select){
        			dragged = false;
        			this.the_gui.getOverlay().setSelectedNode(test, appendSelectedNode);
        			repaint();
        		}
        		
        		dragging = false;
        		if(win_select){
        			float startX, startY, endX, endY;
                	if(selectStart.x<selectEnd.x){
                		startX = selectStart.x;
                		endX = selectEnd.x;
                	}
                	else{
                		startX = selectEnd.x;
                		endX = selectStart.x;
                	}
                	
                	if(selectStart.y<arrowEnd.y){
                		startY = selectStart.y;
                		endY = selectEnd.y;
                	}
                	else{
                		startY = selectEnd.y;
                		endY = selectStart.y;
                	}
                	
                	the_gui.getOverlay().setSelectedNodes(new XYAddress(startX, startY), new XYAddress(endX, endY), appendSelectedNode);
                	
        			win_select = false;
        			repaint();
        		}
            	
	            return;
        	}
            if( edit_mode != EDIT_LINE ) return;
            if( lineStart == null ) return;
            
            if( test != null ) {
            	
                //can do an connection
                XYAddress sXY = lineStart.getAddress();
                XYAddress eXY = scaleToNode(p);
                float dist = sXY.distanceTo( eXY );
                synchronized(the_gui.getOverlay()){
                    the_gui.getOverlay().addEdge( lineStart, test, dist );
                }
                repaint();
            }else{
           	    Graphics g = this.getGraphics();
                g.setColor(GUI.dwColor);
                Point startP = scaleToPoint( lineStart.getAddress() );
                g.drawLine( startP.x, startP.y, linePrev.x, linePrev.y);
            }
            lineStart = null;
        }
    }
	MenuItem []mi_sim ;

    /**
     * mouse event summary
     * a. Left button
     *      a.1 pressed: line start if in line mode
     *      a.2 dragged: line draw if in line mode
     *      a.3 released: line end if in line mode
     *                    node add if in node mode
     * b. Right button
     *      b.1 released: popup menu depending on the state
     */
    private void updatePopupMenu(){
    	/*
        if( null != mi_sim ){
            for( int i=0; i<mi_sim.length; i++){
                pm_edit.remove( mi_sim[i] );
                mi_sim[i]=null;
            }
            mi_sim=null;
        }

        if( null == the_gui.getOverlay().getMenus() ) return;
        // compose the menu
        String[] menuName = the_gui.getOverlay().getMenus();
        mi_sim = new MenuItem[menuName.length];
        for( int i=0; i<menuName.length; i++){
            mi_sim[i] = new MenuItem( menuName[i] );
            pm_edit.add(mi_sim[i]);
            mi_sim[i].addActionListener(this);
        }
        */
    	pm_edit = the_gui.getOverlay().getMenus();

    }

    public void mouseEntered(MouseEvent e) {

    }

    public void mouseExited(MouseEvent e) {

    }

    public void mouseClicked(MouseEvent e) {

    }

    public void mouseMoved(MouseEvent e) {
    }

    void reset(){
        restore();
    }
	
	void setMobileMode(String mode){
		mobilityMode = mode;
	}
    /** @return the current edit mode */
    int getEditMode()  { return edit_mode; }
    
    /** set node type 
      * added by Guangyu
      */
    void setNodeType(int type){
    	node_type=type;
    }
    
    
    /** set the edit mode */
    void setEditMode( int mode ) {
    	
    	if(mode!=GUI_DisplayWindow.EDIT_MOVE){
    		the_gui.getOverlay().setSelectedNode(null, false);
    	}
    	if(mode==GUI_DisplayWindow.EDIT_MOVE){
    		the_gui.printWarning("Hold CTRL to select multiple nodes");
    	}
        switch (mode){
        case GUI_DisplayWindow.EDIT_ARROW:
        case GUI_DisplayWindow.EDIT_MOVE:
        	this.setCursor(new Cursor(Cursor.HAND_CURSOR) );
            break;
         
        default:
        	this.setCursor( new Cursor( Cursor.DEFAULT_CURSOR) );
            
        }
        edit_mode = mode;
        if( mode == EDIT_SRC || mode==EDIT_DST ) multicastRoot = null;
        if( mode == EDIT_TREE ){
        	routeSrc=null;
        	routeDst=null;	
        }
    }

    /**
     * change of the view  point and scale
     */
    public void scaleBy( float factor ) {
        scale *= factor;
        the_gui.printWarning("scale:"+scale);
    }

    /**
     * restore the scale to the default value
     */
    public void restore( ) {
        scale = 1.0f;
        viewX = 0;
        viewY = 0;
        the_gui.printWarning("scale:"+scale + "; viewX:" + viewX + "; viewY:" + viewY);
    }

    /**
     * fit the display to the window
     */
    void bestfit( ) {
    //scale the topology to fit into the display
        if( the_gui.getOverlay().getPhyNodes().size() <= 1 ) return;

       	Enumeration nodes = the_gui.getOverlay().getPhyNodes().elements();

        int fxmax = Integer.MIN_VALUE;
        int fymax = Integer.MIN_VALUE;
        int fxmin = Integer.MAX_VALUE;
        int fymin = Integer.MAX_VALUE;

        while( nodes.hasMoreElements() ) {
            OverlayNode onode = (OverlayNode) nodes.nextElement();
            XYAddress addr = onode.getAddress();

            if( addr.getX() > fxmax ) fxmax = addr.getX();
            if( addr.getY() > fymax ) fymax = addr.getY();
            if( addr.getX() < fxmin ) fxmin = addr.getX();
            if( addr.getY() < fymin ) fymin = addr.getY();
        }

        Dimension dim = this.getSize();
        double dimy = dim.getHeight();
        double dimx = dim.getWidth();

        viewX = fxmin - 10;
        viewY = fymin - 10;

        double scaleX = dimx/(fxmax - fxmin+20);
        double scaleY = dimy/(fymax - fymin+20);
        if( scaleX > scaleY ) scale = (float)scaleY;
        else scale = (float)scaleX;
        the_gui.printWarning("scale:"+scale + "; viewX:" + viewX + "; viewY:" + viewY);
    }

    void setView( int x, int y ){
        viewX = x; viewY = y;
    }

    Point scaleToPoint( XYAddress addr ) {
        return new Point( (int)( (addr.getX() - viewX)*scale ),
                          (int)( (addr.getY() - viewY)*scale ) );
    }

    XYAddress scaleToNode( Point p ) {
        return new XYAddress( (int) ( p.x/scale + viewX ),
                              (int) ( p.y/scale + viewY ) );
    }
	
    
    
} // End of GUI_DisplayWindow class
	


