/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team
  * July, 2002 
  */
 
package lotos;

import lotos.*; 
import java.awt.*;

/**
 *  This class consists of a bunch of helper functions for working with Points
 */

class G_Point
{

/**
 * compare two points
 */
public final static boolean greaterThan(Point a, Point b) {
	return (a.y > b.y) || ((a.y == b.y) && (a.x > b.x));
    }

public static double distance(Point a, Point b)
  {
    int dx = a.x - b.x;
    int dy = a.y - b.y;

    return Math.sqrt(dx*dx + dy*dy);
  }

    /** angle of datum to positive X axis at pivot */
static public double angleToHorizon(Point pivot, Point datum)
       // calculates the angle from this address to the given address
       // returns an angle between 0 and 360
  {
    return Math.atan2((long) datum.y - (long) pivot.y, (long) datum.x - (long) pivot.x);
  }

/**
 * angle of reference, pivot, datum.
 * @return positive value for datum CW of reference, negative for datum CCW of reference
 */
static public double angleRelative(Point reference, Point pivot, Point datum)
    {
        double angle_reference_pivot_datum = angleToHorizon(pivot, datum) - angleToHorizon(pivot, reference);

        if (angle_reference_pivot_datum > Math.PI)
            return angle_reference_pivot_datum - 2.0 * Math.PI;
        else if (angle_reference_pivot_datum <= -Math.PI)
            return angle_reference_pivot_datum + 2.0 * Math.PI;
        else
            return angle_reference_pivot_datum;
    }

    /**
     * calculates the angle from this address to the given address
     * @return an angle between 0 and 360
     */
    static public double angleToHorizon(DTAddress pivot, DTAddress datum)
    {
        return Math.atan2( datum.y - pivot.y, datum.x -  pivot.x);
    }

    static public double angleRelative( DTAddress reference, DTAddress pivot, DTAddress datum ){
        double angle_reference_pivot_datum = angleToHorizon(pivot, datum) - angleToHorizon(pivot, reference);

        if (angle_reference_pivot_datum > Math.PI)
            return angle_reference_pivot_datum - 2.0 * Math.PI;
        else if (angle_reference_pivot_datum <= -Math.PI)
            return angle_reference_pivot_datum + 2.0 * Math.PI;
        else
            return angle_reference_pivot_datum;
    }

/**
 * Returns true if datum lies on the circle inscribing the points a, b, and c.
 * (Actually, it affords an error of +/-G_DoublePoint.EPSILON.)
 * If no circle exists (all points are colinear) then it returns false.
 */

static public boolean isOnCircle(Point a, Point b, Point c, Point datum)
{
    G_DoublePoint center = findCenterOfCircle(a, b, c);

    if (center == null) // a,b,c are colinear
        return false;

    double center_to_datum = G_DoublePoint.distance(center, new G_DoublePoint(datum));
    double center_to_a = G_DoublePoint.distance(center, new G_DoublePoint(a));


    if ( center_to_datum < center_to_a + G_DoublePoint.EPSILON &&
        center_to_datum > center_to_a - G_DoublePoint.EPSILON)
        return true;
    else
        return false;
}


/**
 * Returns true if datum lies inside the circle inscribing the points a, b, and c.
 * There is no accountment for error here.
 * If no circle exists (all points are colinear) then it returns true.
 */
static public boolean isInCircle(Point a, Point b, Point c, Point datum)
{

    G_DoublePoint center = findCenterOfCircle(a, b, c);

    if (center == null) // a,b,c are colinear
        return true;


    if (G_DoublePoint.distance(center, new G_DoublePoint(datum)) < G_DoublePoint.distance(center, new G_DoublePoint(a)))
        return true;
    else
        return false;
}


/**
 * Returns true if datum lies inside or on the circle inscribing the points a, b, and c.
 * There is no accountment for error here.
 * If no circle exists (all points are colinear) then it returns true.
 */
static public boolean isInOrOnCircle(Point a, Point b, Point c, Point datum)
{

    G_DoublePoint center = findCenterOfCircle(a, b, c);

    if (center == null) // a,b,c are colinear
        return true;


    if (G_DoublePoint.distance(center, new G_DoublePoint(datum)) < G_DoublePoint.distance(center, new G_DoublePoint(a)) + G_DoublePoint.EPSILON)
        return true;
    else
        return false;
}



/** Find the center of the circle inscribing 3 points.
 * If all 3 points are the same, it returns that point.
 * Else, if 3 points lie on a line, it returns null.
 */

public static G_DoublePoint findCenterOfCircle(Point a, Point b, Point c)
{
    // The center is the point of intersection of any 2 of the 3 perpendicular
    // bisectors of line segments connecting a, b, and c.
    // We now select 2 line segments so that their perpendicular bisectors
    // do not have infinite slopes.
    // p1, p2 are the endpoints of the 1st line segment
    // q1, q2 are the endpoints of the 2nd line segment

	Point p1, p2;
	Point q1, q2;
    if (a.y == b.y && b.y == c.y) // all colinear
    {
        if (a.x == b.x && b.x == c.x)  // all the same point!
            return new G_DoublePoint(a);
        else
            return null;
    }
    else if (a.y != b.y && a.y != c.y)
    {
	    p1 = a;
	    p2 = b;
	    q1 = a;
	    q2 = c;
    }
    else if (a.y != b.y && b.y != c.y)
    {
	    p1 = a;
	    p2 = b;
	    q1 = b;
	    q2 = c;
    }
    else // a.y == b.y && b.y != c.y && a.y != c.y
    {
	    p1 = b;
	    p2 = c;
	    q1 = a;
	    q2 = c;
    }


    // calculate x-intercepts and slopes for two lines.
    // (NOTE: divide by zero should never happen, but
    // given that the slopes could be infinitely small,
    // I'll catch the expeption anyway.)

    try
    {
        double f, s;   // x intercept and slope for first line

        s = ((long)p2.x - (long)p1.x) / (double) ((long)p1.y - (long)p2.y);
        f = ((long)p1.y + (long)p2.y)/2.0 + (((long)p1.x + (long)p2.x)*-0.5*s);

        double g, t;   // x intercept and slope for second line

        t = ((long)q2.x - (long)q1.x) / (double) ((long)q1.y - (long)q2.y);
        g = ((long)q1.y + (long)q2.y)/2.0 + (((long)q1.x + (long)q2.x)*-0.5*t);

        /*
          System.out.print("Line 1: Yintercept = ");
          System.out.print(f);
          System.out.print("   Slope = ");
          System.out.println(s);
          System.out.print("Line 2: Yintercept = ");
          System.out.print(g);
          System.out.print("   Slope = ");
          System.out.println(t);
          */

        if (s == t)  // either 0 or infinite solutions...
            return null;
        else
            return G_DoublePoint.findInterceptOfLines(f, s, g, t);
    }
    catch (ArithmeticException ae)
    {
        System.err.println("Caught ArithmeticException in Point.findCenterOfCircle: " + ae.getMessage());
        return null;  //seems safe to say there is no center...
    }
}

public static double distanceToLine(Point line_end1, Point line_end2, Point data)
        {
            if (line_end1.y == line_end2.y)
            {
                if ((line_end1.x < line_end2.x && line_end1.x < data.x && data.x < line_end2.x) ||
                    (line_end2.x < line_end1.x && line_end2.x < data.x && data.x < line_end1.x))
                    return Math.abs(data.y - line_end1.y);
            }
            else if (line_end1.x == line_end2.x)
            {
                if ((line_end1.y < line_end2.y && line_end1.y < data.y && data.y < line_end2.y) ||
                    (line_end2.y < line_end1.y && line_end2.y < data.y && data.y < line_end1.y))
                    return Math.abs(data.x - line_end1.x);
            }
            else
            {
                double slope = (line_end1.y - line_end2.y)/ (double) (line_end1.x - line_end2.x);
                double slope_perpendicular = -1.0/slope;

                G_DoublePoint intercept = G_DoublePoint.findInterceptOfLines(line_end1.y - (slope*line_end1.x), slope,
                                                                              data.y - (slope_perpendicular*data.x), slope_perpendicular);


                if ((line_end1.y < line_end2.y && line_end1.y < intercept.y && intercept.y < line_end2.y) ||
                    (line_end2.y < line_end1.y && line_end2.y < intercept.y && intercept.y < line_end1.y))
                    return G_DoublePoint.distance(intercept, new G_DoublePoint(data));
            }

            return Math.min(G_Point.distance(line_end1, data), G_Point.distance(line_end2, data));
        }
}

class G_DoublePoint {
public static double EPSILON = 1e-7;

    double x;
    double y;

    public G_DoublePoint(Point p) {
	x = p.x;
	y = p.y;
    }

    public G_DoublePoint(double xx, double yy) {
	x = xx;
	y = yy;
    }


public static double distance(G_DoublePoint a, G_DoublePoint b)
  {
    double dx = a.x - b.x;
    double dy = a.y - b.y;

    return Math.sqrt(dx*dx + dy*dy);
  }

public String toString()
    {
        return "" + x + "," + y;
    }


/** Returns the point at the interception of two lines.
 * (f,s) and (g,t) are the y-intercept and slopes of two lines.
 *
 * <p> <code> x = (g-f)/(s-t); </code>
 * <p> <code> y = f + s*x; </code>
 *
 * @throws ArithmeticException if s == t
 */
public static G_DoublePoint findInterceptOfLines(double f, double s, double g, double t)
    {
        // Initial solution:  (caused divide by zero error)
        //double y = (f/s - g/t)/(1/s - 1/t);
        //double x = (y - f)/s;

        // solve without dividing by an individual slope.
        double x = (g-f)/(s-t);
        double y = f + s*x;

        return new G_DoublePoint(x, y);
    }

}




