/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 
package lotos;

import lotos.*; 
import java.awt.Color;

import hypercast.HyperCastConfig;
import hypercast.I_Node;
import hypercast.I_UnicastAdapter;

/**
 * 
 * @author HyperCast Team
 * @author Guangyu, Dong
 *
 * User implements this interface to put a new HyperCast protocol into the 
 * LoTos simulator.
 * 
 * The instance of this interface will be dynamically loaded, so use constructor
 * without parameter only.
 * 
 */
public interface I_SimHyperCastAgent {
	
	/**
	 * Get the name of this protocol
	 * @return
	 */
	public String getName();
	/**
	 * Set the network object
	 * @param network
	 */
	public void setNetwork(simHCNetwork network);
	/**
	 * create the next I_Node object to be put into the simulator.
	 * 
	 * If needServers() return true, then this function create a server node
	 * that is needed to run the protocol.
	 * 
	 * @param coords	the coordinates of this node on the LoTos window
	 * @param adapter	the node adapter
	 * @param info		if is not null, use this info to create a new node
	 * @return
	 */
	public I_Node createNode(XYAddress coords, HyperCastConfig config, I_UnicastAdapter adapter, String info);
	
	/**
	 * Check if need to create some servers before running the protocol.
	 * If the result is true, those servers can be created by calling createNode
	 * @return
	 */
	public boolean needServers(HyperCastConfig config);
	/**
	 * Get the information string of the current node, which can be used to restore it in another session
	 * @return
	 */
	public String getInfoForNode(I_Node node);
	
	/**
	 * Get the labels of the menu items for this protocol.
	 * 
	 * There should not have duplicate menu names in this list
	 * 
	 * @return
	 */
	public String [] getMenuItems();
	
	/**
	 * Fires the menu item by the menu's name.
	 * 
	 * @param menu
	 */
	public void menuFired(String menu);
	
	/**
	 * Get the color for a I_Node object. The resulting color should depends on
	 * the current state of the node.
	 * @param node
	 * @return
	 */
	public Color getColorForNode(I_Node node);
	
	/**
	 * Get the shape for a I_Node object. The resulting shape should depends on
	 * the type of the node.
	 * 
	 * @param node
	 * @return
	 */
	public int getShapeForNode(I_Node node);
	
	/**
	 * Get the display label for a node. 
	 * @param node
	 * @return
	 */
	public String getLabelForNode(I_Node node);
	
	
	/**
	 * Called when a protocol node is removed from LoTos
	 * @param node
	 * @return true if this protocol node can be removed
	 */
	public void nodeRemoved(I_Node node);
	
	
}
