/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team
  * July 2002
  */

package lotos;

import lotos.*; 
import java.util.*;

/**
 * A priority queue that is implemented with dijkstra algorithm
 */

class L_PriorityQueue_Dijkstra implements L_PriorityQueue
{
    /**the vector that stores each key/object pair*/
    L_PDQ_Entry the_queue[];
    int queue_end;

    /** a mapping from an object to its L_PDQ entry */
    Hashtable object_to_L_PDQ_Entry;

    /**
     * Constructor
     * @param size the size of the queue
     */

    L_PriorityQueue_Dijkstra(int size)
    {
        the_queue = new L_PDQ_Entry[size+1];
        queue_end = 1;

        object_to_L_PDQ_Entry = new Hashtable();
    }

    public final boolean isEmpty()
    {
        return queue_end == 1;
    }

    private final float keyAt(int i) {
        return the_queue[i].key;
    }

private final void swap(int i, int j) {
    L_PDQ_Entry old_i = the_queue[i];
    L_PDQ_Entry old_j = the_queue[j];

    old_i.index = j;
    old_j.index = i;

    the_queue[i] = old_j;
    the_queue[j] = old_i;
}

private final void moveUpward(int i) {
	int j = i/2;
	while (i > 1 && keyAt(i) < keyAt(j)) {
	    swap(i, j);
	    i = j;
	    j = i/2;
	}
}

private final void moveDownward(int i) {
    int left = i*2;
    int right = i*2 + 1;

    float i_key = keyAt(i);

    while (left < queue_end) {
        float left_key = keyAt(left);
        float min_key = left_key;

        if (right < queue_end)
        {
            float right_key = keyAt(right);
            if (right_key < min_key)
                min_key = right_key;
        }

        // if its proper, we're done
        if (i_key <= min_key)
            break;

        // else swap with left or right
        if (min_key == left_key) {
            swap(i, left);
            i = left;
        }
        else {
            swap(i, right);
            i = right;
        }

        left = i*2;
        right = i*2 + 1;
    }
}

public final void insert(float key, Object o)
    {
        int i = queue_end;
        the_queue[i] = new L_PDQ_Entry(key, o, i);
        queue_end++;

        object_to_L_PDQ_Entry.put(o, the_queue[i]);

        moveUpward(i);
    }

    // returns -1 if empty.
    public final float minKey()
    {
        if (isEmpty())
            return -1;
        else
            return keyAt(1);
    }

    public final Object minValue()
    {
        if (isEmpty())
            return null;
        else
            return the_queue[1].value;
    }

    public final Object removeMin()
    {
        if (isEmpty())
            return null;

        Object datum = the_queue[1].value;

        int last_place = queue_end-1;
        swap(1, last_place);

        queue_end--;
        object_to_L_PDQ_Entry.remove(datum);

        if (last_place != 1)
            moveDownward(1);

        return datum;
    }

public final float getKeyOf(Object o)
    {
        L_PDQ_Entry ptr = (L_PDQ_Entry) object_to_L_PDQ_Entry.get(o);

        if (ptr == null)
            return -1;
        else
            return ptr.key;
    }

public final void setKeyOf(Object o, float value)
    {
        remove(o);
        insert(value, o);
    }

public final boolean contains(Object o)
    {
        return object_to_L_PDQ_Entry.containsKey(o);
    }


public final void remove(Object o)
    {
        L_PDQ_Entry ptr = (L_PDQ_Entry) object_to_L_PDQ_Entry.get(o);
        if (ptr == null)
            return;

        int index = ptr.index;

        int last_place = queue_end-1;
        swap(index, last_place);

        queue_end--;
        object_to_L_PDQ_Entry.remove(o);

        if (index != last_place) {
            // only one of these should do anything...
            // calling the other doesn't hurt.
            moveUpward(index);
            moveDownward(index);
        }
    }

private final void printState() {
	int i = 1;
	int j = 1;
	int d = 2;

	while (i < queue_end)
    {
        if (the_queue[i] == null)
            System.out.print("null ");
        else
            System.out.print(the_queue[i].value.toString() + " ");

        if (i == j) {
            System.out.print("| ");
            j += d;
            d *= 2;
        }
        i++;
    }
    lotusDebug.println();
}

public static void main( String[] argv) {
	L_PriorityQueue_Dijkstra q = new L_PriorityQueue_Dijkstra(20);

	q.insert((float) 1.0, "a");
	q.insert((float) 10.0, "j");
	q.insert((float) 3.0, "c");
	q.insert((float) 11.0, "k");
	q.insert((float) 12.0, "l");
	q.insert((float) 4.0, "d");
	q.insert((float) 14.0, "n");
	q.insert((float) 15.0, "o");
	q.insert((float) 2.0, "b");
	q.insert((float) 9.0, "i");
	q.insert((float) 17.0, "q");
	q.insert((float) 7.0, "g");
	q.insert((float) 16.0, "p");

	q.remove("b");
	q.remove("k");
	q.printState();

	lotusDebug.println((String) q.removeMin());
	lotusDebug.println((String) q.removeMin());
	lotusDebug.println((String) q.removeMin());
	lotusDebug.println((String) q.removeMin());
 	lotusDebug.println((String) q.removeMin());

	q.insert((float) 8.0, "h");
	q.insert((float) 5.0, "e");
	q.insert((float) 6.0, "f");
	q.insert((float) 13.0, "m");

	lotusDebug.println((String) q.removeMin());
	lotusDebug.println((String) q.removeMin());
	lotusDebug.println((String) q.removeMin());
	lotusDebug.println((String) q.removeMin());
	lotusDebug.println((String) q.removeMin());

	lotusDebug.println((String) q.removeMin());
	lotusDebug.println((String) q.removeMin());
	lotusDebug.println((String) q.removeMin());
	lotusDebug.println((String) q.removeMin());
    }
}

/** an entry of the priority queue, including a key/object pair*/
class L_PDQ_Entry
{
    float key;
    Object value;
    int index;

    L_PDQ_Entry(float k, Object o, int i)
    {
        key = k;
        value = o;
        index = i;
    }
}
