/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team
  * 
  */

package lotos;

import lotos.*; 
import hypercast.HyperCastConfig;
import hypercast.HyperCastStatsException;
import hypercast.I_MulticastAdapter;
import hypercast.I_Node;
import hypercast.I_UnicastAdapter;
import hypercast.DT.DT_LogicalAddress;
import hypercast.HC.HC_LogicalAddress;
import hypercast.HC.HC_Node;
import hypercast.Pastry.Pastry_LogicalAddress;
import hypercast.Pastry.Pastry_Node_BuddyList;
import hypercast.SPT.SPT_Node;
import hypercast.util.XmlUtil;

import java.awt.Color;
import java.util.StringTokenizer;

import javax.xml.transform.TransformerException;


public class PastryAgent implements I_SimHyperCastAgent {

	simHCNetwork network;
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getName()
	 */
	public String getName() {
		return "Pastry";
	}
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getLabelForNode(hypercast.I_Node)
	 */
	public String getLabelForNode(I_Node node) {
		return "pastry";
	}
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#setNetwork(simHCNetwork)
	 */
	public void setNetwork(simHCNetwork network) {
		this.network = network;
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#createNode(OverlayNode, HC_Adapter)
	 */
	public I_Node createNode(XYAddress coords, HyperCastConfig config, I_UnicastAdapter adapter, String info) {
		if(info==null)
			return new Pastry_Node_BuddyList(config, (I_MulticastAdapter)adapter, "/Public/Node/Pastry/", null);
		else{
			StringTokenizer tokens = new StringTokenizer(info, ":");
			String type = tokens.nextToken();
			String laStr = tokens.nextToken();
			Pastry_LogicalAddress la = new Pastry_LogicalAddress(laStr);
			return new Pastry_Node_BuddyList(config, (I_MulticastAdapter)adapter, "/Public/Node/Pastry/", la);
		}
	}
	
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getInfoForNode(hypercast.I_Node)
	 */
	public String getInfoForNode(I_Node node) {
		String laStr = node.getAddressPair().getLogicalAddress().toString();
		return this.getName()+":"+laStr.substring(3, laStr.length()-1);
	}
	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getMenuItems(java.lang.String)
	 */
	public String[] getMenuItems() {
		return new String[0];
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#menuFired(java.lang.String)
	 */
	public void menuFired(String menu) {
		//no menu
		//do nothing
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getColorForNode(hypercast.I_Node)
	 */
	public Color getColorForNode(I_Node node) {
		
		String stateString = ""+Pastry_Node_BuddyList.STATE_STOPPED;
		try {
			stateString = XmlUtil.getScalarStatsResult(node, "/Node/State");
		} catch (HyperCastStatsException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (TransformerException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		int state = Integer.parseInt(stateString);
		Color color;
		
		switch (state)
        {
            case Pastry_Node_BuddyList.STATE_STOPPED:
                color = Color.lightGray;
                break;
            case Pastry_Node_BuddyList.STATE_BUDDY_SEARCH:
                color = Color.red;
                break;
            case Pastry_Node_BuddyList.STATE_BUDDY_FOUND:
                color = Color.pink;
                break;
            case Pastry_Node_BuddyList.STATE_JOIN_INITIATED:
                color = Color.orange;
                break;
            case Pastry_Node_BuddyList.STATE_JOINED:
                color = Color.yellow;
                break;
            default:
                // TODO: error?
            	color = Color.GRAY;
        }
		return color;
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#getShapeForNode(hypercast.I_Node)
	 */
	public int getShapeForNode(I_Node node) {
		return protocolInterface.shapeCircle;
	}

	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#nodeRemoved(hypercast.I_Node)
	 */
	public void nodeRemoved(I_Node node) {
		
	}


	/* (non-Javadoc)
	 * @see I_SimHyperCastAgent#needServers()
	 */
	public boolean needServers(HyperCastConfig config) {
		return false;
	}
	

}
