/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, July 2002 
  * 
  */
package lotos;

import lotos.*; 
/**
 * This class implements message event by extending SimEvent
 */

abstract class SimEventMsg extends SimEvent
{
    //this is the physical layer routing
    /** source of the message, note that src can be different from sender in the
     *  SimEvent because a message can be forwarded for multiple hops before
     *  reaching the destination. While src and dst of the message are the same
     *  during the whole process, sender and recipient are changed at each forwarding
     *  node
     */
    OverlayNode src;
    /** destination of the message */
    OverlayNode dst;
    /** type of the message */
    int msgType;
    
    /** added by Guangyu */
    /** TTL */
    int TTL=-1;
    
    /** type of transmission */
    int transType;
    final static int TRANS_WIRED=0;
    final static int TRANS_WIRELESS=1;
    final static int TRANS_BOTH=2;
    
    public void setTransType(int tt){
    	transType=tt;	
    }
    public boolean isWired(){
    	return transType==TRANS_WIRED || transType==TRANS_BOTH;
    }
    public boolean isWireless(){
    	return transType==TRANS_WIRELESS || transType==TRANS_BOTH;
    }
	
	public int getTTL(){
		return TTL;
	}
	public void setTTL(int ttl){
		TTL=ttl;
	}
	
    public SimEventMsg(
                       OverlayNode osrc,
                       OverlayNode odst,
                       int mtype )
    {
        super( 0, null, null, SimEvent.MESSAGE_EVENT );
        src = osrc;
        dst = odst;
        msgType = mtype;
     }


    public OverlayNode getSrc()
    {
        return src;
    }

    public void setSrc( OverlayNode o)
    {
        src = o;
    }
    //public void setSrc( OverlayNode osrc ) { src = osrc; }

    public OverlayNode getDst()
    {
        return dst;
    }

    public void setDst( OverlayNode o)
    {
        dst = o;
    }

    //public void setDst( OverlayNode odst ) { dst = odst; }
    public int getMsgType(){ return msgType; }

	 public int getSize(){ return 8; }

    abstract public String getMsgLabel(int type);
    
    public abstract Object clone();
    
    public boolean equals(Object o){
    	return false;
    }
   
}
