/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, July 2002 
  * 
  */
package lotos;

import lotos.*; 
import java.util.*;

/**
 * The discrite-time simulator
 */
/*
 * revised by Guangyu Dong, Jan 6, 2004
 * to get precise microsecond level simulation time
 */
class SimEventProcessor
{
    /** current simulation time */
    /** changed to "long" by Guangyu to get precise microsecond level time */
    /** 1000000 = 1sec */
    private long sim_time;
    //float sim_time;
    
    /* set sim time with microseconds */
    void setMSSimTime(long t){
    	sim_time=t;	
    }
    
    /* set sim time with seconds */
    void setSimTime(float t){
    	sim_time=(long)(t*SimEvent.TIME_PER_SEC);	
    }
	/* @return current simulation time in second */
    float getSimTime(){
    	return (float)sim_time/SimEvent.TIME_PER_SEC;
    }
    /* @return current simulation time in miccrosecond */
    long getMSSimTime(){
    	return sim_time;	
    }
        
    /** the queue containing the simulation event */
    Vector the_queue;
	
	/** Constructor */
    public SimEventProcessor()
    {
        setMSSimTime(0);
        the_queue = new Vector();
    }

    /** @return those pending events */
    public Enumeration getPendingEvents() {
        return the_queue.elements();
    }


    /** insert an event into the simulator */
    public void insertEvent( SimEvent se)
    {
    	
        if (se.getMSTime() < getMSSimTime())
        {
            lotusDebug.error("ERROR: Event inserted for time " + se.getTime() );
            //lotusDebug.error(se.getTime());
            lotusDebug.error(" when clock was at " + sim_time );
            //lotusDebug.error(sim_time);
            lotusDebug.errorln(": Event Ignored.");
            return;
        }
        
        //added by Guangyu, Feb 10th, 2004
        //to remove duplicated timer event
		clearEvent(se);
		
        int i;
        for( i=0; i<the_queue.size(); i++){
            SimEvent event = (SimEvent) the_queue.elementAt(i);
            if( event.getTime() > se.getTime() ) break;
        }
        if( i == the_queue.size() ) the_queue.addElement(se);
        else the_queue.add( i, se );
    }

    /** @return true when the event queue is empty */
    public boolean queueIsEmpty()
    {
        if (the_queue.size() == 0 ) return true; else return false;
    }

    


    /** @return the next pending event and advance the sim time */
    public SimEvent getNextEvent()
    {
        if( the_queue.size() ==  0 ) return null;
        SimEvent se = (SimEvent) the_queue.elementAt(0);
        the_queue.remove(0);
        setMSSimTime(se.getMSTime()); //advance the time;
        return se;
    }

    /** @return the next event within the endTime
     *  If the return is null, advance the time to end time
     *  If the return is not-null, advance the time using getNextEvent
     */

    public SimEvent getNextEvent(float endTime ){

        if( the_queue.size() == 0 ) {
            setSimTime(endTime);
            return null;
        }
        SimEvent se = (SimEvent) the_queue.elementAt(0);
        if( se.getTime() > endTime ) { setSimTime(endTime); return null; }
        return getNextEvent();
    }

    /** reset the discrete-time simulator */
    public void reset(){
        the_queue.clear();
        sim_time = 0;
    }
    /**
     * Find a specified type of event
     * @param e		the event
     * @return		the delay of the event, or -1
     */
    public float getEvent(SimEvent e){
		//lotusDebug.errorln("get a event"+e);
    	for(int i=0;i<the_queue.size();i++){
    		SimEvent se=(SimEvent)the_queue.elementAt(i);
    		if(se.eventType()==e.eventType() && se.equals(e)){
    			//lotusDebug.errorln("find a event"+se);
    			return se.getTime()-this.getSimTime();
    		}
    	}
    	return -1;
    }
    /**
     * Cleare a specified type of event
     * @param e		the event
     */
    public void clearEvent(SimEvent e){
    	//if(!(e instanceof SimEventTimer))return;
    	
		for(int i=0;i<the_queue.size();i++){
			SimEvent se=(SimEvent)the_queue.elementAt(i);
			if(se.eventType()==e.eventType() && se.equals(e))
				the_queue.removeElementAt(i);
		}
	
    }
    

}
