/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, July 2002 
  * 
  */
package lotos;

import lotos.*; 

/**
 * 
 * A simple MAC layer
 * 
 * revised by Guangyu Dong, Jan 6, 2004
 * to adapt to the change of MACInterface
 */

class SimpleMAC implements MACInterface
{
	public final int MAC_QUEUE_SIZE = 10;
	public final int MAX_BACKOFF_STEPS = 16;

	OverlayNode node;
	SimEventMsg msgQueue[];
	int queueHead, queueTail;
	int backoffStep;
	int backoffInterval;

	public SimpleMAC(OverlayNode olNode)
	{
		msgQueue = new SimEventMsg[MAC_QUEUE_SIZE];
		node = olNode;
		backoffStep = 1;
		backoffInterval = 0;
		queueHead = 0;
		queueTail = 0;
	}

	public boolean sendMsg(SimEventMsg msg)
	{
		if(insertIntoMsgQueue(msg))
		{
			return true;
		}
		else
		{
			return false;
		}
	}

	public boolean insertIntoMsgQueue(SimEventMsg msg)
	{
		if((queueTail+1)%MAC_QUEUE_SIZE == queueHead) /* Queue Full */
			return false;
		else
		{
			msgQueue[queueTail] = msg;
			queueTail = (queueTail+1)%MAC_QUEUE_SIZE;
			return true;
		}
	}

	public void sendMessages()
	{
		if((backoffInterval == 0) && (queueHead != queueTail) )
		{
			if(node.net.phyLayer.sendMsg(msgQueue[queueHead]))
			{
				queueHead = (queueHead+1)%MAC_QUEUE_SIZE;
				backoffStep = 1;
			}
			else
			{
				if(backoffStep < MAX_BACKOFF_STEPS)
					backoffStep *= 2;
				backoffInterval = backoffStep + (int)(Math.random()*backoffStep);
			}
		}
	}
	
	/* init MAC layer */
	public void MACInit(){
		
	}
	
	public void MACStep()
	{
		sendMessages();
		if(backoffInterval != 0)
			backoffInterval--;
	}
	/* added by Guangyu
	 */
	public void handleMACTimer(SimEventMAC mac_event){
		//no timer to handle	
	}
	
	public void recordMessage(long ms_start_time, long ms_end_time){
		
	}
	public void senseMessage(long ms_start_time, long ms_end_time){
		
	}
	public int getHeaderSize(){
		return 0;	
	}
}

