/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team
  * 
  */
package lotos;

import lotos.*; 
import java.lang.*;
import java.util.*;

/**
 *  This class is used for statistic collection. Three performance metrics,
 *  RDP, Stress & Degree, are collected. In addition, an extention to stress,
 *  proportional stress, is also measured.
 */


class Stats{

    /**
     * a reference to the overlay network
     */
    OverlayNetwork onetwork;
    /**
     * all the nodes in the overlay
     */
    Vector active_nodes;

    boolean pStress = false;

    /**
     * Constructor
     * @param onet the overlay network to be evaluated
     */
    Stats( OverlayNetwork onet ){
        onetwork = onet;
        active_nodes = onet.getActiveNodes();
    }

    /** measure PStress */
    public void setPStress() { pStress = true; }
    /** measure Stress */
    public void unsetPStress()  { pStress = false; }

    /** @return an stress array which contains the stresses of
     *  all the delivery trees rooted at each overlay node
     */

   float[] reportStress(){
/*        if( 0 == active_nodes.size() ) return null;

        lotusDebug.println("stress");
        float [] stress = stressArray();
        float [] data = new float[3];

        //stress
        Arrays.sort(stress);
        data[0] = averageFloat( stress );
        data[1] = percentileFloat( stress, 0.9f);
        data[2] = percentileFloat( stress, 0.99f);
        return data;*/
        return null;
    }

    /**
     * @return an RDP array that records all the pair-wise delays of
     * all the overlay nodes
     */
    float[] reportRDP(){
        if( 0 == active_nodes.size() ) return null;
				
        lotusDebug.println("rdp");
        float [] rdp = rdpArray();

        if( rdp == null ) {
            lotusDebug.errorln("rdp can not be calculated, possibily because physical in-routability" );
            return null;
        }

        float [] data = new float[3];
        lotusDebug.println("report rdp");
        //rdp
        Arrays.sort(rdp);
        data[0] = averageFloat( rdp );
        data[1] = percentileFloat( rdp, 0.9f );
        data[2] = percentileFloat( rdp, 0.99f);
        return data;
    }

    /**
     * @return all the performance metric are returned in a float array,
     * including: </p>
     * RDP ( average, 90%, 99% ) </p>
     * Stress ( average, 90%, 99% ) </p>
     * PStress ( average, 90%, 99% ) </p>
     * Degree ( average, 90%, 99% )
     */
    float[] reportFloat(){
        if( 0 == active_nodes.size() ) return null;


        lotusDebug.println("degree");
        int [] degree = degreeArray();
		
		/** modified by Guangyu to enable calculating rdp in ad hoc network */
		boolean net_fwd=onetwork.forwarding;
        onetwork.forwarding=true;
        onetwork.buildPhysicalRoute();
        lotusDebug.println("rdp");
        float [] rdp = rdpArray();
        onetwork.forwarding=net_fwd;
        
        
        lotusDebug.println("stress");
        float [][] stress = stressArray();
        
		/** modified by Guangyu to adapt to ad hoc network */
        if( rdp == null ) {
            lotusDebug.errorln("rdp can not be calculated, possibily because physical in-routability" );
            //return null;
        }

        float [] data = new float[12];
        lotusDebug.println("report degree");
        //degree
        Arrays.sort(degree);
        data[0] = averageInt( degree );
        data[1] = percentileInt( degree, 0.9f);
        data[2] = percentileInt( degree, 0.99f);
		
		if(rdp!=null){
		    lotusDebug.println("report rdp");
	        //rdp
	        Arrays.sort(rdp);
	        data[3] = averageFloat( rdp );
	        data[4] = percentileFloat( rdp, 0.9f );
	        data[5] = percentileFloat( rdp, 0.99f);
    	}

        //stress
        lotusDebug.println("report stress");
        Arrays.sort(stress[0]);
        Arrays.sort(stress[1]);
		data[6] = averageFloat( stress[0] );
        data[7] = percentileFloat( stress[0], 0.9f);
        data[8] = percentileFloat( stress[0], 0.99f);

        //proportional stress
        lotusDebug.println("report proportional stress");
        data[9] = averageFloat( stress[1] );
        data[10] = percentileFloat( stress[1], 0.9f);
        data[11] = percentileFloat( stress[1], 0.99f);
        return data;
    }


    /**
     * @return a degree array that records the number of logical edges
     * of each overlay node
     */
    private int[] degreeArray(){
        int[] degree = new int[active_nodes.size()];
        for( int i=0; i<active_nodes.size(); i++){
            degree[i] = ((OverlayNode)(active_nodes.elementAt(i))).getLogicalNeighborNumber();
        }
        return degree;
    }

    /**@return a two dimensional stress array
     * stress[0][] is an array that contains the stress of all nodes
     * stress[1][] is an array that contains the pstress of all nodes
     */
    private float[][] stressArray(){
        float[][] stress = new float[2][active_nodes.size()];
        float[] nStress;

        for( int i=0; i<active_nodes.size(); i++){
            nStress = nodeStress( (OverlayNode) active_nodes.elementAt(i));
            stress[0][i] = nStress[0];
            stress[1][i] = nStress[1];
        }

        return stress;
    }

    /**
     * measures the stress of a delivery tree
     * @param  the root of the delivery tree
     * @return 1st element is the stress and 2rd element is the pstress
     */
    private float[] nodeStress( OverlayNode root ){

        Vector phyNodes = onetwork.getPhyNodes();
        OverlayNode [] nodequeue = new OverlayNode [active_nodes.size()];
        int head, tail;

        head=1;
        tail=0;
        nodequeue[0] = root;

        int [][]edgestress= new int[phyNodes.size()][];

        //lotusDebut.println("allocate buffer");

        for( int i=0; i<phyNodes.size(); i++){
            OverlayNode eachnode = (OverlayNode) phyNodes.elementAt(i);
            edgestress[i] = new int[eachnode.getPhyNeighborNumber()];
            for( int j=0; j<eachnode.getPhyNeighborNumber(); j++) edgestress[i][j]=0;
        }

        //lotusDebut.println("traverse the tree");
        while( head != tail ){

            OverlayNode parent = nodequeue[tail++];

            Enumeration e = parent.getLogicalNeighbor();

            while( e.hasMoreElements() ){

            OverlayNode child = (OverlayNode)e.nextElement();
            //test if the node is on the parents of some of its neighbor
            if( parent == onetwork.getLogicalNextHop( child, root )){
                //lotusDebut.println("children" + child.getAddress().toString() + "|" );
                //if it is, then we push it into the queue first
                nodequeue[head++] = child;
                //calculate the stress generated by parent & children

                OverlayNode nextHop = parent.getPhyNextHop( child);
                OverlayNode prevHop = parent;

                while( nextHop != child ){
                //lotusDebut.println( "From:" + prevHop.getAddress().toString() +
                //"To:" + nextHop.getAddress().toString() );
				
                edgestress[phyNodes.indexOf(prevHop)][prevHop.getPhyNeighborIndex(nextHop)]++;
                prevHop = nextHop;
                nextHop = nextHop.getPhyNextHop(child);
                }
            //lotusDebut.println( "From:" + prevHop.getAddress().toString() +
            //		"To:" + nextHop.getAddress().toString() );
                edgestress[phyNodes.indexOf(prevHop)][prevHop.getPhyNeighborIndex(nextHop)]++;

              //  lotusDebut.println();
            }
            }
        }

        //lotusDebut.println("Search for the max stress at the node");

        //find the maximum stress
        double totalstress = 0;
        double edgenum = 0;

        double overlayLen = 0;
        double physicalLen = 0;

        for( int i=0; i<edgestress.length; i++){
            //lotusDebut.println("node at" + i );
            OverlayNode ni = (OverlayNode)phyNodes.elementAt(i);
            for( int j=0; j < edgestress[i].length; j++){
                //lotusDebut.println("array length" + edgestress[i].length);
                //lotusDebut.println("node at" + j );
                if( 0 != edgestress[i][j] ) {
                    totalstress += edgestress[i][j];
                    edgenum++;
                    physicalLen += ni.getPhyEdgeLen(j);
                    overlayLen += ni.getPhyEdgeLen(j)*edgestress[i][j];
                }
            }
        }

        //lotusDebut.println( "stress at node" + root.getAddress().toString() + "is" + maxstress );
        float [] retVal = new float[2];
        retVal[0] = ((float)totalstress)/((float)edgenum);
        retVal[1] = (float)(overlayLen/physicalLen);
        return retVal;
    }

    /*private int[] stressArray(){
        int[] stress = new int[active_nodes.size()];

        for( int i=0; i<active_nodes.size(); i++)
            stress[i] = nodeStress( (OverlayNode) active_nodes.elementAt(i));

        return stress;
    }

    private int nodeStress( OverlayNode root ){

        Vector phyNodes = onetwork.getPhyNodes();
        OverlayNode [] nodequeue = new OverlayNode [active_nodes.size()];
        int head, tail;

        head=1;
        tail=0;
        nodequeue[0] = root;

        int [][]edgestress= new int[phyNodes.size()][];

        //lotusDebut.println("allocate buffer");

        for( int i=0; i<phyNodes.size(); i++){
            OverlayNode eachnode = (OverlayNode) phyNodes.elementAt(i);
            edgestress[i] = new int[eachnode.getPhyNeighborNumber()];
            for( int j=0; j<eachnode.getPhyNeighborNumber(); j++) edgestress[i][j]=0;
        }

        //lotusDebut.println("traverse the tree");
        while( head != tail ){

            OverlayNode parent = nodequeue[tail++];

            Enumeration e = parent.getLogicalNeighbor();

            while( e.hasMoreElements() ){

            OverlayNode child = (OverlayNode)e.nextElement();
            //test if the node is on the parents of some of its neighbor
            if( parent == onetwork.getLogicalNextHop( child, root )){
                //lotusDebut.println("children" + child.getAddress().toString() + "|" );
                //if it is, then we push it into the queue first
                nodequeue[head++] = child;
                //calculate the stress generated by parent & children

                OverlayNode nextHop = parent.getPhyNextHop( child);
                OverlayNode prevHop = parent;

                while( nextHop != child ){
                //lotusDebut.println( "From:" + prevHop.getAddress().toString() +
                //"To:" + nextHop.getAddress().toString() );

                edgestress[phyNodes.indexOf(prevHop)][prevHop.getPhyNeighborIndex(nextHop)]++;
                prevHop = nextHop;
                nextHop = nextHop.getPhyNextHop(child);
                }
            //lotusDebut.println( "From:" + prevHop.getAddress().toString() +
            //		"To:" + nextHop.getAddress().toString() );
                edgestress[phyNodes.indexOf(prevHop)][prevHop.getPhyNeighborIndex(nextHop)]++;

              //  lotusDebut.println();
            }
            }
        }

        //lotusDebut.println("Search for the max stress at the node");

        //find the maximum stress
        int maxstress = 1;
        for( int i=0; i<edgestress.length; i++){
            //lotusDebut.println("node at" + i );
            for( int j=0; j < edgestress[i].length; j++){
            //lotusDebut.println("array length" + edgestress[i].length);
            //lotusDebut.println("node at" + j );
            if( maxstress < edgestress[i][j] ) maxstress = edgestress[i][j];
            }
        }

        //lotusDebut.println( "stress at node" + root.getAddress().toString() + "is" + maxstress );
        return maxstress;
    }
    */

    float [][]oDist;

    /**
     * @param src source node
     * @param dst destination node
     * @return RDP between src & dst
     */
    private float rdpCalc( OverlayNode src, OverlayNode dst){
        if( src == dst ) return 1;
        float logicalDist = 0;

        OverlayNode nextHop = onetwork.getLogicalNextHop( src, dst );
        OverlayNode prevHop = src;
        int prevI;
        int nextI;
        while( nextHop != dst ){
            if( null == nextHop ){
            	lotusDebug.println("next hop info unavailable");
            	break;
            }
            prevI = active_nodes.indexOf(prevHop);
            nextI = active_nodes.indexOf(nextHop);
            logicalDist += oDist[prevI][nextI];
            prevHop = nextHop;
            nextHop = onetwork.getLogicalNextHop(nextHop, dst);
        }
        prevI = active_nodes.indexOf(prevHop);
        nextI = active_nodes.indexOf(nextHop);
        logicalDist += oDist[prevI][nextI];

        prevI = active_nodes.indexOf(src);
        nextI = active_nodes.indexOf(dst);
        float phyDist = oDist[prevI][nextI];
        return logicalDist/phyDist;
    }

    /**
     * @return an RDP array that records all the pair-wise RDP of all
     * the overlay nodes
     */
    private float[] rdpArray(){

        int len = (active_nodes.size()) * (active_nodes.size()-1)/2;
        float[] rdp= new float[len];
        oDist = new float[active_nodes.size()][active_nodes.size()];

        for( int i=0; i<active_nodes.size(); i++){
            OverlayNode src = (OverlayNode) active_nodes.elementAt(i);
            for( int j=0; j<i; j++) {
                OverlayNode dst = (OverlayNode) active_nodes.elementAt(j);
                //lotusDebug.println("get distance from "+i+" to "+j);
                oDist[i][j] = src.getDistanceTo(dst);
                oDist[j][i] = oDist[i][j];
                if( oDist[i][j] < 0 ) return null;
            }
            oDist[i][i] = 0;
        }
        int index = 0;
        for( int i=0; i<active_nodes.size(); i++){
            for( int j=i+1; j<active_nodes.size(); j++){
            //lotusDebug.println("rdp from node:" + i + " to " + j );
            rdp[index++] = rdpCalc( (OverlayNode) active_nodes.elementAt(i),
                          (OverlayNode) active_nodes.elementAt(j) );
            //lotusDebug.println("done");
            //if( rdp[index] < 0 ) return null;
            }
        }
        oDist=null;
        return rdp;
    }


    /*float averageRDP() {
	float [] rdp = rdpArray();
	float sum=0;
	for( int i=0; i<rdp.length; i++)
	    sum += rdp[i];

	return sum/rdp.length;
    }

    float averageStress(){
	int [] stress = stressArray();

	float sum=0;
	for( int i=0; i<stress.length; i++)
	    sum += stress[i];

	return sum/stress.length;
    }

    float averageDegree(){
	int [] degree = degreeArray();

	float sum=0;
	for( int i=0; i<degree.length; i++)
	    sum += degree[i];

	return sum/degree.length;
    }

    float maxRDP(){
	float [] rdp = rdpArray();

	float max = 1;
	for( int i=0; i<rdp.length; i++){
	    if( rdp[i] > max ) max = rdp[i];
	}

	return max;
    }

    int maxStress(){
	int [] stress = stressArray();

	int max = 1;
	for( int i=0; i<stress.length; i++){
	    if( stress[i] > max ) max = stress[i];
	}
	return max;
    }

    int maxDegree(){

	int [] degree = degreeArray();
	int max = 0;

	for( int i=0; i<degree.length; i++){
	    if( degree[i] > max ) max = degree[i];
	}

	return max;
    }

    static void sortInt( int []input, float percent ){
        int tmp;
        int sortLen = input.length - (int)(input.length * percent) + 1;
        int compare;
        for( int i=0; i<sortLen; i++){
            compare = input.length-i-1;
            for( int j=0; j<compare; j++){
                if( input[j] > input[compare] ) {
                    tmp = input[compare];
                    input[compare] = input[j];
                    input[j] = tmp;
                }
            }
        }
    }

    static void sortFloat( float []input, float percent){
        float tmp;
        int sortLen = input.length - (int)(input.length * percent) + 1;
        int compare;
        for( int i=0; i<sortLen; i++){
            compare = input.length-i-1;
            for( int j=0; j<compare; j++){
                if( input[j] > input[compare] ) {
                    tmp = input[compare];
                    input[compare] = input[j];
                    input[j] = tmp;
                }
            }
        }
    }
*/
    /**
     * @param input a sorted integer array
     * @param percent a percentile value
     * @return the index of the element in the array that has the
     * percentile value
     */
    static int percentileInt( int [] input, float percent){
        //sortInt(input, percent);
        int index = (int)(input.length*percent);
        return input[index];
    }

    /**
     * @param input a sorted float array
     * @param percent a percentile value
     * @return the index of the element in the array that has the
     * percentile value
     */
    static float percentileFloat( float [] input, float percent ){
        //sortFloat(input, percent);
        int index = (int)(input.length*percent);
        return input[index];
    }

    /**
     * @return the average of an integer array
     */
    static float averageInt( int[] input ){
        double sum = 0;
        for( int i=0; i<input.length; i++) sum+=input[i];
        return (float)(sum/input.length);
    }
    /**
     * @return the average of a float array
     */
    static float averageFloat( float[]input){
    	double sum=0;
        for( int i=0; i<input.length; i++)sum+=input[i]; 
        return (float)(sum/input.length);
    }
/*
    float []statReport(){
        lotusDebug.println("degree");
        int [] degree = degreeArray();
        lotusDebug.println("rdp");
        float [] rdp = rdpArray();
        lotusDebug.println("stress");
        int [] stress = stressArray();

        float [] data = new float[9];
        lotusDebug.println("report degree");
        //degree
        Arrays.sort(degree);
        data[0] = averageInt( degree );
        data[1] = percentileInt( degree, 0.9f);
        data[2] = percentileInt( degree, 0.99f);

        lotusDebug.println("report rdp");
        //rdp
        Arrays.sort(rdp);
        data[3] = averageFloat( rdp );
        data[4] = percentileFloat( rdp, 0.9f );
        data[5] = percentileFloat( rdp, 0.99f);

        //stress
        Arrays.sort(stress);
        data[6] = averageInt( stress );
        data[7] = percentileInt( stress, 0.9f);
        data[8] = percentileInt( stress, 0.99f);
        return data;
    }

    void report(){
	lotusDebug.println( "Average Degree:" + averageDegree() );
	lotusDebug.println( "Average Stress:" + averageStress() );
	lotusDebug.println( "Average RDP:" + averageRDP() );
	lotusDebug.println( "Max Degree:" + maxDegree() );
	lotusDebug.println( "Max Stress:" + maxStress() );
	lotusDebug.println( "Max RDP:" + maxRDP() );
    }

    String reportString(){
        return
        	"Average Degree:" + averageDegree() + "\n" +
            "Average Stress:" + averageStress() + "\n" +
            "Average RDP:" + averageRDP() + "\n" +
            "Max Degree:" + maxDegree() + "\n" +
            "Max Stress:" + maxStress() + "\n" +
            "Max RDP:" + maxRDP();
    }*/

}
