/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, July 2002 
  * 
  */
package lotos;

import lotos.*; 
import java.io.*;
import java.util.*;
import java.lang.*;

/**
 * An abstract that abstracts the shared elements of all topology descriptions.
 */
abstract class TopoAdaptor{

    /** file name suffix: home-made format */
    static String suffix_HM = ".hm";
    /** file name suffix: GT_ITM */
    static String suffic_GT = ".alt";
    /** file name suffix: landmark */
    static String suffic_LM = ".lm";
    /** file name suffix: INET */
    static String suffic_IT = ".it";

    /** physical node array */
    PhyNode[] nodes;
    /** physical edge array */
    PhyEdge[] edges;
    //Vector activeNodes;

    PhyNode[] getNodes( ){ return nodes;}
    PhyEdge[] getEdges( ){ return edges;}
    //Vector getActiveNodes() { return activeNodes;}

    TopoAdaptor(){ //activeNodes = new Vector();
    }

    /**
     * Abstract function, implenented by each topology adaptor to
     * load its topology description file.
     * @param filename the description file to be loaded
     * @return true if loading is successful, false otherwise
     */
    abstract boolean loadFile( String filename );

    /**
     * Save the file to the default description format( home-made )
     * This function can be overiden by subclasses to save a topology
     * description to a specific format.
     * @param filename the file to save the topology description
     * @return true if the file is saved successfully, false otherwise
     */
    boolean saveFile( String filename ){
	//lotusDebut.println("in save to File");
	//lotusDebut.println("filename" + filename);

        try {
            FileWriter output = new FileWriter(filename);
            BufferedWriter bw_output = new BufferedWriter(output);
            PrintWriter pw_output = new PrintWriter(bw_output);

            saveToStream( pw_output  );

	    //lotusDebut.println("before close");
            pw_output.close();  // I assume this closes output and bw_output
	    //lotusDebut.println("after close");

        }
        catch (Exception e){
	        lotusDebug.println("error when writing the file");
	        return false;
        }
        return true;
    }

    private void saveToStream( PrintWriter pw )
    {
	//lotusDebut.println("in save to stream");
        pw.println("#Lotus save file");

	//lotusDebut.println("print the title");


        // Write out all the nodes.
	pw.print( nodes.length );
	pw.print( " " );
	if( null == edges ) pw.print( 0 );
    else pw.print( edges.length );
	pw.println();

	int index = 0;


        while (index < nodes.length)
        {

		    PhyNode p = nodes[index];
		    pw.print(p.GetX());
            pw.print(" ");
            pw.print(p.GetY());
            
            pw.print(" ");
            pw.print(p.getMobilityNum());
            
            pw.print(" ");
            pw.print(p.speed);
            
            pw.print(" ");
            pw.print(p.wired?"Y":"N");
            
            pw.print(" ");
            pw.print(p.wireless?"Y":"N");
            
            /* protocol setting*/
            /*
            protocolInterface protocol=((OverlayNode)p).getProtocol();
            if(protocol!=null){
            	pw.print(" ");
            	pw.print(protocol.getConfigString());
            };
            */
            
            /*
            if (activeNodes.contains(p))
                pw.print(" Y ");
            else
                pw.print(" N ");
            */
            pw.println();
    	    index++;

	    //lotusDebut.println("save node" + p.x + ":" + p.y );
        }

        // Write out all the connections
        index = 0;
        if( null == edges ) return;
        while (index < edges.length)
        {
	    PhyEdge p = edges[index];
	    pw.print(p.srcNode);
            pw.print(" ");
            pw.print(p.dstNode);
	    pw.print(" ");
	    pw.print(p.distance);
            pw.println();

	    index++;

	    //lotusDebut.println("Save edges" + p.srcNode + ":" + p.dstNode );
        }

	//lotusDebut.println("end of save");

    }



}

