/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, July 2002 
  * 
  */
package lotos;

import lotos.*; 
import java.util.*;
import java.lang.*;
import java.io.*;

/**
 * Topology adaptor for home-made description
 */
class TopoAdaptor_HM extends TopoAdaptor{

    TopoAdaptor_HM( String filename ){
		loadFile(filename);
    }

    TopoAdaptor_HM( ){}

    boolean loadFile( String filename ){
        try
        {
            FileReader input = new FileReader(filename);
            BufferedReader br_input = new BufferedReader(input);
            processLoadStream( br_input  );
            br_input.close();  // I assume this closes input.
        }
        catch (Exception e)
        {
            lotusDebug.println("exception at TopoAdaptor_HM" + e );
            return false;
        }
	    return true;

    }

    private void processLoadStream(BufferedReader br) throws Exception
    {
		String line="#";
		
		while(line.startsWith("#") || line.equals(""))line = br.readLine().trim();
		
		StringTokenizer token = new StringTokenizer(line);
	
		int numnodes = (Integer.valueOf( token.nextToken() )).intValue();
		int numedges = (Integer.valueOf( token.nextToken() )).intValue();
		
		//lotusDebug.println(new Integer(numnodes).toString());
		
		nodes = new PhyNode[numnodes];
		edges = new PhyEdge[numedges];
	
		int i;
		for( i=0; i<numnodes; i++){
	
		    line = br.readLine().trim();
		    
		    if(line.startsWith("#"))continue;
	
		    token = new StringTokenizer(line);
	
		    //token.nextToken();
		float speed = 0;
		    float xc = (Float.valueOf( token.nextToken() )).floatValue();
		    float yc = (Float.valueOf( token.nextToken() )).floatValue();
		    int mobilityMode = (Integer.valueOf( token.nextToken() )).intValue();
		    speed =	(Float.valueOf( token.nextToken() )).floatValue();
			
			String wr=token.nextToken().trim().toUpperCase();
			String wl=token.nextToken().trim().toUpperCase();
			
		    nodes[i] = new PhyNode(xc,yc, 
		    					   PhyNode.getMobilityStr(mobilityMode), 
		    					   speed, wr.startsWith("Y"), wl.startsWith("Y")
		    					   );
		    
		    //lotusDebug.println( "Add nodes:" + xc + "," + yc );
		    /*
		    if( token.hasMoreTokens() ) {
				String tag = (token.nextToken()).trim();
				if( tag.equals( "Y" ) ){
				    activeNodes.addElement(nodes[i]);
				}
		    }*/
		}
		for( i=0; i<numedges; i++){
		    line = br.readLine();
		    lotusDebug.println(line);
	
		    token = new StringTokenizer(line);
	
		    int from = (Integer.valueOf( token.nextToken() )).intValue();
		    int to = (Integer.valueOf( token.nextToken() )).intValue();
		    float dist = (Float.valueOf( token.nextToken() )).floatValue();
	
		    lotusDebug.println( "Add edges:" + from + "," + to );
	
		    edges[i] = new PhyEdge(from, to, dist);
	
		}

	
    }

}
