/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, 2002 
  * 
  */
package lotos;

import lotos.*; 
import java.util.*;
/*
 * A wireless physical layer implementation
 * 
 * @author Guangyu Dong, Jan 6, 2004
 */
class WirelessPhysicalLayer implements PhysicalLayer
{
   /* Packet Speed in m/s */
   // final static float PACKET_SPEED = 1000000f/8;
	final static float PACKET_SPEED = 1000f/8;
   /* Transmission Range in m */
   final static float SIGNAL_RANGE = 100f;
   /* Signal Sensing Range in m */
   final static float SENSE_RANGE = 200f;
   
   /* the network */
   OverlayNetwork net;
   
	/*
	 * Constructor
	 */
	WirelessPhysicalLayer(OverlayNetwork olNet)
	{
		net = olNet;
	}
	/*
	 * @see PhysicalLayer#sendMsg(SimEventMsg)
	 */
	public boolean sendMsg(SimEventMsg msg)
	{
		long time_start = net.getMSSimTime();
		long time_end = time_start + (long)((msg.getSize()+getPreamble())*SimEvent.TIME_PER_SEC/PACKET_SPEED);
		OverlayNode srcNode = msg.getSender();

		
		/* Detect collisions with previous transmissions from neighbors */
		{
			Vector neighbors = srcNode.getWirelessNeighbors();
			if(neighbors!=null){
	   	   		for(int i=0;i<neighbors.size();i++) {
		      	    OverlayNode onode = (OverlayNode) neighbors.elementAt(i);
					onode.wl_MAC.recordMessage(time_start, time_end);
		      	}
		    }
		    neighbors = srcNode.getSensingNeighbors();
			if(neighbors!=null){
	   	   		for(int i=0;i<neighbors.size();i++) {
		      	    OverlayNode onode = (OverlayNode) neighbors.elementAt(i);
					onode.wl_MAC.senseMessage(time_start, time_end);
		      	}
		    }
		
         	msg.setMSTime( time_end );
         	net.insertEvent(msg);
         	
         	
         	srcNode.wl_MAC.recordMessage(time_start, time_end);
         	
			
			return true;
		}
		
	}
	/*
	 * @return the preamble in Bytes
	 */
	int getPreamble(){
		return 0;	
	}
}
