/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, 2002
  * 
  */
package lotos;

import lotos.*; 
import java.io.*;
import java.util.*;

/**
 * Convert the data file from AMP project to standard DT Precise overlay topology
 * description file
 */

class ampGen {
    final static int numOfHost = 92;
    final static int numOfLandmarks = 7;
    final static String formatTag = "dtprecise";

    private static double [][] removeGarbage( double [][]distance){

        boolean existGarbage = false;
        for( int i=0; i<distance.length; i++)
            for( int j=0; j<distance.length; j++) {
                if( distance[i][j] < 0 ) existGarbage = true;
            }

        if( false == existGarbage ) return distance;

        //search row by row and then column by column
        int hostIndex = -1;
        int hostGarbage = -1;

        for( int i = 0; i < distance.length; i++){
            int garbage = 0;
            for( int j = 0; j<distance.length; j++) {
                if( distance[i][j] < 0 ) garbage++;
            }
            if( garbage > hostGarbage ) {
                hostGarbage = garbage;
                hostIndex = i;
            }
        }

        for( int i = 0; i < distance.length; i++){
            int garbage = 0;
            for( int j = 0; j<distance.length; j++) {
                if( distance[j][i] < 0 ) garbage++;
            }
            if( garbage > hostGarbage ) {
                hostGarbage = garbage;
                hostIndex = i;
            }
        }

        double [][] newDist = new double [distance.length-1][distance.length-1];
        int [] map = new int[distance.length-1];
        for( int i=0; i<=map.length; i++){
            if( i<hostIndex ) map[i]=i;
            else if( i>hostIndex ) map[i-1] = i;
        }

        for( int i=0; i<map.length; i++ )
            for( int j=0; j<map.length; j++){
                newDist[i][j] = distance[map[i]][map[j]];
            }

        return newDist;
    }

    public static void main(String argv[]) throws Exception{
        double [][] distance = new double[numOfHost][numOfHost];

        //read the input to the array
        FileReader input = new FileReader("ampreport.txt");
        BufferedReader br = new BufferedReader(input);
        String line;

        for( int i=0; i<numOfHost; i++ ) {
	        line = br.readLine();
        	StringTokenizer token = new StringTokenizer(line);
            token.nextToken();
            for( int j=0; j<numOfHost; j++){
                double delay;
                try{
                    delay = Double.valueOf(token.nextToken()).doubleValue();
                }catch( NumberFormatException e ){
                    delay = -1;
                }
                distance[i][j] = delay;
            }
        }

        br.close();

        //fill out those unavailable using symmetric
        for( int i=0; i<numOfHost; i++){
            for( int j=0; j<numOfHost; j++){
                if( distance[i][j] < 0 ) {
                    if( distance[j][i] >= 0 ) distance[i][j] = distance[j][i];
                }
            }
        }

        //iterate to kick out those unsable data
        int hostnum;
        do{
            hostnum = distance.length;
            distance = removeGarbage( distance );
        }while( hostnum != distance.length );

        lotusDebug.println( "usable number:" + distance.length );

        //take the average of bidirection measurement
        for( int i=0; i<distance.length; i++)
            for( int j=0; j<i; j++){
                distance[i][j] = (distance[i][j]+distance[j][i])/2;
                distance[j][i] = distance[i][j];
            }

        //generate the landmarks & overlaynodes in 7 files
        //record of the landmarks
        boolean []landmarks = new boolean[distance.length];
        boolean []activenodes = new boolean[distance.length];
        for( int i=0; i<distance.length; i++) landmarks[i]=false;

        //selection of the landmarks
        for( int i=0; i<numOfLandmarks; i++){
            while(true){
                int index = (int)(Math.random() * distance.length );
                if( landmarks[index] ) continue;
                else {landmarks[index] = true; break; }
            }
        }

        //generate the file with active_nodes from 10 to 70
        for( int i=10; i<=70; i+=10 ){
            //active_nodes selection
            for( int j=0; j<distance.length; j++) activenodes[j]=false;
            int activeNum = i;
            int index = 0;
            while( activeNum > 0 ){
                if( landmarks[index] == false && activenodes[index] == false ) {
                    if( Math.random() < 0.01*i ) {
                        activenodes[index] = true;
                        activeNum--;
                    }
                }
                index = (index+1)%distance.length;
            }

            //create the file
            String filename = "amp"+i+".txt";
            FileWriter output = new FileWriter(filename);
            BufferedWriter bw_output = new BufferedWriter(output);
            PrintWriter pw = new PrintWriter(bw_output);
            pw.println(formatTag);

            int numedges=(distance.length*(distance.length-1))/2;

        	pw.print( distance.length );
        	pw.print( " " );
            pw.print( numedges );
	        pw.println();

	        index = 0;
            while (index < distance.length)
            {
	            pw.print((int)(1000*Math.random()));
                pw.print(" ");
                pw.print((int)(1000*Math.random()));
                if( landmarks[index] ) pw.print(" L");
                if( activenodes[index] ) pw.print(" Y");
                pw.println();
        	    index++;
            }

            for( int j=0; j<distance.length; j++)
                for( int k=0; k<j; k++) {
                    pw.print( j );
                    pw.print(" ");
                    pw.print( k );
                    pw.print(" ");
                    pw.print(((float)((int)(distance[j][k]*100)))/100);
                    pw.println();
                }
            pw.close();
        }

    }
}
