/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, 2002 
  * 
  */
package lotos;

import lotos.*; 
import java.lang.*;
import java.util.*;
import java.io.*;

/**
 * This class implements the GNP algorithm, encapsulates the details of
 * the algorithm and provides simple interface for the outside
 */

class landmark{

  /**total landmarks used*/
  int numLandmarks;
  /**the coordinate dimensions*/
  int dimensions;

  /**landmark coordinates    dim = numLandmarks * dimensions*/
  float [][]coordinates;
  /**delays of the landmarks dim = numLandmarks * numLandmarks */
  float [][]lmDelays;

  /** target coorndinates     dim = dimensions */
  float []targetCoors;
  /** delays between the landmarks and the target dim=numLandmarks */
  float []targetDelays;
  /** only those delays that are valid are used */
  boolean []validDelays;

  /*float []hostCoors; //the two hosts coordiantes
  float []hostDelay; //the two hosts */

  //static configuration parameters
  static int mytry=3;
  static int range=500;
  static int lambda=2000;
  static int mytry2=3;
  static int restarts=1;
  static float ftol=0.00001f;
  static float MAX_ERR=1e20f;
  static float TINY=1.0e-10f;
  static int NMAX=500000;
  static int NR_END = 1;

  /** whether the starting point of the optimization problem is randomly
   *  selected or not
   */
  boolean randomInit = true; //false;
  /*static final int objTarget = 0;
  static final int objLandmark = 1;
  static final int objTwohost = 2;
*/

    /**
     * Constructor
     * @param delay pair-wise delays between the landmarks
     * @param dim the dimension of the coordinate
     */
  landmark( float [][]delay, int dim ){

    dimensions = dim;
    lmDelays = delay;
    numLandmarks = lmDelays.length;
    coordinates = new float[numLandmarks][dim];
    targetCoors = new float[dimensions];

    lotusDebug.println("dimension:"+dim + "  " + "probes:" + numLandmarks );
    //estabish the coordinates of the landmarks
    fitModel();
  }

  /**
   * use randomly selected starting point
   */
  void randomInitiation(){
    randomInit = true;
  }

  /**
   * Use origin as the fixed starting point
   */
  void fixInitiation(){
    randomInit = false;
  }


   /** return the coordinates of the landmarks */
  float [][] getLMCoordinates() { return coordinates; }

  /** solve the optimization problem to obtain the logical coordinates
   *  of the landmarks
   */
  void fitModel() {

    float [] model = new float[ numLandmarks * dimensions ];
    lotusDebug.println("solve the model coordinates");
    float error = solve(model, true, mytry);

    if (error >= MAX_ERR) {
      lotusDebug.println("error too large");
    } else {
	//lotusDebug.println("error within bound");
    }

    for( int i=0; i<numLandmarks; i++){
      for( int j=0; j<dimensions; j++){
	coordinates[i][j]= model[i*dimensions+j];
      }
    }
  }

  //solve the coordinates for a single target
  /*float[] fitTargetData( float [] tDelays ) {

      if( tDelays.length != numLandmarks ) {
	  lotusDebug.println(" measurement dimenstion does not match");
	  return null;
      }

      targetDelays = tDelays;

      float [] xy = new float[dimensions];

      float error = solve( xy, false, mytry2 );

      if (error >= MAX_ERR) {
    	  lotusDebug.println("error too large");
      } else {
          //lotusDebug.println("error within bound");
      }

      //copy the result back to the class
      for( int i=0; i<dimensions; i++) targetCoors[i]=xy[i];
      return targetCoors;
  }*/

  /** solve the optimization to obtain a target's logical coordinates */
  float[] fitTargetData( float [] tDelays, boolean [] index ) {

      if( tDelays.length != numLandmarks ) {
          lotusDebug.println(" measurement dimenstion does not match");
          return null;
      }

      targetDelays = tDelays;
      validDelays = index;

      float [] xy = new float[dimensions];

      float error = solve( xy, false, mytry2 );

      if (error >= MAX_ERR) {
	    lotusDebug.println("error too large");
      } else {
	  //lotusDebug.println("error within bound");
      }

      //copy the result back to the class
      for( int i=0; i<dimensions; i++) targetCoors[i]=xy[i];

      return targetCoors;
  }

  /**
   * the solver of the optimization problem.
   * @param solution    the return array of the optimization problem
   * @param lmObjective which objective function to use, either landmark's or the target's
   * @param mytry       the times to repeat the optimization problem
   *
   */
  float solve(float []solution, boolean lmObjective, int mytry) {
    float min, fit, localftol;
    float []myxy;
    float [][]p;
    float []y;
    int i, j, restarted, num;

    num=solution.length;
    localftol = ftol;
    myxy = new float[num+1];
    p = new float[num+1+1][]; //hold the num+1 initial points

    p[1] = myxy; // the first point is from  myxy, initialized randomly
    for (i=2;i<=num+1;i++) {
      p[i]= new float[num+1];
    }
    y = new float[num+1+1]; //the corresponding function value

    //lotusDebug.println(" start evaluate");
    while(true){
        min = MAX_ERR;
        for (i=0; i<mytry; i++) {
        	restarted = 0;
            for (j=1; j<=num; j++) {
                if( randomInit ) myxy[j] = (float) (Math.random()*range) - (range/2);
                else myxy[j] = 0;
                //myxy[j]=((((int)(0.7f*j*500))%500) - 500/2)-1;
            }

        	boolean restart;
	        restart=true;
	        while(restart){
                float []tmpstore = copyOne(myxy);
                y[1] = funObjective( tmpstore, lmObjective);
                tmpstore=null;

                for (j=2; j<=num+1; j++) {
                    for( int tmpc=0; tmpc<=num; tmpc++) p[j][tmpc]=myxy[tmpc];
                //memcpy(p[j],myxy,sizeof(float)*(num+1));
                    p[j][j-1] += lambda;
                    tmpstore = copyOne(p[j]);
                    y[j] = funObjective(tmpstore, lmObjective);
                    tmpstore=null;
                }
                //lotusDebug.println("evaluate before amoeba");
                amoeba(p,y,num,localftol,lmObjective,j);
                //lotusDebug.println("evaluate after amoeba");
                if (j < 0) {
                    lotusDebug.println("No answer");
                    restart=false;
                    break; //to the outmost loop
                }

                if (restarted < restarts) {
                    restarted++;
                    continue;
                }else{
                    break;
                }

            }

        	fit = funObjective(copyOne(myxy), lmObjective);
        	//lotusDebug.println("Fit = " + fit);
        	if( restart ){
                float []tmpstore = copyOne( myxy );
	            fit = funObjective(tmpstore, lmObjective);

            	if (fit < min) {
                    min = fit;
                    for (j=0; j<num; j++) {
                      solution[j] = myxy[j+1];
                    }
                }
            }

        }

        if (min == MAX_ERR) {
        /* we did not converge, relax ftol and try again */
            localftol = localftol*10;
            continue;
        }
        else{
            break;
        }

    }

        //lotusDebug.println("finishing the solve function");
    y=null;
    myxy=null;
    for (i=2;i<=num+1;i++) {
      p[i]=null;
    }
    p=null;
    myxy=null;
    return min;
  }

  private float [] vector(int nl, int nh){
    float []v;
    v = new float[nh-nl+1+NR_END];
    return v;
  }

  private void free_vector( float []v, int nl, int nh ){
    v=null;
  }

  /**
   * the coorndiates of the two nodes are expressed as
   * two arrays, their length should be equal to dimensions
   */
  float host_dist( float[] x, float [] y) {
    float dist=0;
    for (int i=0; i<x.length; i++) {
      dist += Math.pow(Math.abs(x[i]-y[i]), 2); //Eu norm
    }
    return (float) Math.pow(dist, 1.0/2.0);

  }

  /**
   * the distance of a host to the landmark with the designated index
   * @param host    logical address of the host
   * @param lmIndex the index of the landmark
   */

  float host_dist( float[] host, int lmIndex ){
    float dist=0;
    for ( int i=0; i<host.length; i++){
      dist += Math.pow(Math.abs(host[i]-coordinates[lmIndex][i]), 2);
    }

    return (float) Math.pow(dist, 1.0/2.0);
  }

  // evaluate the objective function of landmark
  // array: dimension * numLandmarks, the coordinates of the
  ///       landmarks

  /**
   * Evaluate the objective function of the optimization problem
   * @param array the value of the optimization variables
   * @param lmObjective whether to use the landmark or target's objective function
   * @return the objective function value evaluated with the values in array
   */
  private float funObjective( float [] array, boolean lmObjective){
      if( lmObjective)
	  return landmarkObjective(array);
      else
	  return targetObjective(array);
      //return (array[0]-1)*(array[0]-1)+array[1]*array[1];
  }

  /**
   * The objective function of the optimization problem for the landmark
   * @param array the value of the optimization variables
   */
  private float landmarkObjective( float [] array ){
    float dist, sum, error;
    int i, j;

    if( array.length != dimensions * numLandmarks ){
      lotusDebug.println(" land mark objective evaluation error");
      System.exit(1);
    }

    sum = 0;

    for (i=0; i<numLandmarks; i++) {
      for (j=0; j<i; j++) {
	float [] lmx = new float[dimensions];
	float [] lmy = new float[dimensions];

	for( int k=0; k<dimensions; k++){
	  lmx[k] = array[i*dimensions+k];
	  lmy[k] = array[j*dimensions+k];
	}
	dist = host_dist( lmx, lmy );
	error = (lmDelays[i][j] - dist)/lmDelays[i][j];
	sum += error*error;
      }
    }
    return sum;
  }

  //the cost function for the target optimization
  //array: the cooridinates of a single target

  /**
   * The objective function of the optimization problem for the target
   * @param array the value of the optimization variables
   */
  float targetObjective(float []array) {
    float dist, sum, error;
    int i;

    if( array.length != dimensions ){
      lotusDebug.println(" target objective evaluation error");
      System.exit(1);
    }

    sum = 0;
    for (i=0; i<numLandmarks && validDelays[i] ; i++) {
      dist = host_dist(array, i );
      error = (targetDelays[i]-dist)/targetDelays[i];
      sum += error*error;
    }
    return sum;
  }

  /**
   * A subroutine invoked by the solver
   * @param p   input of the n+1 points, each is represented in n-dimension coordinates
   * @param y   objective function value, evaluated at the n+1 points
   * @param ndim    the dimenstion of the optimization, i.e. # of opt variables.
   * @param lmObject    which objective function, the landmark's or the target's
   * @nfunk number of evalution during the invokation
   * @stuff
  */

  private float amotry(float [][]p,
		       float []y,
		       float []psum,
		       int ndim,
		       boolean lmObjective,
		       int ihi,
		       float fac) {

    int j;
    float fac1,fac2,ytry;
    float [] ptry;

    /*lotusDebug.println("ihi " + ihi );
    for( int i = 0; i<ndim; i++)
      System.out.print(psum[i+1] + " " );
    lotusDebug.println();
    */

    ptry=vector(1,ndim);

    fac1=(1.0f-fac)/ndim;
    fac2=fac1-fac;

    for (j=1;j<=ndim;j++) ptry[j]=psum[j]*fac1-p[ihi][j]*fac2;

    //    ytry= fit_p1(&ptry[1], ndim, stuff);
    float []tmpstore = copyOne( ptry );
    ytry = funObjective( tmpstore, lmObjective);
    tmpstore = null;

    //lotusDebug.println("ytry " + ytry);
    if (ytry < y[ihi]) {
      y[ihi] = ytry;
      for (j=1;j<=ndim;j++) {
	psum[j] += ptry[j]-p[ihi][j];
	p[ihi][j]=ptry[j];
      }
    }
    free_vector(ptry,1,ndim);
    return ytry;
  }

  /* p: input of the n+1 points, each is represented in n-dimension coordinates
     y: objective function value, evaluated at the n+1 points
     ndim: the dimenstion of the optimization, i.e. # of opt variables.
     tolerance:
     funk: objective function
     nfunk: number of evalution during the invokation
     stuff:

  */

  void amoeba(float [][]p, float []y, int ndim, float ftol,
	      boolean lmObjective, int nfunk) {

    int i, ihi, ilo, inhi, j, mpts=ndim+1;
    float rtol, sum, swap, ysave, ytry;
    float []psum;

    //lotusDebug.println("amoeba input");
    /*for( i = 1; i<=mpts; i++){
	for( j = 1; j<=ndim; j++){
	    System.out.print(p[i][j] + " " );
	}
	lotusDebug.println("||" + y[i] );
	}*/


    psum = vector(1,ndim);
    nfunk=0;

    for (j=1;j<=ndim;j++) {
      for(sum=0.0f,i=1;i<=mpts;i++) sum += p[i][j];
      psum[j]=sum;
    }

    for(;;) {
      ilo=1;

      //ihi = y[1]>y[2]?(inhi=2,1):(inhi=1,2);
      if( y[1]>y[2]) {
	ihi = 1;
	inhi = 2;
      }
      else{
	ihi = 2;
	inhi = 1;
      }

      for (i=1;i<=mpts;i++) {
	if (y[i] <= y[ilo]) ilo=i;
	if (y[i] > y[ihi]) {
	  inhi=ihi;
	  ihi=i;
	} else if (y[i]>y[inhi] && i!=ihi) inhi = i;
      }

      rtol=(float)(2.0*Math.abs(y[ihi]-y[ilo])/(Math.abs(y[ihi])+Math.abs(y[ilo])+TINY));
      //lotusDebug.println("rtol " + rtol );
      if (rtol < ftol || y[ilo] < 1.0e-6) {
	SWAP( y[1], y[ilo] );
	for (i=1;i<=ndim;i++) {SWAP(p[1][i],p[ilo][i]);}
	break;
      }
      if (nfunk>=NMAX) {
	SWAP( y[1], y[ilo] );
	for (i=1;i<=ndim;i++) {SWAP(p[1][i], p[ilo][i]);}
	break;
      }
      nfunk +=2;
      ytry=amotry(p,y,psum,ndim,lmObjective,ihi,-1.0f);  //c
      //lotusDebug.println("ytry " + ytry);
      if (ytry <= y[ilo])
	ytry=amotry(p,y,psum,ndim,lmObjective,ihi,2.0f); //a
	else if (ytry >= y[inhi]) {
	  ysave=y[ihi];
	  ytry=amotry(p,y,psum,ndim,lmObjective,ihi,0.5f); //b
	  if (ytry >= ysave) {
	    for (i=1;i<=mpts;i++) {
	      if (i != ilo) {
		for (j=1;j<=ndim;j++)
		  p[i][j]=psum[j]=0.5f*(p[i][j]+p[ilo][j]);

		float [] tmpstore = copyOne( psum );
		y[i]=funObjective( tmpstore, lmObjective);
		tmpstore=null;
	      }
	    }
	    nfunk += ndim;
	    for (j=1;j<=ndim;j++) {
	      for(sum=0.0f,i=1;i<=mpts;i++) sum += p[i][j];
	      psum[j]=sum;
	    }

	  }
	}
      else --(nfunk);

      //lotusDebug.println(nfunk);
    }
    free_vector(psum,1,ndim);

    /*lotusDebug.println("amoeba output");
    for( i=1; i<=ndim; i++){
	System.out.print(p[1][i] + " " );
    }

    lotusDebug.println();
    System.exit(0);*/
  }


  // used to accommodate the problem of using array starting from index 1
  private float[] copyOne( float src[] ){
    float []tmpstore = new float[src.length-1];
    for( int tmpc=0; tmpc<src.length-1; tmpc++) tmpstore[tmpc]=src[tmpc+1];
    return tmpstore;
  }

  float tmp;
  private void SWAP( float a, float b) {
    tmp = a;
    a = b;
    b = tmp;
  }




  public static void main(String argv[]) throws Exception{

/*      lotusDebug.println("start testing");
      int numProbe = 4;
      float [][]probes = new float[4][4];

      try{
	  FileReader input = new FileReader("matrix");
	  BufferedReader br = new BufferedReader(input);

	  for( int i=0; i<numProbe; i++){
	      String line = br.readLine();
	      StringTokenizer token = new StringTokenizer(line);

	      for( int j=0; j<i; j++){
		  probes[i][j] = (Float.valueOf( token.nextToken())).floatValue();
	      }
	  }

	  br.close();

      }
      catch( Exception e) { return;}

      landmark lm = new landmark( probes, 2 );

      for( int i=0; i<4; i++){
	  lotusDebug.println( lm.coordinates[i][0] + ":" + lm.coordinates[i][1] );

      }

      for( int i=0; i<4; i++){
	  for( int j=0; j<i; j++){
	      lotusDebug.println( i +
				  " to " +
				  j +
				  " is " +
				  lm.host_dist( lm.coordinates[i], lm.coordinates[j]) +
				  " " +
				  lm.lmDelays[i][j]);
	  }
      }

      FileReader input = new FileReader("combined");
      BufferedReader br = new BufferedReader(input);
      float [] target = new float[4];

      for( int i=0; i<10; i++){
	  String line = br.readLine();
	  StringTokenizer token = new StringTokenizer(line);
	  String ipAddr = token.nextToken();

	  for( int j=0; j<4; j++){
	      target[j] = (Float.valueOf( token.nextToken())).floatValue();
	  }

	  float [] co= lm.fitTargetData( target );

	  lotusDebug.println( ipAddr + ":" + co[0] + "," + co[1] );
	  for( int j=0; j<4; j++){
	      lotusDebug.println( "\t" +
				  lm.host_dist( co, j) +
				  "\t" +
				  target[j] );
				  }

      }

      br.close();

    */
  }

}
