/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, 2002 
  * 
  */
package lotos;

import lotos.*; 
import java.awt.*;
import java.util.*;

/**
 * This interface abstracts the the services provided by the physical layer.
 * A logical node can only access services from the physical network from
 * this interface.
 */

public interface physicalInterface {

    //services to the logical layer
    /** the logical links of this node */
    Enumeration logicalLinks();
    /** the physical links of this node */
    Enumeration physicalLinks();
    /** insert a logical edge
     *  @param n the other end of the edge
     *  @param c the color of the edge, for display purpose
     *  @param width the width of the edge, for display purpose
     */
    boolean insertLogicalEdge( OverlayNode n, Color c, int width);
    /**
     * remove a logical edge
     * @param n the logical neighbor
     */
    boolean removeLogicalEdge( OverlayNode n );
    /** <0 if the physical node is not a neighbor
     *  otherwise, the link weight
     */
    //float physicalWeight( physicalInterface );
    /**
     * @return the physical address of the node
     */
    XYAddress physicalAddress();
    //both the timer and the messages
    /**
     * send a message
     * @param msg the message
     * @return whether the message is sent successfully or not
     */
    boolean sendMsg( SimEventMsg msg);
    /**
     * set a timer event
     * @param interval the time after which this timer expires
     * @param the type of the timer
     */
    void setTimer( float interval, Object index );
    /**
     * @return the current simulation time
     */
    float simTime( );
    
    /** added by Guangyu to provide support for Virtual Adapter to run
     *  hypercast node.
     */
    /**
     * Get the delay for the next specific timer event
     * @param index
     * @return
     */
    float getTimer(Object index);
	/**
	 * Clear the specific timer events
	 * @param index
	 * @return
	 */
	void clearTimer(Object index);
    
}
