/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, 2002 
  * 
  */
package lotos;

import lotos.*; 
import java.awt.*;

/**
 * This is the interfaces implemented by all the overlay protocols.
 * The physical network communiates with the protocol through this interface.
 * The interface abstracts two main functions: </p>
 * a. display related function, i.e. how to display the logial node</p>
 * b. protocol-related function, i.e. how to handle timer/packet event and
 * the start and reset of the protocol
 */
/**
 * revised by Guangyu Dong, Jan 6, 2004
 * add getName().
 */
public interface protocolInterface {

    static final int shapeCircle = 0;
    static final int shapeTriangle = 1;
    static final int shapeSquare = 2;
	
	/* added by Guangyu */
	/** @return a label to be displayer on the GUI */
	public String getName();

    //*display properties
    /** the color of the logical node */
    public Color getColor();
    /** the shape of the logical node */
    public int getShape();

    //*protocol handling
    /** handle timer event */
    public void handleTimer( Object index);

    /** handle message event */
    public void handleMsg( SimEventMsg msg );

    /** start the protocol at this logical node */
    public void startProtocol( );
    /** reset the protocol to its default state */
    public void resetProtocol();
    
    /** convert the protocol setting to a String */
    /** it's used to save the topology to a file */
    //String getConfigString();
    /** configurate the protocol with a String */
    /** it's used to read the topology from a file*/
    //void setConfigString(String config);
    
}
