/***********************************************************************/
/* Copyright (c) 1999-2005, The Rector and Board of Visitors of the    */
/* University of Virginia.                                             */
/* All rights reserved.                                                */
/*                                                                     */
/* This program is being made available under the terms of version 2.1 */
/* of theGNU Lesser General Public License ("LGPL") as published as of */
/* February, 1999 by the Free Software  Foundation; and subject to the */
/* restrictions of the LGPL, you are authorized to  modify and/or      */
/* redistribute it.                                                    */
/*                                                                     */
/* As stated in the LGPL, THIS PROGRAM IS BEING MADE AVAILABLE "AS IS",*/
/* AND WITHOUT A WARRANTY OF ANY KIND, EITHER EXPRESSED OR IMPLIED,    */
/* INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF            */
/* MERCHANTABILITY AND  FITNESS FOR A PARTICULAR PURPOSE. FURTHER, YOU */
/* AGREE NOT TO HOLD THE COPYRIGHT  OWNER(S) LIABLE FOR DAMAGES,       */
/* INCLUDING ANY GENERAL, SPECIAL, INCIDENTAL OR CONSEQUENTIAL DAMAGES,*/
/* ARISING OUT OF THE USE OR INABILITY TO USE  THE PROGRAM.            */
/* See the GNU General Public License for more details.                */
/*                                                                     */
/* You should have received a copy of version 2.1 of the GNU General   */
/* Public License along with this program; if not, you may obtain a    */
/* copy  from the Free Software Foundation website at the following URL*/
/* address (link  active as of September 1, 2005):                     */
/* http://www.fsf.org/licensing/licenses/lgpl.html; or that failing,   */
/* you can request a copy from Free Software Foundation,  Inc., 51     */
/* Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.                */
/*                                                                     */
/* This program is part of the HyperCast software distribution. To     */
/* learn more about HyperCast, please visit the website at             */
/* http://www.hypercast.org.                                           */
/***********************************************************************/


 /* 
  * @author HyperCast Team, 2002 
  * 
  */
package lotos;

import lotos.*; 

import hypercast.ByteArrayUtility;
import hypercast.HTTP_ServerUtility;
import hypercast.HyperCastConfig;
import hypercast.HyperCastStatsException;
import hypercast.I_AdapterCallback;
import hypercast.I_Address;
import hypercast.I_AddressPair;
import hypercast.I_Message;
import hypercast.I_MulticastAdapter;
import hypercast.I_NetworkAddress;
import hypercast.I_Node;
import hypercast.I_PhysicalAddress;
import hypercast.I_Stats;
import hypercast.I_UnderlayAddress;
import hypercast.I_UnicastAdapter;
import hypercast.KeyVault;
import hypercast.NotificationHandler;
import hypercast.SecurityProcessor;
import hypercast.SimpleStats;
import hypercast.StatsProcessor;
import hypercast.events.NODE_NEIGHBORHOODCHANGED;
import hypercast.events.NOTIFICATION_EVENT;
import hypercast.util.XmlUtil;

import java.awt.BorderLayout;
import java.awt.Button;
import java.awt.CheckboxMenuItem;
import java.awt.Color;
import java.awt.Dialog;
import java.awt.FlowLayout;
import java.awt.Frame;
import java.awt.Label;
import java.awt.List;
import java.awt.Menu;
import java.awt.MenuItem;
import java.awt.Panel;
import java.awt.PopupMenu;
import java.awt.TextArea;
import java.awt.TextField;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;
import java.io.ByteArrayOutputStream;
import java.io.CharArrayWriter;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.text.ParseException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.LinkedList;
import java.util.Properties;

import org.apache.xpath.XPath;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;


/**
 * This class implements a network running live Hypercast node
 * 
 * @author Guangyu Dong, Feb 3, 2004
 */

public class simHCNetwork extends OverlayNetwork implements ActionListener, ItemListener{
	
	/** Config object */
	private SimHypercastConfig config;
	
	/** Adapter manger handling creating adapters */
	/*package*/ HC_AdapterManager adpt_mngr;
	
	/** Current protocol */
	private ProtocolEntry currentProtocol = null;
	private SecuritySettingEntry currentSecurity = null;
	
	/** Display properties */
	boolean displayNodeProtocol = true;
	boolean displayNodeLogicalAddress = true;
	boolean displayNodePhysicalAddress = true;
	boolean displaySecuritySetting = true;
	
	private boolean loadSaveFlag_logical = true;
	private boolean loadSaveFlag_physical = true;
	
	GUI the_gui;
	
	MenuItem selectedProtocolMenu;
	CheckboxMenuItem security_menus[];

	/** Constructor
	 * build the network with some topology
	 * @param ta    The topolology adapter
	 * @throws IOException
	 * @throws ParseException
	 * @throws FileNotFoundException
	 */
	public simHCNetwork(TopoAdaptor ta){
		super(ta);
		
		initConfig();
		
	}
	/** Constructor
	 * build the network with some topology
	 * @throws IOException
	 * @throws ParseException
	 * @throws FileNotFoundException
	 */
	public simHCNetwork(){
		super();
		initConfig();
	}
	private void initConfig(){
		try {
			config = new SimHypercastConfig(this, "SimHypercast.xml");
			adpt_mngr = new HC_AdapterManager();
		} catch (FileNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (ClassNotFoundException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (InstantiationException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		} catch (IllegalAccessException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		if(config!=null && config.getProtocolCount()>0){
			//currentProtocol = config.getProtocolByName(config.getProtocolNames()[0]);
		}
		
	}
	/**
	 * Activate a protocol node
	 * @param o			the physical node
	 * @param type		the type of hypercast node
	 * @return		true if activation succeeded.
	 */
	
	private boolean activateNode(OverlayNode o, ProtocolEntry entry, SecuritySettingEntry security, String info, String paStr) {
		//if( dtServer == null ) return false;
		if (o == null)
			return false;

		if (active_nodes.contains(o)) {
			//return false;
			deactivateNode(o);
		}
		//create adapter
		if (adpt_mngr == null) {
			adpt_mngr = new HC_AdapterManager();
		}
		if (config == null) {
			initConfig();
		}

		HC_Adapter adapter;
		
		if(paStr==null)
			adapter = adpt_mngr.createAdapter(o);
		else
			adapter = adpt_mngr.createAdapter(o, Integer.parseInt(paStr));

		//create protocol node
		HyperCastConfig hc_config = entry.getConfig();
		hc_config = new HyperCastConfig(hc_config.getDocument());
		StringWriter tempOutputBuf = new StringWriter();
		PrintWriter tempOutput = new PrintWriter(tempOutputBuf);
		hc_config.err = tempOutput;
		hc_config.log = tempOutput;
		
		simHCNode prot_node = createProtocolNode(o, entry, security, hc_config, adapter, info);
		prot_node.protocolName = entry.getName();
		if(security!=null)
			prot_node.securitySetting = security.name;
		else
			prot_node.securitySetting = "No security";
		
		prot_node.getOutput().print(tempOutputBuf.toString());
		
		hc_config.err = prot_node.getOutput();
		hc_config.log = prot_node.getOutput();
		
		
		o.setProtocol(prot_node);
		active_nodes.addElement(o);

		if (this.isRunning())
			prot_node.startProtocol();

		return true;
	}
	/**
	 * Activate a node without assigning an I_Node object.
	 * @param o
	 * @return
	 */
	private simHCNode activateEmptyNode(OverlayNode o, HyperCastConfig hc_config){
		if (o == null)
			return null;

		if (active_nodes.contains(o)) {
			//return false;
			deactivateNode(o);
		}
		//create adapter
		if (adpt_mngr == null) {
			adpt_mngr = new HC_AdapterManager();
		}
		if (config == null) {
			initConfig();
		}
		
		HC_Adapter adapter;
		
		adapter = adpt_mngr.createAdapter(o);

		//create protocol node
		hc_config = new HyperCastConfig(hc_config.getDocument());
		
		simHCNode prot_node = new simHCNode(this, o, adapter, null, null);
		
		hc_config.err = prot_node.getOutput();
		hc_config.log = prot_node.getOutput();
		
		
		o.setProtocol(prot_node);
		active_nodes.addElement(o);

		return prot_node;
	}
	private simHCNode createProtocolNode(OverlayNode o, ProtocolEntry entry, SecuritySettingEntry security, HyperCastConfig config, HC_Adapter adapter, String info){
		I_SimHyperCastAgent agent = entry.getAgent();
		
		if(security!=null)
			return this.createSecuredNode(o, agent, security, config, adapter, info);
		else
			return new simHCNode(this, o, adapter, agent.createNode(o.getAddress(), config, adapter, info), agent);
	}
	
	private simHCNode createSecuredNode(OverlayNode o, I_SimHyperCastAgent agent, SecuritySettingEntry security, HyperCastConfig config, HC_Adapter adapter, String info){
		HyperCastConfig sec_config = security.config;
		StringWriter tempOutputBuf = new StringWriter();
		PrintWriter tempOutput = new PrintWriter(tempOutputBuf);
		sec_config.err = tempOutput;
		sec_config.log = tempOutput;
		
		KeyVault keyvaultObj = new KeyVault (security.config);
		SecurityProcessor secProc= new SecurityProcessor (security.config, keyvaultObj, adapter);
		I_Node newI_Node = agent.createNode(o.getAddress(), config, secProc, info);
		simHCNode newNode = new simHCNode(this, o, adapter, newI_Node, agent);
		newNode.addNotificationHandler(secProc);
		
		newNode.getOutput().print(tempOutputBuf.toString());
		
		return newNode;
	}
	
	/**
	 * Deactivate the protocol node
	 * @param o		the physical node
	 * @return		true if deactivation succeeded.
	 */
	public boolean deactivateNode(OverlayNode o) {

		if (active_nodes.contains(o)) {
			
			simHCNode hc_node = (simHCNode)o.getProtocol();
			if(hc_node==null)
				return false;
			
			hc_node.agent.nodeRemoved(hc_node.hc_node);
			
			//get adapter
			HC_Adapter adapter = ((simHCNode)o.getProtocol()).getAdapter();
			o.clearLogicalEdges();
			o.getProtocol().resetProtocol();
			o.setProtocol(null);
			o.forwarding = true;

			//remove adapter from the manager
			adpt_mngr.removeAdapter(adapter);

			active_nodes.remove(o);
			return true;

		} else {
			return false;
		}
	}

	/* (non-Javadoc)
	 * @see OverlayNetwork#getLogicalNextHop(OverlayNode, OverlayNode)
	 */
	public OverlayNode getLogicalNextHop(OverlayNode src, OverlayNode dst) {
		simHCNode srcp, dstp;
		srcp = (simHCNode)src.getProtocol();
		dstp = (simHCNode)dst.getProtocol();

		if (srcp != null && dstp != null) {
			I_AddressPair []aps =
				srcp.getHCNode().getParent(
					dstp.getHCNode().getAddressPair().getLogicalAddress());
			if (aps != null && aps.length==1) {
				OverlayNode dstn =
					(OverlayNode)adpt_mngr.searchAddress(
						(SimPhysicalAddress)(aps[0].getPhysicalAddress()));
				return dstn;
			}
		}

		return null;
	}

	/* (non-Javadoc)
	 * @see OverlayNetwork#start()
	 */
	boolean start() {
		this.simRunning = true;
		for (int i = 0; i < active_nodes.size(); i++) {
			OverlayNode onode = (OverlayNode)active_nodes.elementAt(i);
			onode.getProtocol().startProtocol();
		}
		return true;
	}

	/* (non-Javadoc)
	 * @see OverlayNetwork#reset()
	 */
	void reset() {
		super.reset();
		this.simRunning = false;
		for (int i = 0; i < active_nodes.size(); i++) {
			OverlayNode onode = (OverlayNode)active_nodes.elementAt(i);
			onode.getProtocol().resetProtocol();
		}
	}

	/* (non-Javadoc)
	 * @see OverlayNetwork#buildOverlay()
	 */
	boolean buildOverlay() {
		return false;
	}
	/*
	private final String[] treeMenu;
	=
		{
			"Activate",
			"Deactivate",
			"Stats",
			"-----------",
			"SPT",
			"DTServerJoin",
			"DTServer",
			"DTMulticast Node",
			"DTBuddyList Node",
			"HyperCube Node" };
			*/
	//private String currentProtocol = null;
	//private String currentSecuritySetting = null;
	/* (non-Javadoc)
	 * @see OverlayNetwork#getMenus()
	 */
	PopupMenu protocolMenu = null;
	PopupMenu securityMenu;
	public PopupMenu getProtocolMenu(){
		if(protocolMenu==null){
			protocolMenu = new PopupMenu("Protocols");
			String protocols[] = config.getProtocolNames();
			CheckboxMenuItem pm = new CheckboxMenuItem("No protocol");
			protocolMenu.add(pm);
			pm.addItemListener(this);
			
			pm.setState(true);
			for(int i=0;i<protocols.length;i++){
				pm = new CheckboxMenuItem(protocols[i]);
				protocolMenu.add(pm);
				pm.addItemListener(this);
					
			}
		}
		
		return protocolMenu;
	}
	public PopupMenu getSecurityMenu(){
		if(securityMenu==null){
			securityMenu = new PopupMenu("Security");
			CheckboxMenuItem secSubMenu = new CheckboxMenuItem("No security");
			secSubMenu.addItemListener(this);
			if(currentSecurity == null)
				secSubMenu.setState(true);
			securityMenu.add(secSubMenu);
			String secNames[] = config.getSecuritySettingNames();
			security_menus = new CheckboxMenuItem[secNames.length+1];
			security_menus[0] = secSubMenu;
			
			for(int i=0;i<secNames.length;i++){
				secSubMenu =  new CheckboxMenuItem(secNames[i]);
				secSubMenu.addItemListener(this);
				securityMenu.add(secSubMenu);
				security_menus[i+1]=secSubMenu;
				if(currentSecurity!=null && currentSecurity.name.equals(secNames[i])){
					secSubMenu.setState(true);
				}
			}
		}
		
		return securityMenu;
	}
	
	public PopupMenu getMenus() {
		/* initiate menu */
		OverlayNode selectedNode = this.getFirstSelectedNode();
		
		String[] protocolNames = config.getProtocolNames();
		
		PopupMenu allmenu = new PopupMenu("SimHypercast");
		MenuItem mi;
		
		
		mi = new MenuItem("Activate");
		if(selectedNode==null || selectedNode.getProtocol()!=null)
			mi.setEnabled(false);
		
		mi.addActionListener(this);
		allmenu.add(mi);
		mi = new MenuItem("Deactivate");
		if(selectedNode!=null){
			simHCNode hcNode = (simHCNode)selectedNode.getProtocol();
			if(hcNode==null)
				mi.setEnabled(false);
		}else
			mi.setEnabled(false);
		
		mi.addActionListener(this);
		allmenu.add(mi);
		mi = new MenuItem("Node Statistics");
		if(selectedNode!=null){
			simHCNode hcNode = (simHCNode)selectedNode.getProtocol();
			if(hcNode==null)
				mi.setEnabled(false);
		}else
			mi.setEnabled(false);
		
		mi.addActionListener(this);
		allmenu.add(mi);
		
		mi = new MenuItem("Event History");
		if(selectedNode!=null){
			simHCNode hcNode = (simHCNode)selectedNode.getProtocol();
			if(hcNode==null)
				mi.setEnabled(false);
		}else
			mi.setEnabled(false);
		
		mi.addActionListener(this);
		allmenu.add(mi);
		
		if(selectedNodes.size()==1 && selectedNode.getProtocol()!=null){
			simHCNode pNode = (simHCNode)selectedNode.getProtocol();
			
			
			
			for(int i=0;i<protocolNames.length;i++){
				ProtocolEntry entry = config.getProtocolByName(protocolNames[i]);
				if(entry.getAgent().getName().equals(pNode.agent.getName())){
					
					String protocolMenus[] = entry.getAgent().getMenuItems();
					
					if(protocolMenus!=null && protocolMenus.length>0){
					
						//mi = new MenuItem("------");
						//allmenu.add(mi);
						
						Menu pmenu = new Menu(entry.getAgent().getName());
						selectedProtocolMenu = pmenu;
						
						allmenu.add(pmenu);
						
						for(int j=0;j<protocolMenus.length;j++){
							mi = new MenuItem(protocolMenus[j]);
							mi.addActionListener(this);
							pmenu.add(mi);
						}
						
						allmenu.add(pmenu);
					}
				}
			}
		}
		
		
		return allmenu;
	}

	public void showStats(OverlayNode onode) {
		
		if (onode == null)
			return;
		if (onode.getProtocol() == null)
			return;

		I_Node hcnode = ((simHCNode)onode.getProtocol()).getHCNode();
		if (hcnode == null)
			return;
		
		StatsWindow statswin = new StatsWindow(hcnode);
		statswin.show();
	}
	
	public void showEventHistory(OverlayNode onode) {
		
		if (onode == null)
			return;
		if (onode.getProtocol() == null)
			return;

		simHCNode hcnode = ((simHCNode)onode.getProtocol());
		if (hcnode == null)
			return;
		
		EventWindow eventwin = new EventWindow(hcnode);
		eventwin.show();
	}
	
	public void setLoadSaveFlag(boolean logical, boolean physical){
		
		if(!logical && !physical){
			throw new IllegalArgumentException("Need load/save either logical or physical network.");
		}
		
		this.loadSaveFlag_logical = logical;
		this.loadSaveFlag_physical = physical;
	}
	
	/* (non-Javadoc)
	 * @see OverlayNetwork#protocolReader(java.lang.String)
	 */
	boolean protocolReader(String fname) {
		Document doc;
		InputStream xmlinput = FileResourceManager.openFile(fname);
		try {
			doc = XmlUtil.createDocument(xmlinput, fname);
		} catch (FileNotFoundException e) {
			e.printStackTrace();
			return false;
		} catch (IOException e) {
			e.printStackTrace();
			return false;
		}
		
		if(this.loadSaveFlag_physical){
			nodes.clear();
		}
		
		active_nodes.clear();
		adpt_mngr.reset();
		
		Element LoTosProtocolSetting = (Element)doc.getElementsByTagName("OverlayNetwork").item(0);
		NodeList nl = LoTosProtocolSetting.getElementsByTagName("OverlayNode");
		
		for(int i=0;i<nl.getLength();i++){
			Element ovlynode = (Element)nl.item(i);
			String xyaddress = ovlynode.getElementsByTagName("XYAddress").item(0).getChildNodes().item(0).getNodeValue();
			OverlayNode onode =null;
			
			if(this.loadSaveFlag_physical && ovlynode.getElementsByTagName("Physical").getLength()>0){
				Element phyNode = (Element)ovlynode.getElementsByTagName("Physical").item(0);
				Element mobiMode = (Element)phyNode.getElementsByTagName("MobilityMode").item(0);
				String mobiStr = mobiMode.getChildNodes().item(0).getNodeValue();
				float speed = 0;
				
				onode = this.addNode(new XYAddress(xyaddress));
				
				if(!mobiStr.equals("None")){
					Element mobiSpeed = (Element)phyNode.getElementsByTagName("MobilitySpeed").item(0);
					String mobiSpeedStr = mobiSpeed.getChildNodes().item(0).getNodeValue();
					speed = Float.parseFloat(mobiSpeedStr);
					onode.setMobility(mobiStr, speed);
				}else{
					onode.setMobility("");
				}
					
				
				Element wired = (Element)ovlynode.getElementsByTagName("WiredConnection").item(0);
				String wiredStr = wired.getChildNodes().item(0).getNodeValue().trim();
				
				Element wireless = (Element)ovlynode.getElementsByTagName("WirelessConnection").item(0);
				String wirelessStr = wireless.getChildNodes().item(0).getNodeValue().trim();
				
				int linkType=0;
				
				if(wiredStr.equalsIgnoreCase("Yes") && wirelessStr.equalsIgnoreCase("Yes"))
					linkType = OverlayNode.INTERFACE_BOTH;
				else if(wiredStr.equalsIgnoreCase("Yes"))
					linkType = OverlayNode.INTERFACE_WIRED;
				else if(wirelessStr.equalsIgnoreCase("Yes"))
					linkType = OverlayNode.INTERFACE_WIRELESS;
				
				onode.setLinkType(linkType);
				
				
			}else{	
				onode = findClosest(new XYAddress(xyaddress), 0);
			}
			
			if(onode==null)
				onode = this.addNode(new XYAddress(xyaddress));
			
			if(onode!=null && this.loadSaveFlag_logical && ovlynode.getElementsByTagName("Protocol").getLength()>0){
				Element protocol = (Element)ovlynode.getElementsByTagName("Protocol").item(0);
				
				String protocolName = protocol.getElementsByTagName("ProtocolName").item(0).getChildNodes().item(0).getNodeValue();
				String securitySetting = protocol.getElementsByTagName("SecuritySetting").item(0).getChildNodes().item(0).getNodeValue();
				String protocolInfo = protocol.getElementsByTagName("ProtocolInfo").item(0).getChildNodes().item(0).getNodeValue();
				String physicalAddress = protocol.getElementsByTagName("PhysicalAddress").item(0).getChildNodes().item(0).getNodeValue();
				
				ProtocolEntry pEntry = config.getProtocolByName(protocolName);
				SecuritySettingEntry sEntry = config.getSecuritySettingByName(securitySetting);
				
				this.activateNode(onode, pEntry, sEntry, protocolInfo, physicalAddress);
				
				if(pEntry!=null){
					
				}else{
					lotusDebug.println("Invalid protocol: "+protocolName);
				}
				
			}else{
				//lotusDebug.println("Cannot find a overlay node for the protocol!");
			}
		}
		nl = LoTosProtocolSetting.getElementsByTagName("OverlayNode");
		//read links
		for(int i=0;i<nl.getLength();i++){
			Element ovlynode = (Element)nl.item(i);
			String xyaddress = ovlynode.getElementsByTagName("XYAddress").item(0).getChildNodes().item(0).getNodeValue();
			OverlayNode onode = findClosest(new XYAddress(xyaddress), 0);;
			
			//physical links
			if(onode!=null && this.loadSaveFlag_physical && ovlynode.getElementsByTagName("Physical").getLength()>0){
				onode.physical_neighbors.clear();
				Element phyNode = (Element)ovlynode.getElementsByTagName("Physical").item(0);
				NodeList phyLinks = phyNode.getElementsByTagName("PhysicalLink");
				for(int j=0;j<phyLinks.getLength();j++){
					String endxyaddress = ((Element)phyLinks.item(j)).getElementsByTagName("EndPointXYAddress").item(0).getChildNodes().item(0).getNodeValue();
					OverlayNode endNode = findClosest(new XYAddress(endxyaddress), 0);;
					
					if(endNode!=null){
					
						float dist = onode.getDirectDistanceTo( endNode );
						
						this.addEdge(onode, endNode, dist);
					}
				}
			}
			
			simHCNode pnode = (simHCNode)onode.getProtocol();
			if(onode!=null && pnode!=null && this.loadSaveFlag_logical && ovlynode.getElementsByTagName("Protocol").getLength()>0){
				onode.clearLogicalEdges();
				Element protocol = (Element)ovlynode.getElementsByTagName("Protocol").item(0);
				NodeList logLinks = protocol.getElementsByTagName("LogicalLink");
				for(int j=0;j<logLinks.getLength();j++){
					String endxyaddress = ((Element)logLinks.item(j)).getElementsByTagName("EndPointXYAddress").item(0).getChildNodes().item(0).getNodeValue();
					OverlayNode endNode = findClosest(new XYAddress(endxyaddress), 0);;
					
					float dist = onode.getDirectDistanceTo( endNode );
					
					onode.insertLogicalEdge(endNode, Color.yellow, 2);
				
				}
			}
			
		}
		
		return true;
	}
	final static String formatTag = "hypercast";
	/* (non-Javadoc)
	 * @see OverlayNetwork#protocolWriter(java.lang.String)
	 */
	boolean protocolWriter(String fname) {
		Document doc = XmlUtil.createDocument();
		Element LoTosProtocolSetting = doc.createElement("OverlayNetwork");
		doc.appendChild(LoTosProtocolSetting);
		
		for(int i=0;i<this.nodes.size();i++){
			OverlayNode onode = (OverlayNode)nodes.get(i);
			
			Element ovlynode = doc.createElement("OverlayNode");
			
			
			Element xyaddress = doc.createElement("XYAddress");
			xyaddress.appendChild(doc.createTextNode(""+onode.getAddress()));
			ovlynode.appendChild(xyaddress);
			
			if(this.loadSaveFlag_physical){
				LoTosProtocolSetting.appendChild(ovlynode);
				
				Element phynode = doc.createElement("Physical");
				ovlynode.appendChild(phynode);
				
				String mobiStr = onode.getMobilityStr();
				if(mobiStr==null)
					mobiStr = "None";
				
				Element mobiMode = doc.createElement("MobilityMode");
				mobiMode.appendChild(doc.createTextNode(mobiStr));
				phynode.appendChild(mobiMode);
				
				if(!mobiStr.equals("None")){
					Element mobiSpeed = doc.createElement("MobilitySpeed");
					mobiSpeed.appendChild(doc.createTextNode(""+onode.mobile_speed));
					phynode.appendChild(mobiSpeed);
				}
				
				String wiredStr = onode.wired_link?"Yes":"No";
				String wirelessStr = onode.wireless_link?"Yes":"No";
				
				Element wired = doc.createElement("WiredConnection");
				wired.appendChild(doc.createTextNode(wiredStr));
				phynode.appendChild(wired);
				
				Element wireless = doc.createElement("WirelessConnection");
				wireless.appendChild(doc.createTextNode(wirelessStr));
				phynode.appendChild(wireless);
				
				//save the physical links
				for(int j=0;j<onode.physical_neighbors.size();j++){
					OverlayNode endNode = (OverlayNode)onode.physical_neighbors.get(j);
					Element phyEdge = doc.createElement("PhysicalLink");
					Element endXy = doc.createElement("EndPointXYAddress");
					endXy.appendChild(doc.createTextNode(endNode.getAddress().toString()));;
					phyEdge.appendChild(endXy);
					phynode.appendChild(phyEdge);
				}
				
				
			}
			
			simHCNode pnode = (simHCNode)onode.getProtocol();
			if(this.loadSaveFlag_logical && pnode!=null){
				if(!this.loadSaveFlag_physical)
					LoTosProtocolSetting.appendChild(ovlynode);
				
				Element protocol = doc.createElement("Protocol");
				ovlynode.appendChild(protocol);
				
				Element protocolName = doc.createElement("ProtocolName");
				protocolName.appendChild(doc.createTextNode(""+pnode.protocolName));
				protocol.appendChild(protocolName);
				
				Element securitySetting = doc.createElement("SecuritySetting");
				securitySetting.appendChild(doc.createTextNode(""+pnode.securitySetting));
				protocol.appendChild(securitySetting);
				
				Element info = doc.createElement("ProtocolInfo");
				info.appendChild(doc.createTextNode(""+pnode.agent.getInfoForNode(pnode.hc_node)));
				protocol.appendChild(info);
				
				Element pa = doc.createElement("PhysicalAddress");
				pa.appendChild(doc.createTextNode(""+pnode.adapter.createPhysicalAddress()));
				protocol.appendChild(pa);
				
				//save the logical links
				for(int j=0;j<onode.logical_neighbors.size();j++){
					OverlayNode endNode = (OverlayNode)onode.logical_neighbors.get(j);
					Element logEdge = doc.createElement("LogicalLink");
					Element endXy = doc.createElement("EndPointXYAddress");
					endXy.appendChild(doc.createTextNode(endNode.getAddress().toString()));;
					logEdge.appendChild(endXy);
					protocol.appendChild(logEdge);
				}
				
			}
		}
		
		
		try {
			FileOutputStream output = new FileOutputStream(fname);
			XmlUtil.writeXml(doc, output);
		} catch (Exception e) {
			lotusDebug.errorln("Error when saving network to file: "+fname);
	        return false;
		}
		return true;
	}
	/* (non-Javadoc)
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public synchronized void actionPerformed(ActionEvent arg0) {
		if(arg0.getSource() instanceof MenuItem){
			MenuItem mi = (MenuItem)arg0.getSource();
			String menu_name = mi.getLabel();
			
			Enumeration e = selectedNodes.elements();
			
			while(e.hasMoreElements()){
				OverlayNode selectedNode = (OverlayNode)e.nextElement();
				
				if(menu_name.equals(("Activate"))){
					activateNode(selectedNode);
				}else if(menu_name.equals(("Deactivate"))){
					deactivateNode(selectedNode);
				}else if(menu_name.equals("Node Statistics")){
					showStats(selectedNode);
				}
				else if(menu_name.equals("Event History")){
					showEventHistory(selectedNode);
				}
			}
			
		}
		if(the_gui!=null){
			the_gui.repaint();
			if(the_gui.dw!=null)
				the_gui.dw.repaint();
		}
		
	}
	
	
	/* (non-Javadoc)
	 * @see OverlayNetwork#activateNode(OverlayNode)
	 */
	public boolean activateNode(OverlayNode o) {
		if(o==null)return false;
		if(currentProtocol!=null)
			return activateNode(o, currentProtocol, currentSecurity, null, null);
		else
			return false;
	}
	/* (non-Javadoc)
	 * @see java.awt.event.ItemListener#itemStateChanged(java.awt.event.ItemEvent)
	 */
	public synchronized void itemStateChanged(ItemEvent arg0) {
		
		CheckboxMenuItem mi = (CheckboxMenuItem)arg0.getSource();
		String menu_name = mi.getLabel();
		Object parentMenu = mi.getParent();
		
		if(selectedProtocolMenu!=null && parentMenu == selectedProtocolMenu){
			
			Enumeration e = selectedNodes.elements();
			while(e.hasMoreElements()){
				OverlayNode selectedNode = (OverlayNode)e.nextElement();
			
				simHCNode pnode = (simHCNode)selectedNode.getProtocol();
				if(pnode!=null){
					pnode.agent.menuFired(menu_name);
				}
			}
		}else if(parentMenu == securityMenu){
			chooseSecurity(menu_name);
			
		}
		
		else if(parentMenu == protocolMenu){
			chooseProtocol(menu_name);
		}
	}
	public String getCurrentProtocolName(){
		if(this.currentProtocol!=null){
			return currentProtocol.getName();
		}else{
			return "No protocol";
		}
	}
	public String getCurrentSecurity(){
		if(this.currentSecurity!=null){
			return this.currentSecurity.name;
		}else{
			return "No security";
		}
		
	}
	public void chooseSecurity(String menu_name){
		getSecurityMenu();
		
		if(menu_name.equals("No security")){
			currentSecurity = null;
		}else{
		
			String[] secNames = config.getSecuritySettingNames();
			
			currentSecurity = config.getSecuritySettingByName(menu_name);
			
		}
		for(int i=0;i<security_menus.length;i++){
			if(security_menus[i].getLabel().equals(menu_name)){
				security_menus[i].setState(true);
			}else{
				security_menus[i].setState(false);
			}
		}
		
		Enumeration e = selectedNodes.elements();
		while(e.hasMoreElements()){
			OverlayNode selectedNode = (OverlayNode)e.nextElement();
			simHCNode pnode = (simHCNode)selectedNode.getProtocol();
			
			if(pnode!=null){
				if(!pnode.securitySetting.equals(menu_name)){
					ProtocolEntry entry = config.getProtocolByName(pnode.protocolName);
					String info = pnode.agent.getInfoForNode(pnode.hc_node);
					String paStr = pnode.adapter.createPhysicalAddress().toString();
					if(entry!=null){
						this.deactivateNode(selectedNode);
						this.activateNode(selectedNode, entry, currentSecurity, info, paStr);
					}
				}
						
			}
		}
		if(the_gui!=null){
			the_gui.repaint();
			if(the_gui.dw!=null)
				the_gui.dw.repaint();
		}
	}
	
	public void chooseProtocol(String menu_name){
		getProtocolMenu();
		ProtocolEntry entry=null;
		
		if(menu_name.equals("No protocol")){
			entry = null;
		}else{
		
			String[] protocolNames = config.getProtocolNames();
			
			entry = config.getProtocolByName(menu_name);
		}
		currentProtocol = entry;
		Enumeration e = selectedNodes.elements();
		while(e.hasMoreElements()){
			OverlayNode selectedNode = (OverlayNode)e.nextElement();
			simHCNode pnode = (simHCNode)selectedNode.getProtocol();
				
			if(pnode!=null){
				if(entry!=null && pnode.agent != entry.getAgent()){
					this.deactivateNode(selectedNode);
					this.activateNode(selectedNode);
				}else if(entry==null){
					this.deactivateNode(selectedNode);
				}
						
			}else if(entry!=null){
				this.activateNode(selectedNode);
			}
		}
				
		for(int i=0;i<protocolMenu.getItemCount();i++){
			CheckboxMenuItem pmi = (CheckboxMenuItem)protocolMenu.getItem(i);
			if(pmi.getLabel().equals(menu_name)){
				pmi.setState(true);
			}else{
				pmi.setState(false);
			}
		}
		if(the_gui!=null){
			the_gui.repaint();
			if(the_gui.dw!=null)
				the_gui.dw.repaint();
		}
		
	}

}
interface EventHistoryEntry{
	public float getTimestamp();
	public String getSummary();
	public String getDetails();
}

class NotificationHistoryEntry implements EventHistoryEntry{
	float timestamp;
	NOTIFICATION_EVENT notification;
	
	public NotificationHistoryEntry(float timestamp, NOTIFICATION_EVENT notification){
		this.timestamp = timestamp;
		this.notification = notification;
	}

	/* (non-Javadoc)
	 * @see EventHistoryEntry#getTimestamp()
	 */
	public float getTimestamp() {
		return timestamp;
	}

	/* (non-Javadoc)
	 * @see EventHistoryEntry#getSummary()
	 */
	public String getSummary() {
		return "Notification Event" ;
	}

	/* (non-Javadoc)
	 * @see EventHistoryEntry#getDetails()
	 */
	public String getDetails() {
		String details = "";
		details += "Type = " +notification.getClass().getName() +"\n\n";
		details += "Occurred at " + ((float)notification.getTimestamp()/1000) +"s\n\n";
		details += "Information = " + notification.getInfo() +"\n\n";
		
		return details;
	}
}
class TimerHistoryEntry implements EventHistoryEntry{
	float timestamp;
	boolean sent;
	Object timer;
	float delay;
	
	public TimerHistoryEntry(float timestamp, boolean sent, Object timer, float delay){
		this.timestamp = timestamp;
		this.sent = sent;
		this.timer = timer;
		this.delay = delay;
	}
	public float getTimestamp(){
		return timestamp;
	}
	public String getSummary(){
		if(sent){
			return "Timer set delay="+delay+"s"; 
		}else{
			return "Timer expires";
		}
	}
	/* (non-Javadoc)
	 * @see EventHistoryEntry#getDetails()
	 */
	public String getDetails() {
		String details = "";
		details += "Type = "+timer.getClass().getName() +"\n\n";
		details += "Timer.toString() = " +timer +"\n\n";
		if(sent)
			details += "Delay = "+delay+"s";
		
		return details;
	}
	
}
class MessageHistoryEntry implements EventHistoryEntry{
	float timestamp;
	boolean sent;
	HC_EventMsg eventMsg;
	
	String src, dst;
	
	byte payload[];
	
	public MessageHistoryEntry(float timestamp, boolean sent, HC_EventMsg msg, String src, String dst){
		this.timestamp = timestamp;
		this.sent = sent;
		this.eventMsg = msg;
		
		this.src = src;
		this.dst = dst;
		
		payload = eventMsg.getHCMessage().toByteArray();
	}
	public float getTimestamp(){
		return timestamp;
	}
	public String getSummary(){
		if(sent){
			return "Message sent to "+dst; 
		}else{
			return "Message received from "+src;
		}
	}
	public String getDetails(){
		String details = "";
		
		details += "Src="+src+"    Dst="+dst+"    Size="+payload.length+"\n\n";
		
		details += "Message.toString() = "+eventMsg.getHCMessage().toString()+"\n\n";
		
		for(int i=0;i<payload.length;i++){
			byte []test = new byte[1];
			test[0]=payload[i];
			details += HTTP_ServerUtility.toHexString(test)+" ";
		}
		
		
		
		return details;
	}
}
/**
 *
 * * This dialog show the timer and messages that have been sent and received by a node
 * @author Guangyu Dong
 *
 */
class EventWindow extends Dialog implements ActionListener, ItemListener{
	simHCNode hcNode;
	List lt_mt;
	TextArea ta_details;
	TextArea ta_output;
	Button b_update;
	Button b_clear;
	Button b_close;
	
	public EventWindow(simHCNode hcNode){
		super(new Frame(), "Event History", false);
		
		this.setSize(800,600);
		this.setLocation(300, 100);
		
		this.setLayout(new BorderLayout());
		
		Panel bottomPanel = new Panel();
		
		b_update = new Button("Update");
		b_update.addActionListener(this);
		bottomPanel.add(b_update);
		
		
		b_clear = new Button("Clear");
		b_clear.addActionListener(this);
		bottomPanel.add(b_clear);
		
		b_close = new Button("Close");
		b_close.addActionListener(this);
		bottomPanel.add(b_close);
		
		this.add(bottomPanel, BorderLayout.SOUTH);
		
		Panel centerPanel = new Panel();
		centerPanel.setLayout(new BorderLayout());
		
		lt_mt = new List();
		lt_mt.addItemListener(this);
		centerPanel.add(lt_mt, BorderLayout.WEST);
		
		Panel displayPanel = new Panel();
		displayPanel.setLayout(new BorderLayout());
		
		Panel detailPanel = new Panel();
		detailPanel.setLayout(new BorderLayout());
		
		Panel outputPanel = new Panel();
		outputPanel.setLayout(new BorderLayout());
		
		displayPanel.add(detailPanel, BorderLayout.NORTH);
		displayPanel.add(outputPanel, BorderLayout.CENTER);
		
		detailPanel.setSize(200, 100);
		
		ta_details = new TextArea("<Empty>", 15 , 80 , TextArea.SCROLLBARS_VERTICAL_ONLY);
		detailPanel.add(ta_details, BorderLayout.CENTER);
		
		ta_output = new TextArea("<No Output>", 25 , 80 , TextArea.SCROLLBARS_VERTICAL_ONLY);
		ta_output.setEditable(false);
		outputPanel.add(ta_output, BorderLayout.CENTER);
		
		centerPanel.add(displayPanel, BorderLayout.CENTER);
		
		this.add(centerPanel, BorderLayout.CENTER);
		
		setHCNode(hcNode);
	}
	public void setHCNode(simHCNode hcnode){
		hcNode = hcnode;
		if(hcNode!=null){
			this.setTitle("Event History: "+hcnode.getName());
			ta_details.setText("<Empty>");
			lt_mt.clear();
			for(int i=0;i<hcnode.eventList.size();i++){
				EventHistoryEntry entry = (EventHistoryEntry)hcNode.eventList.get(i);
				String timestr = "" +entry.getTimestamp();
				while(timestr.length()<9)
					timestr+="0";
				
				lt_mt.add(timestr+"s : " +entry.getSummary());
			}
			ta_output.setText(hcNode.outputBuf.toString());
		}
	}
	
	/* (non-Javadoc)
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent arg0) {
		if (arg0.getSource() == b_close) {
			this.dispose();
		} else if(arg0.getSource() == b_update){
			if(hcNode!=null){
				setHCNode(hcNode);
			}
		}
		else if(arg0.getSource() == b_clear){
			if(hcNode!=null){
				hcNode.eventList.clear();
				hcNode.outputBuf.reset();
				setHCNode(hcNode);
			}
		}
	}

	/* (non-Javadoc)
	 * @see java.awt.event.ItemListener#itemStateChanged(java.awt.event.ItemEvent)
	 */
	public void itemStateChanged(ItemEvent arg0) {
		if(arg0.getSource()==lt_mt){
			int index = lt_mt.getSelectedIndex();
			if(index>=0){
				EventHistoryEntry entry = (EventHistoryEntry)hcNode.eventList.get(index); 
				ta_details.setText(entry.getDetails());
			}else{
				ta_details.setText("<Empty>");
				
			}
		}
		
	}
	
}
/**
 * This dialog can show or set the statistics of a hypcercast node
 * 
 * @author Guangyu Dong, Feb 11, 2004
 */
class StatsWindow extends Dialog implements ActionListener {
	TextField tf_statsname;
	TextArea tf_statsvalue;
	Label l_msg;
	Button b_get;
	Button b_set;
	Button b_readschema;
	Button b_writeschema;
	Button b_close;

	I_Stats stats;

	public StatsWindow(I_Stats _stats) {
		super(new Frame(), "Access Statistics", false);

		stats = _stats;

		this.setSize(650,450);
		this.setLocation(300, 100);

		tf_statsname = new TextField(80);
		tf_statsvalue = new TextArea();
		tf_statsvalue.setColumns(80);
		tf_statsvalue.setRows(20);
		
		l_msg=new Label("                                                                                                                                                                                                          ");
		
		this.setLayout(new FlowLayout());
		this.add(new Label("XPath"));
		this.add(tf_statsname);
		this.add(new Label("Value"));
		this.add(tf_statsvalue);
		b_get = new Button("getStats");
		b_set = new Button("setStats");
		b_readschema = new Button("getReadSchema");
		b_writeschema = new Button("getWriteSchema");
		b_close = new Button("close");
		
		this.add(b_get);
		this.add(b_set);
		this.add(b_readschema);
		this.add(b_writeschema);
		this.add(b_close);
		this.add(l_msg);
		
		b_close.addActionListener(this);
		b_get.addActionListener(this);
		b_set.addActionListener(this);
		b_readschema.addActionListener(this);
		b_writeschema.addActionListener(this);

	}
	
	public void setStats(I_Stats _stats){
		stats = _stats;
	}
	/* (non-Javadoc)
	 * @see java.awt.event.ActionListener#actionPerformed(java.awt.event.ActionEvent)
	 */
	public void actionPerformed(ActionEvent e) {
		l_msg.setText("");
		
		if (e.getSource() == b_close) {
			this.dispose();
			return;
		}
		XPath xpath = XmlUtil.createXPath(this.tf_statsname.getText());
		
		if(xpath==null){
			this.tf_statsvalue.setText("Invalid XPath: "+this.tf_statsname.getText());
			return;
		}
		if (e.getSource() == b_get) {
			
			try {
				Document doc = XmlUtil.createDocument();
				Element [] results = stats.getStats(doc, xpath);
				String result_txt = "";
				
				if(results.length<=0)
					result_txt = "<Empty>";
				
				for(int i=0;i<results.length;i++){
					doc.appendChild(results[i]);
					ByteArrayOutputStream byteOutput = new ByteArrayOutputStream();
					XmlUtil.writeXml(doc, byteOutput, true);
					result_txt += byteOutput.toString() +"\n";
					doc.removeChild(results[i]);
				}
				
				this.tf_statsvalue.setText(result_txt);
					
			} catch (Exception e1) {
				l_msg.setText(e1.toString());
			} 
			
		} 
		else if (e.getSource() == b_readschema) {
			try {
				Document doc = XmlUtil.createDocument();
				Element [] results = stats.getReadSchema(doc, xpath);
				String result_txt = "";
				
				if(results.length<=0)
					result_txt = "<Empty>";
				
				for(int i=0;i<results.length;i++){
					doc.appendChild(results[i]);
					ByteArrayOutputStream byteOutput = new ByteArrayOutputStream();
					XmlUtil.writeXml(doc, byteOutput, true);
					result_txt += byteOutput.toString() +"\n";
					doc.removeChild(results[i]);
				}
				
				this.tf_statsvalue.setText(result_txt);
					
			} catch (Exception e1) {
				l_msg.setText(e1.toString());
			} 
			
		} 
		else if (e.getSource() == b_writeschema) {
			try {
				Document doc = XmlUtil.createDocument();
				Element [] results = stats.getWriteSchema(doc, xpath);
				String result_txt = "";
				
				if(results.length<=0)
					result_txt = "<Empty>";
				
				for(int i=0;i<results.length;i++){
					doc.appendChild(results[i]);
					ByteArrayOutputStream byteOutput = new ByteArrayOutputStream();
					XmlUtil.writeXml(doc, byteOutput, true);
					result_txt += byteOutput.toString() +"\n";
					doc.removeChild(results[i]);
				}
				
				this.tf_statsvalue.setText(result_txt);
					
			} catch (Exception e1) {
				l_msg.setText(e1.toString());
			} 
			
		} 
		else if (e.getSource() == b_set) {
			try {
				Document doc = XmlUtil.createDocumentFromString(tf_statsvalue.getText());
				if(doc.getChildNodes().getLength()>0){
					Element results [] = stats.setStats(
						doc,
						xpath,
						(Element)doc.getChildNodes().item(0));
					
					String result_txt = "";
					
					if(results.length<=0)
						result_txt = "<Empty>";
					
					for(int i=0;i<results.length;i++){
						doc.appendChild(results[i]);
						ByteArrayOutputStream byteOutput = new ByteArrayOutputStream();
						XmlUtil.writeXml(doc, byteOutput, true);
						result_txt += byteOutput.toString() +"\n";
						doc.removeChild(results[i]);
					}
					
					this.tf_statsvalue.setText(result_txt);
					
				}
				
			} catch (Exception e1) {
				l_msg.setText(e1.toString());
			} 
		}
		
	}

}
/**
 * This class implements a Protocol containing
 * a Hypercast Node.
 * 
 * @author Guangyu Dong, Feb 3, 2004
 */
class simHCNode extends NotificationHandler implements protocolInterface{
	
	private static int MAXIMUM_SIZE_OF_EVENT_HISTORY = 100;
	
	/** the agent for the hypercast protocol */
	I_SimHyperCastAgent agent;
	
	String protocolName;
	
	String securitySetting = "No security";
	
	/** the I_Node object to run the protocol */
	I_Node hc_node;

	/** the physical node */
	physicalInterface phy;

	/** the virtual adapter */
	HC_Adapter adapter;
	
	/** if started */
	boolean joined=false;
	
	/** the network object */
	simHCNetwork network;
	
	/** the event list */
	LinkedList eventList;
	
	/** the output of protocol */
	CharArrayWriter outputBuf;
	PrintWriter output;
	
	public String toString(){
		return this.getDetailedName();
	}
	/**
	 * Constructor
	 * @param p			the physical node
	 * @param adapter	the virtual adapter
	 * @param node		the hypercast node
	 */
	public simHCNode(simHCNetwork network, physicalInterface p, HC_Adapter adpt, I_Node node, I_SimHyperCastAgent agent) {
		phy = p;
		adapter = adpt;
		setHCNode(node);
		
		this.agent = agent;
		this.network = network;
		
		eventList = new LinkedList();
		
		outputBuf = new CharArrayWriter();
		output = new PrintWriter(outputBuf);
		
		joined=false;
	}
	public void setHCNode(I_Node node){
		hc_node = node;
		if(hc_node!=null)
			hc_node.setNotificationHandler(this);
	}
	
	public PrintWriter getOutput(){
		return output;
	}
	
	/**
	 * @return	the hypercast node
	 */
	public I_Node getHCNode() {
		return hc_node;
	}
	/**
	 * @return	the adapter
	 */
	public HC_Adapter getAdapter() {
		return adapter;
	}
	/* (non-Javadoc)
	 * @see protocolInterface#getName()
	 */
	public String getName() {
		String label;
		
		label = "";
		if(network.displayNodeProtocol){
			label += agent.getLabelForNode(hc_node);
			if(network.displayNodeLogicalAddress){
				
				if(hc_node.getAddressPair()!=null && hc_node.getAddressPair().getLogicalAddress()!=null)
					label += ":";
			}
		}
		if(network.displayNodeLogicalAddress){
			if(hc_node.getAddressPair()!=null && hc_node.getAddressPair().getLogicalAddress()!=null)
				label +=hc_node.getAddressPair().getLogicalAddress().toString();
		}
		if(network.displayNodePhysicalAddress){
			label +="@"+adapter.createPhysicalAddress();
		}
		if(network.displaySecuritySetting && !securitySetting.equals("No security")){
			label +=":" + securitySetting;
		}
		return label;
		
	}
	public String getDetailedName(){
		String label = "";
		
		label += agent.getLabelForNode(hc_node);
		
		if(hc_node.getAddressPair()!=null && hc_node.getAddressPair().getLogicalAddress()!=null){
			label += ":";
			label +=hc_node.getAddressPair().getLogicalAddress().toString();
		}
		label +="@"+adapter.createPhysicalAddress();
		
		return label;
	}

	/* (non-Javadoc)
	 * @see protocolInterface#getColor()
	 */
	public Color getColor() {
		return agent.getColorForNode(hc_node);
	}

	/* (non-Javadoc)
	 * @see protocolInterface#getShape()
	 */
	public int getShape() {
		return agent.getShapeForNode(hc_node);
	}
	void checkEventListSize(){
		if(eventList.size()>=MAXIMUM_SIZE_OF_EVENT_HISTORY){
			eventList.removeFirst();
		}
	}
	
	public void handler (final NOTIFICATION_EVENT event){
		checkEventListSize();
		this.eventList.add(new NotificationHistoryEntry(network.simTime(), event));
		updateLogicalLinks();
	}
	private void updateLogicalLinks () {
		//update logical links
		//clear all links
		 ((OverlayNode)phy).clearLogicalEdges();
		//update links
		try {

			I_AddressPair[] aps = hc_node.getAllNeighbors();
			if (aps != null) {
				for (int i = 0; i < aps.length; i++) {
					SimPhysicalAddress ha = (SimPhysicalAddress)aps[i].getPhysicalAddress();
					OverlayNode neighbor =
						(OverlayNode)adapter
							.getAdapterManager()
							.searchAddress(
							ha);
					if (neighbor != null) {
						phy.insertLogicalEdge(neighbor, Color.yellow, 2);
						neighbor.insertLogicalEdge(
							(OverlayNode)phy,
							Color.yellow,
							2);
					}
				}
			}
		} catch (Exception e) {

		}
	}
	/* (non-Javadoc)
	 * @see protocolInterface#handleTimer(int)
	 */
	public void handleTimer(Object index) {
		if(!joined)return;
		
		if(index instanceof NeighborsUpdateTimer){
			this.updateLogicalLinks();
			phy.setTimer(NeighborsUpdateTime, new NeighborsUpdateTimer());
		}else{
			adapter.handleTimer(index);
		}
	}

	/* (non-Javadoc)
	 * @see protocolInterface#handleMsg(SimEventMsg)
	 */
	public void handleMsg(SimEventMsg msg) {
		if(!joined)return;
		
		OverlayNode src = msg.getSrc();
		if(src == null )return;
		simHCNode srcProtocol = (simHCNode)src.getProtocol();
		if(srcProtocol==null)return;
		
		if(!srcProtocol.agent.getName().equals(agent.getName()))return;
		
		adapter.handleMsg((HC_EventMsg)msg);

	}

	/** the timer to update the logical links */
	class NeighborsUpdateTimer {
		public boolean equals(Object o){
			if(o==null)
				return false;
			
			return o instanceof NeighborsUpdateTimer;
		}
	}
	private float NeighborsUpdateTime = 5.0f;

	/* (non-Javadoc)
	 * @see protocolInterface#startProtocol()
	 */
	public void startProtocol() {
		if(joined)return;
		try{
			joined=true;
			if(hc_node!=null){
				hc_node.joinOverlay();
				phy.setTimer(NeighborsUpdateTime, new NeighborsUpdateTimer());
			}
			
		}catch(Exception e){
			
		}
	}

	/* (non-Javadoc)
	 * @see protocolInterface#resetProtocol()
	 */
	public void resetProtocol() {
		if(!joined)return;
		try{
			((OverlayNode)phy).clearLogicalEdges();
			if(hc_node!=null)
				hc_node.leaveOverlay();
			this.adapter.Reset(null);
			this.eventList.clear();
			this.outputBuf.reset();
			
			joined=false;
		}catch(Exception e){
			e.printStackTrace();
		}
	}

}


/**
 * This class implements a virtual physical address
 * used to run Hypercast Node in Lotos
 * 
 * @author Guangyu Dong, Feb 2, 2004
 */
class SimPhysicalAddress implements I_PhysicalAddress, I_UnderlayAddress {
	/** the ID */
	private int ID;
	
	static final int HC_ADDR_SIZE=4;
	
	/** 
	 * Constructor
	 * @param _id	the Id
	 * @param p		the physical node
	 */
	public SimPhysicalAddress(int _id){
		ID=_id;
	}
	public SimPhysicalAddress(byte []array, int offset){
		ID=ByteArrayUtility.toInteger(array, offset);
	}
	/**
	 * @return	the id associated with this address
	 */
	public int getId(){
		return ID;
	}
	/**
	 * @return	string value
	 */
	public String toString(){
		return ""+ID;
	}
	
	/**
	 * @return	the physical node using this address
	 */
	/*
	public physicalInterface getPhyNode(){
		return phy;
	}
	*/
	/* (non-Javadoc)
	 * @see hypercast.I_Address#greaterThan(hypercast.I_Address)
	 */
	public boolean greaterThan(I_Address a) {
		if (a == null || !(a instanceof SimPhysicalAddress))
			throw new IllegalArgumentException("Attempted to compare LotosAddress with something of a different type!");
		return getId()>((SimPhysicalAddress)a).getId();
	}

	/* (non-Javadoc)
	 * @see hypercast.I_Address#toByteArray()
	 */
	public byte[] toByteArray() {
		byte[] array = new byte[4];

		System.arraycopy(ByteArrayUtility.toByteArray(getId()), 0, array, 0, 4);
        
		return array;
	}

	/* (non-Javadoc)
	 * @see hypercast.I_Address#getSize()
	 */
	public int getSize() {
		return 4;
	}
	/**
	 * compare two addresses
	 */
	public boolean equals(Object o){
		if((o instanceof SimPhysicalAddress)){
			return ((SimPhysicalAddress)o).getId()==this.getId();
		}else return false;
		
	}
	/* (non-Javadoc)
	 * @see hypercast.I_PhysicalAddress#addressCheck(java.lang.Object)
	 */
	public boolean addressCheck(Object o) {
		return this.equals(o);
	}
		
	
	
}


/**
 * This class implements a virtual adapter
 * used to run Hypercast Node in Lotos
 * 
 * @author Guangyu Dong, Feb 2, 2004
 */

 class HC_Adapter implements I_MulticastAdapter {

	/** the physical node using this adapter */
	private physicalInterface phy;
	/** the physical address of this adapter */
	private SimPhysicalAddress phy_addr;
	/** callback */
	private I_AdapterCallback callback;

	/** state */
	private boolean started;

	/** adapter manager */
	private HC_AdapterManager manager;
	
	String statsName = "NodeAdapter";
	
	/**
	 * A StatsProcessor instance which handles the statistics in this object.
	 */
	protected StatsProcessor statsPro;
	
	/**
	 *  Traffic Statistics 
	 */
	long UReceivedPackets = 0;
	long MReceivedPackets = 0;
	long USentPackets = 0;
	long MSentPackets = 0;
	long UReceivedBytes = 0;
	long MReceivedBytes = 0;
	long USentBytes = 0;
	long MSentBytes = 0;
	
	
	/**
	 * Constructor
	 * @param p		the physical node using this adapter
	 * @param addr	the physical address assigned to this adapter
	 */
	public HC_Adapter(
		physicalInterface p,
		SimPhysicalAddress addr,
		HC_AdapterManager man) {
		
		phy = p;
		phy_addr = addr;
		started = false;
		manager = man;
		
		this.InitStatisticsStructure();
	}
	
	private void InitStatisticsStructure()
	{
		
		/*
		 * Initial StatsProcessor instance.
		 */
		statsPro = new StatsProcessor(this/* I_Stats object */, true/* readable flag */, true/* writable flag */);
		
		statsPro.addStatsElement("UPacketsReceived", new UnicastPacketsReceived()/* I_Stats object */, 1, 1);

		statsPro.addStatsElement("UBytesReceived", new UnicastBytesReceived()/* I_Stats object */, 1, 1);

		statsPro.addStatsElement("UPacketsSent", new UnicastPacketsSent()/* I_Stats object */, 1, 1);

		statsPro.addStatsElement("UBytesSent", new UnicastBytesSent()/* I_Stats object */, 1, 1);
		
		statsPro.addStatsElement("MPacketsReceived", new MulticastPacketsReceived(), 1, 1);

		statsPro.addStatsElement("MBytesReceived", new MulticastBytesReceived(), 1, 1);

		statsPro.addStatsElement("MPacketsSent", new MulticastPacketsSent(), 1, 1);

		statsPro.addStatsElement("MBytesSent", new MulticastBytesSent(), 1, 1);
	}
	
	
	/**
	 * This class manages simple statistic "UPacketsReceived".
	 */
	class UnicastPacketsReceived extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+UReceivedPackets);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchame
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "UBytesReceived".
	 */
	class UnicastBytesReceived extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+UReceivedBytes);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchame
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "UPacketsSent".
	 */
	class UnicastPacketsSent extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+USentPackets);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchame
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "UBytesSent".
	 */
	class UnicastBytesSent extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+USentBytes);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchame
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
	

	/**
	 * This class manages simple statistic "MPacketsReceived".
	 */
	class MulticastPacketsReceived extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+MReceivedPackets);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchame
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "MBytesReceived".
	 */
	class MulticastBytesReceived extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+MReceivedBytes);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchame
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "MPacketsSent".
	 */
	class MulticastPacketsSent extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+MSentPackets);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchame
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}

	/**
	 * This class manages simple statistic "MBytesSent".
	 */
	class MulticastBytesSent extends SimpleStats
	{
		/**
		 * Gets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#getStats()
		 */
		protected String getStats()
		{
			return new String(""+MSentBytes);
		}
	
		/**
		 * Sets the value of a simple statistic managed by the implementation of this class.
		 * 
		 * @see SimpleStats#setStats(String newvalue)
		 */
		protected String setStats(String newValue) throws HyperCastStatsException
		{
			throw new HyperCastStatsException(statisticsName, HyperCastStatsException.WriteToAReadOnlyValue);
		}

		/** 
		 * @see I_Stats#getReadSchame
		 */
		public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException
		{
			return XmlUtil.createSchemaElement(doc/* document instance */, statisticsName/* element name */, "xsd:Long"/* type */, 
												null/* restriction base */, null/* pattern value */);
		}
	}
	
	
	/** 
	 * @return	the adapter manager
	 */
	public HC_AdapterManager getAdapterManager() {
		return manager;
	}
	/**
	 * @return	the physical node
	 */
	public physicalInterface getPhyNode() {
		return phy;
	}
	/**
	 * Handle a message
	 * @param msg
	 */
	public void handleMsg(HC_EventMsg msg) {
		//lotusDebug.errorln("msg=" + msg.getHCMessage());
		try {
			if(msg.dst==null){
				MReceivedPackets ++;
				MReceivedBytes += 8;
			}else{
				UReceivedPackets ++;
				UReceivedBytes += 8;
			}
			simHCNode protocol = (simHCNode)((OverlayNode)phy).getProtocol();
			if(protocol!=null){


				String dststr = protocol.getDetailedName();
				if(msg.getDst()==null)
					dststr = "Broadcast";
				
				String srcstr = "@Unknown";
				OverlayNode srcnode = msg.getSrc();
				if(srcnode !=null){
					simHCNode srcprotocol = (simHCNode)(srcnode).getProtocol();
					if(srcprotocol!=null){
						srcstr = srcprotocol.getDetailedName();
					}
				}
				protocol.checkEventListSize();
				protocol.eventList.add(new MessageHistoryEntry(protocol.network.simTime(), false, msg, srcstr, dststr));
			}
			if(callback!=null)
				callback.messageArrivedFromAdapter(msg.getHCMessage());
			
		} catch (Exception e) {
			lotusDebug.errorln("HC_Adapter#handleMsg: "+msg.getHCMessage().getClass().getName()+": "+e.toString());
			e.printStackTrace();
		}
	}
	/**
	 * Handle a timer
	 * @param timerID
	 */
	public void handleTimer(Object timerID) {
		try {
			
			simHCNode protocol = (simHCNode)((OverlayNode)phy).getProtocol();
			if(protocol!=null){
				protocol.checkEventListSize();
				protocol.eventList.add(new TimerHistoryEntry(protocol.network.simTime(), false, timerID, 0.0f));
			}
			
			callback.timerExpired(timerID);
		} catch (Exception e) {
			lotusDebug.errorln("HC_Adapter#handleTimer: "+e.toString());
			e.printStackTrace();
		}
	}

	/* (non-Javadoc)
	 * @see hypercast.I_MulticastAdapter#sendMulticastMessage(hypercast.I_Message)
	 */
	public void sendMulticastMessage(I_Message mp) {
		if (!started)
			return;
		
		MSentPackets ++;
		MSentBytes +=8;
		
		simHCNode protocol = (simHCNode)((OverlayNode)phy).getProtocol();
		HC_EventMsg msg;
		String label = "unknown";
		if(protocol!=null){
			String src = protocol.getDetailedName();
			String dst = "Broadcast";
			msg = new HC_EventMsg((OverlayNode)phy, null, mp, label);
			protocol.checkEventListSize();
			protocol.eventList.add(new MessageHistoryEntry(protocol.network.simTime(), true, msg, src, dst));
			label = protocol.agent.getName(); 
		}else
			msg = new HC_EventMsg((OverlayNode)phy, null, mp, label);
		
		phy.sendMsg(msg);
	}

	/* (non-Javadoc)
	 * @see hypercast.I_MulticastAdapter#getMulticastAddress()
	 */
	public I_PhysicalAddress getMulticastAddress() {
		return new SimPhysicalAddress(0);
	}

	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#Start()
	 */
	public void Start() {
		started = true;
	}

	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#Stop()
	 */
	public void Stop() {
		started = false;
	}

	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#sendUnicastMessage(hypercast.I_PhysicalAddress, hypercast.I_Message)
	 */
	public void sendUnicastMessage(I_NetworkAddress dst, I_Message mp) {
		if (!started)
			return;
		
		simHCNode protocol = (simHCNode)((OverlayNode)phy).getProtocol();
		
		USentPackets ++;
		USentBytes += 8;
		
		SimPhysicalAddress dstaddr = (SimPhysicalAddress)dst;
		OverlayNode dstnode=(OverlayNode)this.getAdapterManager().searchAddress(dstaddr);
		
		HC_EventMsg msg;
		
		if(protocol!=null){
			String src = protocol.getDetailedName();
			String dststr = "@"+dstaddr;
			if(dstnode !=null){
				simHCNode dstprotocol = (simHCNode)(dstnode).getProtocol();
				if(dstprotocol!=null){
					dststr = dstprotocol.getDetailedName();
				}
			}
			msg =
				new HC_EventMsg(
					(OverlayNode)phy,
					dstnode,
					mp, protocol.agent.getName());
			protocol.checkEventListSize();
			protocol.eventList.add(new MessageHistoryEntry(protocol.network.simTime(), true, msg, src, dststr));
		}else
			msg =
				new HC_EventMsg(
					(OverlayNode)phy,
					dstnode,
					mp, "unknown");
		
		if (dstnode != null) {
			phy.sendMsg(msg);
		}
		//else lotusDebug.errorln("dst not found for "+dstnode);
	}

	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#createPhysicalAddress(byte[], int)
	 */
	public I_PhysicalAddress createPhysicalAddress(
		byte[] byteaddr,
		int offset) {
		return new SimPhysicalAddress(byteaddr, offset);
	}

	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#getPhysicalAddress()
	 */
	public I_PhysicalAddress getPhysicalAddress() {
		return phy_addr;
	}

	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#setCallBack(hypercast.I_AdapterCallBack)
	 */
	public void setCallback(I_AdapterCallback node) {
		callback = node;
	}

	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#setTimer(int, long)
	 */
	public void setTimer(Object timer_id, long delay_ms) {
		if (!started)
			return;
		
		simHCNode protocol = (simHCNode)((OverlayNode)phy).getProtocol();
		if(protocol!=null){
			protocol.checkEventListSize();
			protocol.eventList.add(new TimerHistoryEntry(protocol.network.simTime(), true, timer_id, ((float)delay_ms) / 1000));
		}
		
		phy.setTimer(((float)delay_ms) / 1000, timer_id);
	}

	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#clearTimer(int)
	 */
	public void clearTimer(Object timer_id) {
		phy.clearTimer(timer_id);

	}

	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#getTimer(int)
	 */
	public long getTimer(Object timer_id) {
		return (long)(phy.getTimer(timer_id)*1000);
		
	}

	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#getCurrentTime()
	 */
	public long getCurrentTime() {
		return (long) (phy.simTime() * 1000);
	}
	
	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#startTimer()
	 */
	public void startTimer() {
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#stopTimer()
	 */
	public void stopTimer() {
		// TODO Auto-generated method stub
	}
	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#createPhysicalAddress(java.lang.String)
	 */
	public I_PhysicalAddress createPhysicalAddress(String addrString) {
		try {

			return new SimPhysicalAddress(Integer.parseInt(addrString));

		} catch (Exception e) {
			return null;
		}
	}
	
	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#createPhysicalAddress()
	 */
	public I_PhysicalAddress createPhysicalAddress() {
		return phy_addr;
	}
	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#getPhysicalAddressSize()
	 */
	public int getPhysicalAddressSize() {
		return SimPhysicalAddress.HC_ADDR_SIZE; 
	}
	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#createUnderlayAddress(byte[], int)
	 */
	public I_UnderlayAddress createUnderlayAddress(byte[] byteaddr, int offset) {
		return new SimPhysicalAddress(byteaddr, offset);
	}
	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#createUnderlayAddress(java.lang.String)
	 */
	public I_UnderlayAddress createUnderlayAddress(String addrString) {
		try {

			return new SimPhysicalAddress(Integer.parseInt(addrString));

		} catch (Exception e) {
			return null;
		}
	}
	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#createUnderlayAddress()
	 */
	public I_UnderlayAddress createUnderlayAddress() {
		return phy_addr;
	}
	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#getUnderlayAddressSize()
	 */
	public int getUnderlayAddressSize() {
		return SimPhysicalAddress.HC_ADDR_SIZE;
	}
	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#Reset(java.lang.Object)
	 */
	public boolean Reset(Object socketObj) {
		UReceivedPackets = 0;
		MReceivedPackets = 0;
		USentPackets = 0;
		MSentPackets = 0;
		UReceivedBytes = 0;
		MReceivedBytes = 0;
		USentBytes = 0;
		MSentBytes = 0;
		return true;
	}
	/* (non-Javadoc)
	 * @see hypercast.I_UnicastAdapter#getInfo()
	 */
	public Object getInfo() {
		return phy_addr;
	}
	
	/* (non-Javadoc)
	 * @see hypercast.I_Stats#getStatsName()
	 */
	public String getStatsName() {
		return statsName;
	}
	/* (non-Javadoc)
	 * @see hypercast.I_Stats#setStatsName(java.lang.String)
	 */
	public void setStatsName(String name) {
		statsName = name;
	}
	/* (non-Javadoc)
	 * @see hypercast.I_Stats#getStats(org.w3c.dom.Document, org.apache.xpath.XPath)
	 */
	public Element[] getStats(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getStatsResult(doc, xpath);
	}
	/* (non-Javadoc)
	 * @see hypercast.I_Stats#setStats(org.w3c.dom.Document, org.apache.xpath.XPath, org.w3c.dom.Element)
	 */
	public Element[] setStats(Document doc, XPath xpath, Element newValue) throws HyperCastStatsException {
		return statsPro.setStatsResult(doc, xpath, newValue);
	}
	/* (non-Javadoc)
	 * @see hypercast.I_Stats#getReadSchema(org.w3c.dom.Document, org.apache.xpath.XPath)
	 */
	public Element[] getReadSchema(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getReadSchemaResult(doc, xpath);
	}
	/* (non-Javadoc)
	 * @see hypercast.I_Stats#getWriteSchema(org.w3c.dom.Document, org.apache.xpath.XPath)
	 */
	public Element[] getWriteSchema(Document doc, XPath xpath) throws HyperCastStatsException {
		return statsPro.getWriteSchemaResult(doc, xpath);
	}
	

}


/**
 * This class creates and manages virtual adapters
 * used to run Hypercast Node 
 * 
 * @author Guangyu Dong, Feb 3, 2004
 */

class HC_AdapterManager {

	/** the list of all addresses */
	private Hashtable addr_list;

	/** the Id number used to generate physical address */
	//private int current_id;

	/**
	 * Constructor
	 */
	public HC_AdapterManager() {
		//current_id = 1;
		addr_list=new Hashtable();
	}
	/**
	 * Create an address by specifying the Id value
	 * @param phy		the physical node
	 * @param addr_id	the address Id
	 * @return
	 */
	private SimPhysicalAddress createAddress(int addr_id){
		//if(addr_id<5000)
			//throw new IllegalArgumentException("Specified PhysicalAddress cannot be smaller than [5000]");
		addr_id = this.findFirstAvailableId(addr_id);
		HC_Adapter adpt = (HC_Adapter)addr_list.get(""+addr_id);
		if (adpt != null){
			throw new IllegalArgumentException("PhysicalAddress ["+addr_id+"]already allocated to another node");
		}
		return new SimPhysicalAddress(addr_id);
		
	}
	
	public void reset(){
		addr_list.clear();
	}
	/**
	 * find the first free Id staring from start_id
	 * @param start_id
	 * @return
	 */
	private int findFirstAvailableId(int start_id){
		while(true){
			if(!addr_list.containsKey(""+start_id))
				return start_id;
			
			start_id++;
		}
	}
	/**
	 * create a physical address
	 * @param phy
	 * @return	address
	 */
	private SimPhysicalAddress createAddress() {
		return new SimPhysicalAddress(this.findFirstAvailableId(1));
	}
	/**
	 * create a virtual adapter
	 * @param phy
	 * @return	adapter
	 */
	public HC_Adapter createAdapter(OverlayNode phy) {
		SimPhysicalAddress addr = createAddress();
		HC_Adapter adapter = new HC_Adapter(phy, addr, this);

		addr_list.put(addr.toString(), adapter);

		return adapter;
	}
	/**
	 * Create a virtual adapter by specifying the address Id
	 * @param phy
	 * @return
	 */
	public HC_Adapter createAdapter(OverlayNode phy, int addr_id) {
			SimPhysicalAddress addr = createAddress(addr_id);
			HC_Adapter adapter = new HC_Adapter(phy, addr, this);

			addr_list.put(addr.toString(), adapter);

			return adapter;
	}
	/** 
	 * search the address by a string
	 * @param as	address string
	 * @return	address
	 */
	public physicalInterface searchAddress(String as) {
		HC_Adapter adpt = (HC_Adapter)addr_list.get(as);
		if (adpt != null)
			return adpt.getPhyNode();
		else
			return null;
	}
	/** 
	 * search the address by a string
	 * @param as	address string
	 * @return	address
	 */
	public physicalInterface searchAddress(SimPhysicalAddress ha) {
		HC_Adapter adpt = (HC_Adapter)addr_list.get(ha.toString());
		if (adpt != null)
			return adpt.getPhyNode();
		else
			return null;
	}
	/**
	 * remove an adapter's address from the list
	 * @param adpt	the adapter
	 */
	public void removeAdapter(HC_Adapter adpt) {
		addr_list.remove(adpt.getPhysicalAddress().toString());

	}

}




/**
 * This class wraps an I_Message object to make it transferable 
 * in Lotos.
 * 
 * @author Guangyu Dong, Feb 2, 2004
 */
class HC_EventMsg extends SimEventMsg{
	/** message type: only one type */
	final static int HC_MSG_WRAPPER=0;	
	
	/** the hypercast protocol message */
	private I_Message hc_message;
	
	/** the label */
	private String label;
	
	/**
	 * Constructor
	 * @param osrc		source node
	 * @param odst		destination node
	 * * @param hc_msg	the I_Message object		
	 */
	public HC_EventMsg(OverlayNode osrc, OverlayNode odst, I_Message hc_msg, String label) {
		super(osrc, odst, HC_MSG_WRAPPER);
		hc_message=hc_msg;	
		this.label = label;
	}
	/**
	 * @return	the I_Message object contained in this class
	 */
	public I_Message getHCMessage(){
		return hc_message;
	}
	/* (non-Javadoc)
	 * @see SimEventMsg#getMsgLabel(int)
	 */
	public String getMsgLabel(int type) {
		//return label;
		//return "HC Message";
		return hc_message.toString();
	}

	/* (non-Javadoc)
	 * @see SimEventMsg#clone()
	 */
	public Object clone() {
		return new HC_EventMsg(this.getSrc(), this.getDst(), hc_message, label);
	}
	
	
	
}

/**
 * This class parse the configuration file and create a list of agent class
 * @author Guangyu Dong
 *
 */
class SimHypercastConfig{
	
	private LinkedList protocols;
	
	private LinkedList securitySettings;
	
	private simHCNetwork network;
	
	public SimHypercastConfig(simHCNetwork network, String config_xml) throws ClassNotFoundException, InstantiationException, IllegalAccessException, IOException {
		this.network = network;
		protocols = new LinkedList();
		securitySettings = new LinkedList();
		InputStream configInput = FileResourceManager.openFile(config_xml);
		
		/*
		try {
			configInput = new FileInputStream(config_xml);
		} catch (Exception e) {
			System.out.println("Couldn't find simulation config file in local directory, use the embedded config instead.");
			configInput = getClass().getResourceAsStream(config_xml);
		}
		*/
		parseConfig(XmlUtil.createDocument(configInput, config_xml));
		
	}
	public int getProtocolCount(){
		return protocols.size();
	}
	public int getSecuritySettingCount(){
		return this.securitySettings.size();
	}
	public ProtocolEntry getProtocolByName(String protocolName){
		for(int i=0;i<protocols.size();i++){
			ProtocolEntry agent = (ProtocolEntry)protocols.get(i);
			if(agent.name.equals(protocolName))
				return agent;
		}
		return null;
	}
	public SecuritySettingEntry getSecuritySettingByName(String name){
		for(int i=0;i<securitySettings.size();i++){
			SecuritySettingEntry sse = (SecuritySettingEntry)securitySettings.get(i);
			if(sse.name.equals(name))
				return sse;
		}
		return null;
	}
	public String [] getProtocolNames(){
		String []names = new String[protocols.size()];
		for(int i=0;i<protocols.size();i++){
			ProtocolEntry entry = (ProtocolEntry)protocols.get(i);
			names[i] = entry.getName();
		}
		return names;
	}
	public String [] getSecuritySettingNames(){
		String []names = new String[securitySettings.size()];
		for(int i=0;i<securitySettings.size();i++){
			SecuritySettingEntry sse = (SecuritySettingEntry)securitySettings.get(i);
			names[i] = sse.name;
		}
		return names;
	}
	
	private void parseConfig(Document config_doc) throws ClassNotFoundException, InstantiationException, IllegalAccessException{
		Node simHypercast = config_doc.getElementsByTagName("SimHypercast").item(0);
		
		/** get protocol configs */
		NodeList hypercastProtocols = ((Element)simHypercast).getElementsByTagName("HypercastProtocol");
		
		for(int i=0;i<hypercastProtocols.getLength();i++){
			String name = ((Element)hypercastProtocols.item(i)).getAttribute("name");
			Node agentClass = ((Element)hypercastProtocols.item(i)).getElementsByTagName("AgentClass").item(0);
			Node properties = ((Element)hypercastProtocols.item(i)).getElementsByTagName("Properties").item(0);
			Node configNode = properties.getChildNodes().item(0);
			String agentClassName = "lotos."+agentClass.getChildNodes().item(0).getNodeValue();
			lotusDebug.println("Loading HyperCast protocol agent '"+agentClassName+"'");
			
			Document hc_config_doc = XmlUtil.createDocument();
			Node newConfigNode = hc_config_doc.importNode(configNode, true);
			hc_config_doc.appendChild(newConfigNode);
			
			HyperCastConfig hc_config = new HyperCastConfig(hc_config_doc);
			
			/* dynamically load agent class and set the config */
			Class dynamicAgentClass = Class.forName(agentClassName);
			I_SimHyperCastAgent agentObject = (I_SimHyperCastAgent)dynamicAgentClass.newInstance();
			agentObject.setNetwork(network);
			ProtocolEntry entry = new ProtocolEntry(name, hc_config, agentObject);
			
			protocols.add(entry);
		}
		
		/** get security configs */
		NodeList securitySettings = ((Element)simHypercast).getElementsByTagName("SecuritySetting");
		
		for(int i=0;i<securitySettings.getLength();i++){
			String name = ((Element)securitySettings.item(i)).getAttribute("name");
			Node properties = ((Element)securitySettings.item(i)).getElementsByTagName("Properties").item(0);
			NodeList privateProperties = ((Element)securitySettings.item(i)).getElementsByTagName("PrivateProperty");
			Node configNode = properties.getChildNodes().item(0);
			
			lotusDebug.println("Loading security setting '"+name+"'");
			
			Document hc_config_doc = XmlUtil.createDocument();
			Node newConfigNode = hc_config_doc.importNode(configNode, true);
			hc_config_doc.appendChild(newConfigNode);
			
			HyperCastConfig hc_config = new HyperCastConfig(hc_config_doc);
			
			for(int j=0;j<privateProperties.getLength();j++){
				String ppName = ((Element)privateProperties.item(j)).getAttribute("XPath");
				String ppValue = ((Element)privateProperties.item(j)).getAttribute("value");
				hc_config.setPrivateTextAttribute(XmlUtil.createXPath(ppName), ppValue);
			}
			
			SecuritySettingEntry sse = new SecuritySettingEntry(name, hc_config);
						
			this.securitySettings.add(sse);
		}
		
	}
	
	
}
class SecuritySettingEntry{
	
	String name;
	HyperCastConfig config;
	
	public SecuritySettingEntry(String name, HyperCastConfig config){
		this.name = name;
		this.config = config;
	}
}

class ProtocolEntry{
	String name;
	HyperCastConfig config;
	I_SimHyperCastAgent agent;
	public ProtocolEntry(String name ,HyperCastConfig config, I_SimHyperCastAgent agent){
		this.name = name;
		this.config = config;
		this.agent = agent;
	}
	/**
	 * @return Returns the agent.
	 */
	public I_SimHyperCastAgent getAgent() {
		return agent;
	}
	/**
	 * @return Returns the config.
	 */
	public HyperCastConfig getConfig() {
		return config;
	}
	/**
	 * @return Returns the name.
	 */
	public String getName() {
		return name;
	}
}


